/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "endpointstartstopbutton.h"

namespace Conexus
{
  namespace Gtk
  {

    EndpointStartStopButton::EndpointStartStopButton ( Conexus::Endpoint::pointer endpoint, bool show_text, bool show_image ) :
        m_endpoint( endpoint ),
        m_start_image ( &m_default_start_image ),
        m_default_start_image ( ::Gtk::Stock::CONNECT, ::Gtk::IconSize ( ::Gtk::ICON_SIZE_LARGE_TOOLBAR ) ),
        m_stop_image ( &m_default_stop_image ),
        m_default_stop_image ( ::Gtk::Stock::DISCONNECT, ::Gtk::IconSize ( ::Gtk::ICON_SIZE_LARGE_TOOLBAR ) )
    {
      m_text.set_use_markup();

      this->set_start_text_to_default();
      this->set_stop_text_to_default();

      m_default_start_image.show();
      m_default_stop_image.show();

      if ( show_text )
        this->show_text( );
      else
        this->hide_text( );

      if ( show_image )
        this->show_image( );
      else
        this->hide_image( );

      m_hbox.set_spacing ( 5 );
      m_hbox.pack_start ( m_imagebox, ::Gtk::PACK_SHRINK );
      m_hbox.pack_start ( m_text, ::Gtk::PACK_SHRINK );
      m_hbox.show();
      set_image ( m_hbox );

      this->update_display( );

      this->signal_toggled().connect ( sigc::mem_fun ( *this, &EndpointStartStopButton::on_button_toggled ) );
    }


    EndpointStartStopButton::~EndpointStartStopButton()
    {}

    void EndpointStartStopButton::set_endpoint ( Conexus::Endpoint::pointer io )
    {
      if ( m_endpoint == io ) return;

      if ( m_endpoint_start_stop_connection ) m_endpoint_start_stop_connection.disconnect();

      m_endpoint = io;

      if ( m_endpoint )
      {
        m_endpoint_start_stop_connection =
            m_endpoint->signal_start_stop().connect(sigc::mem_fun( *this, &EndpointStartStopButton::on_external_start_stop));

        if ( this->get_active() )
          m_endpoint->start();
        else
          m_endpoint->stop();
      }
    }

    Conexus::Endpoint::pointer EndpointStartStopButton::endpoint() {
      return m_endpoint;
    }

    const Glib::ustring & EndpointStartStopButton::start_text( )
    {
      return m_start_text;
    }

    const Glib::ustring & EndpointStartStopButton::stop_text( )
    {
      return m_stop_text;
    }

    void EndpointStartStopButton::set_start_text ( const Glib::ustring & text )
    {
      m_start_text = text;
      this->update_display();
    }

    void EndpointStartStopButton::set_start_text_to_default( )
    {
      this->set_start_text ( "Connected  <i><small>( Click to Disconnect )</small></i>" );
    }

    void EndpointStartStopButton::set_stop_text ( const Glib::ustring & text )
    {
      m_stop_text = text;
      this->update_display();
    }

    void EndpointStartStopButton::set_stop_text_to_default( )
    {
      this->set_stop_text ( "Disconnected  <i><small>( Click to Connect )</small></i>" );
    }

    bool EndpointStartStopButton::is_text_visible( )
    {
      return m_text.is_visible();
    }

    void EndpointStartStopButton::show_text( )
    {
      m_text.show();
    }

    void EndpointStartStopButton::hide_text( )
    {
      m_text.hide();
    }

    const ::Gtk::Image & EndpointStartStopButton::start_image( )
    {
      return *m_start_image;
    }

    const ::Gtk::Image & EndpointStartStopButton::stop_image( )
    {
      return *m_stop_image;
    }

    void EndpointStartStopButton::set_start_image ( ::Gtk::Image & image )
    {
      m_start_image = &image;
      this->update_display();
    }

    void EndpointStartStopButton::set_start_image_to_default( )
    {
      m_start_image = &m_default_start_image;
      this->update_display();
    }

    void EndpointStartStopButton::set_stop_image ( ::Gtk::Image & image )
    {
      m_stop_image = &image;
      this->update_display();
    }

    void EndpointStartStopButton::set_stop_image_to_default( )
    {
      m_stop_image = &m_default_stop_image;
      this->update_display();
    }

    bool EndpointStartStopButton::is_image_visible( )
    {
      return m_imagebox.is_visible();
    }

    void EndpointStartStopButton::show_image( )
    {
      m_imagebox.show_all();
    }

    void EndpointStartStopButton::hide_image( )
    {
      m_imagebox.hide();
    }

    void EndpointStartStopButton::on_button_toggled( )
    {
      this->update_display();

      if ( ! m_endpoint ) return;

      if ( this->get_active() && ! m_endpoint->is_running() )
        m_endpoint->start();
      else if ( ! this->get_active() && m_endpoint->is_running() )
        m_endpoint->stop();
    }

    void EndpointStartStopButton::update_display( )
    {
      if ( this->get_active() )
      {
        if ( m_imagebox.get_child() != NULL )
          m_imagebox.remove();
        m_imagebox.add ( *m_start_image );
        m_text.set_markup ( m_start_text );
      }
      else
      {
        if ( m_imagebox.get_child() != NULL )
          m_imagebox.remove();
        m_imagebox.add ( *m_stop_image );
        m_text.set_markup ( m_stop_text );
      }
    }

    void EndpointStartStopButton::on_external_start_stop(StartStop ss)
    {
      this->set_active( ss );
    }

  }

}

