/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSGTKMMENDPOINTSTARTSTOPBUTTON_H
#define CONEXUSGTKMMENDPOINTSTARTSTOPBUTTON_H

#include <map>
#include <gtkmm.h>

#include <conexus/endpoint.h>

namespace Conexus
{
  namespace Gtk
  {

    /**
     * This widget can control the stop/start state of a single Conexus endpoint.
     *
     * @author Rick L Vinyard Jr
     */
    class EndpointStartStopButton : public ::Gtk::CheckButton
    {
      public:

        /**
         * Basic constructor
         * @param show_text If true, the start/stop text label will be shown
         * @param show_image If true, the start/stop image will be shown
         */
        EndpointStartStopButton ( Conexus::Endpoint::pointer io = Conexus::Endpoint::pointer(),
                                  bool show_text = true,
                                  bool show_image = true );

        /** Destructor */
        ~EndpointStartStopButton();

        /**
         * Place a Conexus Endpoint under the control of this widget
         * @param io The Conexus endpoint to control
         * 
         * Note: Removes the previously set endpoint
         *
         * Call set_endpoint() with no parameters to remove the current
         * endpoint and leave the widget controlling nothing.
         */
        void set_endpoint ( Conexus::Endpoint::pointer io = Conexus::Endpoint::pointer() );

        /**
         * Returns the Conexus Endpoint under control
         */
        Conexus::Endpoint::pointer endpoint();
        
        /** The text to be displayed when the widget is in a started state */
        const Glib::ustring& start_text();

        /** The text to be displayed when the widget is in a stopped state */
        const Glib::ustring& stop_text();

        /**
         * Set the start state text. Pango markup is used for display, so if desired
         * the text may also contain pango markup.
         */
        void set_start_text ( const Glib::ustring& text );

        /**
         * Reset the start text to the default value of "Connected  \<i\>\<small\>( Click to Disconnect )\</small\>\</i\>"
         */
        void set_start_text_to_default();

        /**
         * Set the stop state text. Pango markup is used for display, so if desired
         * the text may also contain pango markup.
         */
        void set_stop_text ( const Glib::ustring& text );

        /**
         * Reset the stop text to the default value of "Disconnected  \<i\>\<small\>( Click to Connect )\</small\>\</i\>"
         */
        void set_stop_text_to_default();

        /** True if the text label will be shown, false if it will be hidden. */
        bool is_text_visible();

        /** Shows the text label */
        void show_text( );

        /** Hides the text label */
        void hide_text();

        /** The image to display when the widget is in a started state. */
        const ::Gtk::Image& start_image();

        /** The image to display when the widget is in a stopped state. */
        const ::Gtk::Image& stop_image();

        /** Set the start state image. */
        void set_start_image ( ::Gtk::Image& image );

        /** Reset the start state image to the default stock connect icon. */
        void set_start_image_to_default();

        /** The image to display when the widge tis in a stopped state. */
        void set_stop_image ( ::Gtk::Image& image );

        /** Reset the stopped state image to the default stock disconnect icon. */
        void set_stop_image_to_default();

        /** True if the start/stop image is shown; false if hidden. */
        bool is_image_visible();

        /** Shows the start/stop image */
        void show_image( );

        /** Hides the start/stop image */
        void hide_image( );

      protected:
        /** The endpoint under control */
        Conexus::Endpoint::pointer m_endpoint;

        /** Holds the text and image widgets */
        ::Gtk::HBox m_hbox;

        /** Container for image */
        ::Gtk::EventBox m_imagebox;

        /** Widget displays the current stop/start text */
        ::Gtk::Label m_text;

        /** The text displayed when in a started state */
        Glib::ustring m_start_text;

        /** The text displayed when in a stopped state */
        Glib::ustring m_stop_text;

        /** The image displayed when in a started state */
        ::Gtk::Image* m_start_image;

        /** Default start image */
        ::Gtk::Image m_default_start_image;

        /** The image displayed when in a stopped state */
        ::Gtk::Image* m_stop_image;

        /** Default stop image */
        ::Gtk::Image m_default_stop_image;

        /**
         * Used to disconnect on_external_start_stop() when the endpoint is
         * removed or changed.
         */
        sigc::connection m_endpoint_start_stop_connection;

        /** Callback to handle start/stop clicks */
        void on_button_toggled();

        /** Updates the display */
        void update_display( );

        /**
         * Connected to the controlled endpoint's start/stop signal
         */
        void on_external_start_stop( StartStop );

    };

  }

}

#endif
