/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "file.h"

namespace Conexus
{

  namespace NSPR
  {

    File::File ( const std::string& filename, long unsigned mode ) :
        m_eof ( true ),
        m_throw_eof ( false )
    {
      if ( mode & FILE_UNCHANGED )
        mode = mode & ~FILE_UNCHANGED;
      this->set_filename ( filename, mode );
    }

    File::pointer File::create ( const std::string& filename, long unsigned mode )
    {
      return pointer ( new File ( filename, mode ) );
    }

    File::~File()
    {}

    void File::open ( const std::string name, long unsigned mode ) throw ( open_exception )
    {
      int flags=0;

      if ( is_open() )
        change_state ( ENDPOINT_CLOSED );

      if ( ! ( mode & FILE_UNCHANGED ) )
        m_mode = mode;

      if ( m_mode == 0 ) throw exception::open::invalid_file_mode();

      if ( m_mode & ( FILE_READ|FILE_WRITE ) )
        flags |= PR_RDWR;
      else if ( m_mode & FILE_READ )
        flags |= PR_RDONLY;
      else if ( m_mode & FILE_WRITE )
        flags |= PR_WRONLY;

      if ( m_mode & FILE_APPEND )
        flags |= PR_APPEND;
      if ( m_mode & FILE_CREATE )
        flags |= PR_CREATE_FILE;
      // These flags don't have a NSPR equivalent
//       if ( m_mode & FILE_DIRECT )
//         flags |= O_DIRECT;
//       if ( m_mode & FILE_LARGE_FILE )
//         flags |= O_LARGEFILE;
//       if ( m_mode & FILE_NO_ACCESS_TIME )
//         flags |= O_NOATIME;
//       if ( m_mode & FILE_NO_FOLLOW )
//         flags |= O_NOFOLLOW;
      if ( m_mode & FILE_TRUNCATE )
        flags |= PR_TRUNCATE;
      if ( m_mode & FILE_EXCLUSIVE_CREATE )
        flags |= PR_EXCL;

      // This is the only NSPR specific flag
      if ( m_mode & FILE_SYNC )
        flags |= PR_SYNC;

      m_fd = PR_Open( name.c_str(), flags, 0 );

      // TODO throw a proper exception
      if ( m_fd == NULL )
        throw -1;

      m_eof = false;

      this->set_state_opened();
    }

    void File::open( ) throw ( open_exception )
    {
      if ( m_filename == std::string() )
        throw ( exception::open::no_filename() );
      this->open ( m_filename );
    }

    std::string File::filename( )
    {
      return m_filename;
    }

    long unsigned File::mode( )
    {
      return m_mode;
    }

    void File::set_filename ( std::string d, long unsigned mode ) throw ( open_exception )
    {
      long tempstate;

      // if the filename is the same and the state is the same, then leave
      if ( m_filename == d && m_mode == mode )
        return;

      tempstate = m_state;

      if ( this->is_open() )
        this->close();

      m_filename = d;
      if ( ! ( m_mode & FILE_UNCHANGED ) )
        m_mode = mode;

      this->change_state ( tempstate );

      m_signal_filename_changed.emit();
    }

    sigc::signal<void> File::signal_filename_changed()
    {
      return m_signal_filename_changed;
    }

    Data File::read_data ( long int to, size_t s ) throw ( read_exception )
    {
      Data d = FileDescriptor::read_data ( to, s );
      if ( ! d )
      {
        m_eof = true;
        if ( m_throw_eof ) throw exception::read::eof();
        return Data();
      }
      return d;
    }

    bool File::eof( )
    {
      return m_eof;
    }

    bool File::throw_eof( )
    {
      return m_throw_eof;
    }

    void File::set_throw_eof ( bool teof )
    {
      m_throw_eof = teof;
    }

    bool File::is_file() const
    {
      PRFileInfo info;
      PRStatus result;

      if ( m_filename.size() == 0 ) return false;

      result = PR_GetFileInfo( m_filename.c_str(), &info );

      if ( result == PR_FAILURE ) return false;

      return info.type == PR_FILE_FILE;
    }

    bool File::is_directory() const
    {
      PRFileInfo info;
      PRStatus result;

      if ( m_filename.size() == 0 ) return false;

      result = PR_GetFileInfo( m_filename.c_str(), &info );

      if ( result == PR_FAILURE ) return false;

      return info.type == PR_FILE_DIRECTORY;
    }

    bool File::is_other_file_type() const
    {
      PRFileInfo info;
      PRStatus result;

      if ( m_filename.size() == 0 ) return false;

      result = PR_GetFileInfo( m_filename.c_str(), &info );

      if ( result == PR_FAILURE ) return false;

      return info.type == PR_FILE_OTHER;
    }

    size_t File::size() const
    {
      PRFileInfo info;
      PRStatus result;

      if ( m_filename.size() == 0 ) return 0;

      result = PR_GetFileInfo( m_filename.c_str(), &info );

      if ( result == PR_FAILURE ) return 0;

      return info.size;
    }

  }

}
