/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_NSPRFILE_H
#define CONEXUS_NSPRFILE_H

#include <conexus/enums.h>
#include <conexus-nspr/filedescriptor.h>

namespace Conexus
{

  namespace NSPR
  {

    /**
     * Encapsulates a NSPR file.
     * 
     * TODO review NSPR documentation for completeness
     * TODO add support for PR_GetFileInfo
     * 
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup NSPR
     * @ingroup Endpoint
     */
    class File : public FileDescriptor
    {
      protected:

        File ( const std::string& filename,
               long unsigned mode );

      public:

        typedef ConexusPointer<File> pointer;

        CONEXUS_ENDPOINT_GENERIC_CREATE();

        static File::pointer create ( const std::string& filename = std::string(),
                                      long unsigned mode = 0 );

        virtual ~File();

        /**
        * Opens the named file in the specified mode.
         */
        virtual void open ( const std::string name,
                            long unsigned mode = FILE_UNCHANGED )
        throw ( open_exception );

        /**
        * Will attempt to open the file if a filename has been specified. Otherwise an
        * exception will be thrown.
         */
        virtual void open() throw ( open_exception );

        /** Will return the filename. */
        std::string filename();

        long unsigned mode();

        bool eof();

        bool throw_eof();

        void set_throw_eof ( bool teof = true );

        bool is_file() const;

        bool is_directory() const;

        bool is_other_file_type() const;

        size_t size() const;

        /**
        * Sets the filename.
        *
        * If the file is currently open and d is not equal to the current filename,
        * the current file will be closed.
        *
        * Also, if auto_open is set, the new file will be immediately opened.
        *
        * Additionally, this method checks whether the parameter d is the same as the current filename,
        * and parameter mode is the same as the current mode.
        * If both are equal, the file will not be closed and reopened unless force=true.
         */
        void set_filename ( std::string d, long unsigned mode = FILE_UNCHANGED ) throw ( open_exception );

        sigc::signal<void> signal_filename_changed();

      protected:
        std::string m_filename;
        long unsigned m_mode;
        bool m_eof;
        bool m_throw_eof;

        sigc::signal<void> m_signal_filename_changed;

        /**
        * Reimplements Endpoint::read_data() using FileDescriptor::read_data()
        * to catch the exception where data read size is zero indicating EOF.
        * When encountered throws eof exception.
        * @throw Conexus::exception::read::eof When file EOF is encounter on a zero length read
         */
        virtual Data read_data ( long int timeout, size_t s = 0 ) throw ( read_exception );

    };

  }

}

#endif
