/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include <conexus-nspr/filedescriptor.h>

#include <errno.h>

namespace Conexus
{

  namespace NSPR
  {

    FileDescriptor::FileDescriptor ( PRFileDesc* fd, bool close_on_destruct ) :
        Endpoint ( close_on_destruct ),
        m_fd ( fd )
    {}


    FileDescriptor::~FileDescriptor()
    {
      if ( m_close_on_destruct && this->is_open() )
        this->close();
    }

    PRFileDesc* FileDescriptor::fd() throw ()
    {
      return m_fd;
    }

    void FileDescriptor::close ( bool force ) throw ( close_exception )
    {
      if ( this->is_closed() && !force )
        return ;

      PRStatus i = ::PR_Close ( m_fd );
      // TODO throw a real exception
      if ( i != PR_SUCCESS )
        throw -1;

      this->set_state_closed();
    }

    size_t FileDescriptor::input_available() throw()
    {
      int n = PR_Available ( m_fd );
      if ( n == -1 ) return 0;
      return n;
    }

    size_t FileDescriptor::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
    {
      // TODO honor the timeout parameter
      PRInt32 bytes;

      if ( ! data ) return 0;

      if ( this->is_closed() ) this->open();

      bytes = PR_Write ( m_fd, data, data.size() );

      if ( bytes == -1 )
      {
        PRErrorCode code = PR_GetError();
        throw write_exception ( code );
      }

      return bytes;
    }

    Data FileDescriptor::read_data ( size_t size, Timeout timeout ) throw ( read_exception )
    {
      // If the endpoint is closed, lets try and open it before proceeding
      if ( this->is_closed() ) this->open();

      // Only a non-blocking immediate read needs O_NONBLOCK set
      if ( timeout.is_zero() )
      {
        // Switch to non-blocking if necessary
        PRStatus status;
        PRSocketOptionData opt;
        opt.option = PR_SockOpt_Nonblocking;
        opt.value.non_blocking = PR_TRUE;
        status = PR_SetSocketOption( m_fd, &opt );
        if ( status != PR_SUCCESS ) throw -1;
      }
      else
      {
        PRStatus status;
        PRSocketOptionData opt;
        opt.option = PR_SockOpt_Nonblocking;
        status = PR_GetSocketOption( m_fd, &opt );
        // TODO throw a proper exception
        if ( status != PR_SUCCESS ) throw -1;

        // Switch to blocking if necessary
        if ( opt.value.non_blocking == PR_TRUE )
        {
          opt.value.non_blocking = PR_FALSE;
          status = PR_SetSocketOption( m_fd, &opt );
          if ( status != PR_SUCCESS ) throw -1;
        }

        // If we have a timeout value, then we need to wait in select()
        if ( ! timeout.is_null() )
        {
          int n;
          PRPollDesc poll_desc[ 1 ];
          PRIntervalTime interval_time;

          poll_desc[ 0 ].in_flags = PR_POLL_READ | PR_POLL_EXCEPT;
          poll_desc[ 0 ].fd = m_fd;

          interval_time = ::PR_MicrosecondsToInterval ( timeout.sec * 1000000 + timeout.nsec / 1000 );
          n = ::PR_Poll ( poll_desc, 1, interval_time );

          if ( n == 0 ) throw_read_exception( ETIMEDOUT );
          if ( n < 0 ) throw -1; // TODO throw a real exception
        }
      }

      // We need to size the read
      // if read size is zero then we will read all available
      if ( size == 0 )
      {
        int32_t available;
        if ( (available=PR_Available(m_fd)) > 0 ) size = available;
      }

      // if size is <= zero then we will use the default read size
      if ( size <= 0 ) size = this->fallback_read_size();

      // resize the data object to hold the entire read buffer
      Data data(size);
      PRSize read_result;

      // perform the actual read
      read_result = PR_Read ( m_fd, data, size );

      if ( read_result < 0 ) throw -1; // TODO throw a real exception

      if ( m_timestamp ) data.set_current_time();

      if ( data.size() != ( unsigned ) read_result ) data.resize ( read_result );

      return data;
    }

  }
}
