/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include <conexus-nspr/socket.h>

#include <errno.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>

#include <conexus/utility.h>

namespace Conexus
{

  namespace NSPR
  {

    Socket::Socket( ) throw () :
        FileDescriptor(),
        m_remote_address_set ( false ),
        m_non_blocking ( true )
    {
      m_local_address.signal_changed().connect ( sigc::mem_fun ( *this, &Socket::on_local_address_changed_proxy ) );
      m_remote_address.signal_changed().connect ( sigc::mem_fun ( *this, &Socket::on_remote_address_changed_proxy ) );
    }

    Socket::~Socket() throw ()
    {}

    void Socket::close ( bool force ) throw ( close_exception )
    {
      PRStatus i;

      if ( is_closed() && !force )
        return ;

      i = ::PR_Close ( m_fd );

      // TODO throw a proper exception
      if ( i != PR_SUCCESS ) throw -1;

      m_fd = NULL;

      this->set_state_closed();
    }


    void Socket::bind() throw ( bind_exception )
    {
      this->bind ( m_local_address );
    }

    void Socket::bind ( const Conexus::Address& a ) throw ( bind_exception )
    {
      PRStatus i;
      PRNetAddr nspraddr;

      if ( this->is_closed() )
        try
        {
          this->change_state ( ENDPOINT_OPENED );
        }
        catch ( exception::state::failed )
        {
          throw exception::bind::not_open();
        }

      addr_to_nspr ( a, nspraddr );
      i = PR_Bind ( m_fd, &nspraddr );

      // TODO throw a proper exception
      if ( i != PR_SUCCESS ) throw -1;

      this->set_state_bound();
    }

    void Socket::connect() throw ( connect_exception )
    {
      throw exception::connect::no_socket_connect();
    }

    void Socket::connect ( const Address& a ) throw ( connect_exception )
    {
      PRStatus i;
      PRNetAddr nspraddr;
      //       PRSocketOptionData option;

      if ( this->is_closed() )
        try
        {
          this->change_state ( ENDPOINT_OPENED );
        }
        catch ( exception::state::failed )
        {
          throw exception::connect::not_open();
        }

      addr_to_nspr ( a, nspraddr );
      // TODO come back and put in real timeouts


//       PRSocketOptionData socketOption;
//       socketOption.option = PR_SockOpt_Nonblocking;
//       PRBool old_value;
//       PR_GetSocketOption( m_fd, &socketOption );
//       old_value = socketOption.value.non_blocking;
//       socketOption.value.non_blocking = PR_FALSE;
//       PR_SetSocketOption( m_fd, &socketOption );

      i = PR_Connect ( m_fd, &nspraddr, PR_INTERVAL_NO_TIMEOUT );

      // TODO throw a proper exception
      if ( i != PR_SUCCESS ) throw -1;

      this->set_state_bound();
      this->set_state_connected();

//       socketOption.value.non_blocking = old_value;
//       PR_SetSocketOption( m_fd, &socketOption );
    }

    void Socket::listen ( int backlog )
    {
      PRStatus i;

      if ( ! this->is_bound() )
        try
        {
          this->change_state ( SOCKET_BOUND );
        }
        catch ( exception::state::failed )
        {
          throw exception::listen::not_bound();
        }

      i = PR_Listen ( m_fd, backlog );

      // TODO throw a proper exception
      if ( i != PR_SUCCESS ) throw -1;

      this->set_state_listening();
    }

    ssize_t Socket::writeto ( Address& a, const Data data ) throw ( write_exception )
    {
      PRInt32 ret;
      PRNetAddr nspraddr;

      if ( this->is_closed() )
        try
        {
          this->change_state ( ENDPOINT_OPENED );
        }
        catch ( exception::state::failed )
        {
          throw exception::write::not_socket();
        }

      addr_to_nspr ( a, nspraddr );
      // TODO put in proper interval handling
      ret = PR_SendTo ( m_fd, data, data.size(), 0, &nspraddr, PR_INTERVAL_NO_TIMEOUT );

      // TODO throw a proper exception
      if ( ret == -1 ) throw -1;

      return ret;
    }

    void Socket::on_local_address_changed()
    {
    }

    void Socket::on_remote_address_changed()
    {
    }

    void Socket::on_local_address_changed_proxy()
    {
      this->on_local_address_changed( );
    }

    void Socket::on_remote_address_changed_proxy()
    {
      this->on_remote_address_changed( );
      m_remote_address_set = true;
    }

    void Socket::set_option ( PRSocketOptionData& option )
    {
      PRStatus s;

      s = PR_SetSocketOption ( m_fd, &option );

      // TODO throw a proper exception
      if ( s != PR_SUCCESS ) throw -1;
    }

    void Socket::change_state ( long states ) throw ( state_exception )
    {
      // check for inconsistent states
      // if it's CLOSED it can't be BOUND, CONNECTED, LISTENING or ACCEPTED
      if ( ( states & ENDPOINT_CLOSED ) && ( ( states & SOCKET_BOUND ) || ( states & SOCKET_CONNECTED ) || ( states & SOCKET_LISTENING ) || ( states & SOCKET_ACCEPTED ) ) )
        throw exception::state::inconsistent();

      // call for CLOSED, OPENED and UNCHANGED
      // also checks for existing state
      // if it does the job, go ahead and return
      Endpoint::change_state ( states );

      if ( ! this->is_bound() && states & SOCKET_BOUND )
        bind();
      if ( ! this->is_connected() && states & SOCKET_CONNECTED )
        connect();
      if ( ! this->is_listening() && states & SOCKET_LISTENING )
        listen();
    }

    sigc::signal<void> Socket::signal_bound()
    {
      return m_signal_bound;
    }

    sigc::signal<void> Socket::signal_connected()
    {
      return m_signal_connected;
    }

    sigc::signal<void> Socket::signal_listening()
    {
      return m_signal_listening;
    }

    bool Socket::is_bound()
    {
      return m_state & SOCKET_BOUND;
    }

    bool Socket::is_connected()
    {
      return m_state & SOCKET_CONNECTED;
    }

    bool Socket::is_listening()
    {
      return m_state & SOCKET_LISTENING;
    }

    bool Socket::is_accepted()
    {
      return m_state & SOCKET_ACCEPTED;
    }

    void Socket::set_state_closed()
    {
      m_state &= ~SOCKET_BOUND;
      m_state &= ~SOCKET_CONNECTED;
      m_state &= ~SOCKET_LISTENING;
      m_state &= ~SOCKET_ACCEPTED;
      if ( this->is_closed() ) return;
      Endpoint::set_state_closed();
    }

    void Socket::set_state_bound()
    {
      if ( this->is_bound() ) return;
      m_state |= SOCKET_BOUND;
      m_signal_bound.emit();
    }

    void Socket::set_state_connected()
    {
      if ( this->is_connected() ) return;
      m_state |= SOCKET_CONNECTED;
      m_signal_connected.emit();
    }

    void Socket::set_state_listening()
    {
      if ( this->is_listening() ) return;
      m_state |= SOCKET_LISTENING;
      m_signal_listening.emit();
    }

    bool Socket::set_non_blocking(bool set)
    {
      if ( this->is_open() )
      {
        PRStatus status;
        PRSocketOptionData opt;
        opt.option = PR_SockOpt_Nonblocking;
        opt.value.non_blocking = set ? PR_TRUE : PR_FALSE;
        status = PR_SetSocketOption( m_fd, &opt );
        if ( status == PR_SUCCESS ) m_non_blocking = set;
        return status == PR_SUCCESS;
      }

      m_non_blocking = set;
      return true;
    }

    bool Socket::is_non_blocking()
    {
      if ( m_fd != NULL )
      {
        PRStatus status;
        PRSocketOptionData opt;
        opt.option = PR_SockOpt_Nonblocking;
        status = PR_GetSocketOption( m_fd, &opt );
        // TODO throw a proper exception
        if ( status != PR_SUCCESS ) throw -1;
        m_non_blocking = opt.value.non_blocking == PR_TRUE;
        return m_non_blocking;
      }
      else
      {
        return m_non_blocking;
      }
    }

    Conexus::IPv4::Address & Socket::local_address( )
    {
      return m_local_address;
    }

    Conexus::IPv4::Address & Socket::remote_address( )
    {
      return m_remote_address;
    }

    void Socket::set_remote_address ( Conexus::IPv4::Address addr )
    {
      m_remote_address = addr;
      m_remote_address_set = true;
    }

    void Socket::unset_remote_address()
    {
      m_remote_address = Conexus::IPv4::Address();
      m_remote_address_set = false;
    }

    void Socket::set_local_address ( Conexus::IPv4::Address addr )
    {
      m_local_address = addr;
    }

    size_t Socket::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
    {
      PRIntervalTime t;
      PRInt32 bytes;

      if ( ! data || ! data.data() )
        return 0;

      if ( this->is_closed() )
        this->open();

      if ( timeout )
        t = ::PR_MicrosecondsToInterval ( timeout.sec*1000000 + timeout.nsec/1000 );
      else
        t = PR_INTERVAL_NO_TIMEOUT;

      bytes = PR_Send ( m_fd, data, data.size(), 0, t );

      // TODO throw a proper exception
      if ( bytes == -1 ) throw -1;

      return bytes;
    }

    Data Socket::read_data ( size_t size, Timeout timeout ) throw ( read_exception )
    {
      if ( ! this->is_bound() )
        try
        {
          this->change_state ( SOCKET_BOUND );
        }
        catch ( exception::state::failed )
        {
          throw exception::read::not_bound();
        }

      Data d = FileDescriptor::read_data ( size, timeout );

      if ( d.size() == 0 ) throw exception::read::disconnected();

      return d;
    }

  }

}

