/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSNSPRSOCKET_H
#define CONEXUSNSPRSOCKET_H

#include <sys/types.h>

#include <iostream>

#include <utility>

#include <conexus-nspr/filedescriptor.h>
#include <conexus/ipv4_address.h>
#include <conexus/except.h>

namespace Conexus
{

  namespace NSS
  {
    class SSLSocket;
  }
  
  namespace NSPR
  {

    /**
     * This class encapsulates NSPR's Socket API and serves as the base class for all
     * objects performing socket I/O.
     *
     * This class provides the following propertymm properties:
     *  - domain (int)
     *  - type (int)
     *  - protocol (int)
     *
     * @author Rick L Vinyard Jr
     * @ingroup NSPR
     * @ingroup Endpoint
     */
    class Socket: public FileDescriptor
    {
      protected:

        Socket( ) throw ();

      public:
        
        typedef ConexusPointer<Socket> pointer;

        virtual ~Socket() throw ();

        /**
         * Overloads the parent FileDescriptor class to ensure that the final state
         * clears the BOUND, CONNECTED, LISTENING and ACCEPTED flags.
         */
        virtual void close( bool force = false ) throw ( close_exception );

        /**
         * Binding without an address (autobinding) is a child specific action.
         * By default an attempt to bind without providing an address will result in a thrown exception condition.
         * Therefore children should modify this behavior if they wish to provide autobinding.
         */
        virtual void bind() throw ( bind_exception );

        /**
         * Binds the socket to the provided address.
         *
         * The socket should already be in the OPENED state before this call.
         *
         * If the socket is in the CLOSED state, then set_state(OPENED) will be automatically called.
         */
        virtual void bind( const Conexus::Address& a ) throw ( bind_exception );

        /**
         * Connecting without an address (autoconnecting) is a child specific action.
         * By default an attempt to connect without providing an address will result in a thrown exception condition.
         * Therefore children should modify this behavior if they wish to provide autoconnection.
         */
        virtual void connect() throw ( connect_exception );

        /**
         * Connects the socket to the provided address.
         *
         * If the socket is in the CLOSED state, then set_state(OPENED) will be automatically called.
         *
         * If the provided address is a broadcast address, will also set the broadcast socket option.
         */
        virtual void connect( const Address& a ) throw ( connect_exception );

        /**
         * Places the socket in a listening mode; nearly identical to calling listen() on the socket.
         *
         * The socket should already be in the BOUND state before this call.
         *
         * If the socket is in the CLOSED or OPENED states, then set_state(BOUND) will be automatically called.
         */
        virtual void listen( int backlog = 0 );

        virtual ssize_t writeto(Address& a, const Data data) throw (write_exception);

        virtual void set_option( PRSocketOptionData& option );

        virtual void change_state( long states ) throw ( state_exception );

        sigc::signal<void> signal_bound();

        sigc::signal<void> signal_connected();

        sigc::signal<void> signal_listening();

        bool is_bound();

        bool is_connected();

        bool is_listening();

        bool is_accepted();

        bool set_non_blocking(bool set=true);

        bool is_non_blocking();

        virtual Conexus::IPv4::Address& local_address();

        virtual Conexus::IPv4::Address& remote_address();

        virtual void set_remote_address( Conexus::IPv4::Address addr );

        virtual void unset_remote_address();

        virtual void set_local_address( Conexus::IPv4::Address addr );

      protected:
        friend class Conexus::NSS::SSLSocket;
        
        Conexus::IPv4::Address m_local_address;
        Conexus::IPv4::Address m_remote_address;
        bool m_remote_address_set;

        bool m_non_blocking;

        virtual size_t write_data( const Data data, Timeout timeout ) throw ( write_exception );

        /**
         * Reimplements Conexus::Endpoint::read_data() using Conexus::NSPR::FileDescriptor::read_data().
         *
         * Changes from FileDescriptor::read_data()
         *   - Before reading, checks to see if the socket is at least bound, and if not tries to bind the socket.
         *   - If read size is zero, throws Conexus::exception::read::disconnected
         */
        virtual Data read_data( size_t s, Timeout timeout ) throw ( read_exception );

        virtual void set_state_closed();
        virtual void set_state_bound();
        virtual void set_state_connected();
        virtual void set_state_listening();

        sigc::signal<void> m_signal_bound;
        sigc::signal<void> m_signal_connected;
        sigc::signal<void> m_signal_listening;

        virtual void on_local_address_changed( );
        virtual void on_remote_address_changed( );

        void on_local_address_changed_proxy( );
        void on_remote_address_changed_proxy( );

    };

  }
}

#endif
