/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include <conexus-nspr/udp.h>

#include <iostream>

#include <conexus/ipv4_address.h>

namespace Conexus
  {
  namespace NSPR
    {

    UDP::UDP(uint32_t address, uint16_t localport, uint16_t remoteport):
        m_write_without_connect(false),
        m_multicast_auto_join(true),
        m_multicast_joined(false)
    {
      m_local_address.set_address(address);
      m_local_address.set_port(localport);
      m_remote_address.set_port(remoteport);
    }

    UDP::UDP(const std::string& address, uint16_t localport, uint16_t remoteport):
        m_write_without_connect(false),
    m_multicast_auto_join(true),
    m_multicast_joined(false)
    {
      m_local_address.set_address(address);
      m_local_address.set_port(localport);
      m_remote_address.set_port(remoteport);
    }

    UDP::pointer UDP::create(uint16_t localport, uint16_t remoteport) {
      return pointer( new UDP( INADDR_ANY, localport, remoteport ) );
    }

    UDP::pointer UDP::create(uint32_t address, uint16_t localport, uint16_t remoteport) {
      return pointer( new UDP( address, localport, remoteport ) );
    }

    UDP::pointer UDP::create(const std::string& address, uint16_t localport, uint16_t remoteport) {
      return pointer( new UDP( address, localport, remoteport ) );
    }

    UDP::~UDP() throw()
    {
      if (m_multicast_joined)
        leave();
    }

    void UDP::open() throw (open_exception)
    {
      if ( this->is_open() ) this->change_state( ENDPOINT_CLOSED );
      
      m_fd = PR_NewUDPSocket();

      if ( ! m_fd ) throw open_exception( PR_GetError() );

      // Make sure we set the blocking state
      this->set_non_blocking( m_non_blocking );
      
      this->set_state_opened();
    }

    void UDP::bind() throw (bind_exception)
    {
      if ( m_local_address.is_multicast() )
        {
          if (m_multicast_auto_join)
            {
              IPv4::Address addr( INADDR_ANY, m_local_address.port() );
              Socket::bind(addr);
              join();
            }
          else
          {
            IPv4::Address addr( INADDR_ANY, 0 );
            Socket::bind(addr);
          }
        }
      else
        {
          Socket::bind();
        }
    }

    void UDP::bind(Conexus::Address& a) throw (bind_exception)
    {
      Socket::bind(a);
    }

    void UDP::close() throw (close_exception)
    {
      if (m_multicast_joined)
        leave();
      Socket::close();
    }

//     ssize_t UDP::write(const void* data, size_t size) throw (write_exception)
//     {
//       if (m_write_without_connect)
//         {
//           if ( is_closed() )
//             try
//               {
//                 change_state(ENDPOINT_OPENED);
//               }
//             catch (exception::state::failed)
//               {
//                 throw exception::write::not_opened();
//               }
// 
//           return Socket::writeto(m_remote_address, data, size);
//         }
//       else
//         {
//           if ( ! is_connected() )
//             try
//               {
//                 change_state(SOCKET_CONNECTED);
//               }
//             catch (exception::state::failed)
//               {
//                 throw exception::write::not_connected();
//               }
// 
//           return Socket::write(data, size);
//         }
//     }

//     Data::pointer UDP::read(size_t s) throw (read_exception)
//     {
//       if (! is_bound() )
//         try
//           {
//             change_state(SOCKET_BOUND);
//           }
//         catch (exception::state::failed)
//           {
//             throw exception::read::not_bound();
//           }
// 
// //           std::cout << "Bytes available: " << PR_Available( m_fd ) << std::endl;
// 
//       return Socket::read(this->input_available());
// //       return Socket::read(32767);
//     }

    void UDP::connect( ) throw( connect_exception )
    {
      Socket::connect(m_remote_address);
    }

    void UDP::connect( Address& a ) throw ( connect_exception )
    {
      Socket::connect( a );
    }

    bool UDP::multicast_auto_join() const
      {
        return m_multicast_auto_join;
      }

    void UDP::set_multicast_auto_join(bool value)
    {
      m_multicast_auto_join = value;
    }

    bool UDP::write_without_connect() const
      {
        return m_write_without_connect;
      }


    void UDP::set_write_without_connect(bool value)
    {
      m_write_without_connect = value;
    }

    void UDP::join()
    {
/*      if ( m_multicast_request != NULL)
        leave();

      m_multicast_request = (struct ip_mreq*) calloc( 1, sizeof(struct ip_mreq) );
      m_multicast_request->imr_multiaddr.s_addr=m_local_address.sockaddr_in().sin_addr.s_addr;
      m_multicast_request->imr_address.s_addr=htonl(INADDR_ANY);
      set_option(SOL_IP, IP_ADD_MEMBERSHIP, *m_multicast_request);*/
    }

    void UDP::leave()
    {
/*      if ( m_multicast_request == NULL)
        return;

      set_option(SOL_IP, IP_DROP_MEMBERSHIP, *m_multicast_request);

      free(m_multicast_request);
      m_multicast_request = NULL;*/
    }

    int UDP::multicast_hops()
    {
/*      int i;
      option(SOL_IP, IP_MULTICAST_TTL, i);
      return i;*/
    }

    void UDP::set_multicast_hops(int i)
    {
//       set_option(SOL_IP, IP_MULTICAST_TTL, i);
    }

    bool UDP::multicast_loop()
    {
//       int i;
//       option(SOL_IP, IP_MULTICAST_LOOP, i);
//       return i;
    }

    void UDP::set_multicast_loop(bool b)
    {
//       int i = b?1:0;
//       set_option(SOL_IP, IP_MULTICAST_LOOP, i);
    }

    size_t UDP::write_data(const Data data, Timeout timeout) throw (write_exception)
    {
      if (m_write_without_connect)
      {
        if (!m_remote_address_set)
          throw exception::write::no_default_remote_address();

        if ( is_closed() )
          try
        {
          change_state(ENDPOINT_OPENED);
        }
        catch (exception::state::failed)
        {
          throw exception::write::not_opened();
        }

        return Socket::writeto(m_remote_address, data);
      }
      else
      {
        if ( ! is_connected() )
          try
        {
          change_state(SOCKET_CONNECTED);
        }
        catch (exception::state::failed)
        {
          throw exception::write::not_connected();
        }

        return Socket::write_data(data, timeout);
      }
    }

  }
}
