/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSNSPRUDP_H
#define CONEXUSNSPRUDP_H

#include <conexus-nspr/socket.h>

namespace Conexus
{

  namespace NSPR
  {

    /**
     * This class encapsulates a socket used for UDP/IP communications.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup NSPR
     * @ingroup Endpoint
     */
    class UDP : public Socket
    {
      protected:

        UDP( uint32_t address, uint16_t localport, uint16_t remoteport );

        UDP( const std::string& address, uint16_t localport, uint16_t remoteport );

      public:

        typedef ConexusPointer<UDP> pointer;

        static UDP::pointer create( uint16_t localport = 0, uint16_t remoteport = 0 );

        static UDP::pointer create( uint32_t address, uint16_t localport, uint16_t remoteport );

        static UDP::pointer create( const std::string& address, uint16_t localport=0, uint16_t remoteport=0 );

        virtual ~UDP() throw();

        virtual void open() throw ( open_exception );

        virtual void bind() throw ( bind_exception );

        virtual void bind( Conexus::Address& a ) throw ( bind_exception );

        virtual void close() throw ( close_exception );

        /**
         * Overloads the Socket ancestor send method to ensure that the socket is connected
         * before trying to send.
         */
        //         virtual ssize_t write( const void* data, size_t size ) throw ( write_exception );

        /**
         * Receives a single datagram from the UDP socket up to s bytes in size; allocates
         * any memory necessary as a smart pointer contained in the returned Data object.
         *
         * If s < 0 then receives the next pending datagram, however large it may be.
         */
        //         virtual Data::pointer read( size_t s = -1 ) throw ( read_exception );

        virtual void connect() throw ( connect_exception );

        virtual void connect( Address& a ) throw ( connect_exception );

        /**
         * There are some circumstances in which it may be desirable to set a
         * default address and have the send method transparently call the sendto method
         * without connecting rather than connecting and calling the send method.
         *
         * A specific example of this is when the remote address/port may occassionally
         * be unavailable. In this case the destination may generate an ICMP UNREACHABLE
         * message, and in accordance with RFC 1122 the local machine will fail any
         * future sends on a connected port, throwing a write::connection_refused exception.
         * If you want to ensure that the datagram is sent contrary to RFC 1122 you could
         * set the default address, set send without connect and call send without throwing
         * connection exceptions.
         *
         * But, since this is contrary to RFC 1122 behavior you should really think about
         * your design before setting this.
         */

        virtual void set_write_without_connect( bool value = true );


        virtual bool write_without_connect() const;

        virtual bool multicast_auto_join() const;

        virtual void set_multicast_auto_join( bool value = true );

        /**
         * Returns the multicast ttl of the currently associated interface.
         */
        virtual int multicast_hops();

        /**
         * Sets the multicast ttl of the currently associated interface to @em hops.
         */
        virtual void set_multicast_hops( int hops );

        /**
         * If true multicast packets are looped back to local sockets.
         */
        virtual bool multicast_loop();

        /**
         * If parameter @em loop is true then multicast packets are looped back to local sockets.
         */
        virtual void set_multicast_loop( bool loop );

      protected:
        bool m_write_without_connect;
        bool m_multicast_auto_join;
        bool m_multicast_joined;


        /** Joins the multicast group specified by @em addr. */
        virtual void join();

        /** Leaves the multicast group specified by @em addr. */
        virtual void leave();

        virtual size_t write_data(const Data data, Timeout timeout) throw (write_exception);

    };

  }

}

#endif

