/***************************************************************************
 *   Copyright (C) 2007 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "sslsocket.h"

namespace Conexus
{

  namespace NSS
  {

    SSLSocket::SSLSocket ( Conexus::NSPR::Socket::pointer socket, SSLSocket::pointer model )
    {
      // TODO throw a proper exception
      if ( ! socket ) throw -1;

      if ( ! socket->is_open() ) socket->open();

      if ( model )
        m_fd = SSL_ImportFD ( model->fd(), socket->fd() );
      else
        m_fd = SSL_ImportFD ( NULL, socket->fd() );

      if ( m_fd == NULL ) throw -1;

      // Copy the old socket's state before we close the old socket explicitly
      m_state = socket->m_state;

      // The old socket is now invalid so we will close it
      socket->set_close_on_destruct ( false );
      socket->set_state_closed();

      SECStatus result;

      result = SSL_AuthCertificateHook ( m_fd, SSLSocket::SSL_AuthCertificateHook_proxy, this );
      if ( result != SECSuccess ) throw -1; // Conexus::exception::nss::ssl_authcertificatehook();

      result = SSL_HandshakeCallback( m_fd, SSLSocket::SSL_HandshakeCallback_proxy, this );
      if ( result != SECSuccess ) throw -1;

    }

    SSLSocket::pointer SSLSocket::create ( Conexus::NSPR::Socket::pointer socket, SSLSocket::pointer model )
    {
      return pointer ( new SSLSocket ( socket, model ) );
    }

    SSLSocket::~SSLSocket() throw()
    {
    }

    void SSLSocket::open() throw ( open_exception )
    {
      // TODO throw a proper exception
      throw -1;
    }

    size_t SSLSocket::input_available() throw()
    {
      return SSL_DataPending ( m_fd );
    }

    bool SSLSocket::set_option ( int32_t option, bool on )
    {
      // TODO throw a proper exception
      if ( ! m_fd ) throw -1;

      SECStatus result;
      result = SSL_OptionSet ( m_fd, option, on?PR_TRUE:PR_FALSE );
      return result == SECSuccess;
    }

    bool SSLSocket::is_option_set ( int32_t option )
    {
      // TODO throw a proper exception
      if ( ! m_fd ) throw -1;

      PRBool on;
      SECStatus result;
      result = SSL_OptionGet ( m_fd, option, &on );
      // TODO throw a proper exception
      if ( result != SECSuccess ) throw -1;
      return on;
    }

    bool SSLSocket::enable_cipher ( int32_t cipher, bool enable )
    {
      // TODO throw a proper exception
      if ( ! m_fd ) throw -1;

      SECStatus result;
      result = SSL_CipherPrefSet ( m_fd, cipher, enable );
      return result == SECSuccess;
    }

    bool SSLSocket::is_cipher_enabled ( int32_t cipher )
    {
      // TODO throw a proper exception
      if ( ! m_fd ) throw -1;

      SECStatus result;
      PRBool on;
      result = SSL_CipherPrefGet ( m_fd, cipher, &on );
      // TODO throw a proper exception
      if ( result != SECSuccess ) throw -1;
      return on;
    }

    bool SSLSocket::configure_as_server ( Certificate::pointer cert, PrivateKey::pointer key, KEA kea_type )
    {
      // TODO throw a proper exception
      if ( ! m_fd ) throw -1;

      SECStatus result;
      CERTCertificate* cert_cobj = NULL;
      SECKEYPrivateKey* key_cobj = NULL;

      if ( cert ) cert_cobj = cert->cobj();
      if ( key ) key_cobj = key->cobj();

      result = SSL_ConfigSecureServer ( m_fd, cert_cobj, key_cobj, static_cast<SSLKEAType> ( kea_type ) );

      return result == SECSuccess;
    }

    bool SSLSocket::set_url ( const std::string & url )
    {
      // TODO throw a proper exception
      if ( ! m_fd ) throw -1;

      int result;
      result = SSL_SetURL ( m_fd, url.c_str() );
      return result == 0;
    }

    std::string SSLSocket::url()
    {
      char * url = SSL_RevealURL ( m_fd );
      std::string s;
      if ( url )
      {
        s = url;
        PR_Free ( url );
      }
      return s;
    }

    bool SSLSocket::set_pkcs11_pin_arg ( void * arg )
    {
      // TODO throw a proper exception
      if ( ! m_fd ) throw -1;

      int result;
      result = SSL_SetPKCS11PinArg ( m_fd, arg );
      return result == 0;
    }

    void * SSLSocket::pkcs11_pin_arg()
    {
      return SSL_RevealPinArg ( m_fd );
    }

    bool SSLSocket::invalidate_session()
    {
      int result;
      result = SSL_InvalidateSession ( m_fd );
      return result == 0;
    }

    SECURITY_STATUS SSLSocket::security_status()
    {
      int s;
      SECStatus status;
      status = SSL_SecurityStatus ( m_fd, &s, NULL, NULL, NULL, NULL, NULL );
      if ( status != SECSuccess ) throw -1;
      return static_cast<SECURITY_STATUS> ( s );
    }

    std::string SSLSocket::cipher()
    {
      SECStatus status;
      char* cipher;
      std::string cipherstr;
      status = SSL_SecurityStatus ( m_fd, NULL, &cipher, NULL, NULL, NULL, NULL );
      if ( status != SECSuccess ) throw -1;
      cipherstr = cipher;
      PR_Free ( cipher );
      return cipherstr;
    }

    int SSLSocket::session_key_size()
    {
      SECStatus status;
      int session_key_size;
      status = SSL_SecurityStatus ( m_fd, NULL, NULL, &session_key_size, NULL, NULL, NULL );
      if ( status != SECSuccess ) throw -1;
      return session_key_size;
    }

    int SSLSocket::secret_key_size()
    {
      SECStatus status;
      int secret_key_size;
      status = SSL_SecurityStatus ( m_fd, NULL, NULL, NULL, &secret_key_size, NULL, NULL );
      if ( status != SECSuccess ) throw -1;
      return secret_key_size;
    }

    std::string SSLSocket::issuer()
    {
      SECStatus status;
      char* issuer;
      std::string issuerstr;
      status = SSL_SecurityStatus ( m_fd, NULL, NULL, NULL, NULL, &issuer, NULL );
      if ( status != SECSuccess ) throw -1;
      issuerstr = issuer;
      PR_Free ( issuer );
      return issuerstr;
    }

    std::string SSLSocket::subject()
    {
      SECStatus status;
      char* subject;
      std::string subjectstr;
      status = SSL_SecurityStatus ( m_fd, NULL, NULL, NULL, NULL, NULL, &subject );
      if ( status != SECSuccess ) throw -1;
      subjectstr = subject;
      PR_Free ( subject );
      return subjectstr;
    }

    SECItem* SSLSocket::session_id()
    {
      return SSL_GetSessionID ( m_fd );
    }

    bool SSLSocket::set_peer_id ( const std::string & peer_id )
    {
      int result;
      result = SSL_SetSockPeerID ( m_fd, const_cast<char*> ( peer_id.c_str() ) );
      return result == 0;
    }

    Certificate::pointer SSLSocket::peer_certificate()
    {
      CERTCertificate* cert;

      cert = SSL_PeerCertificate ( m_fd );

      if ( cert )
        return Certificate::create ( cert );

      return Certificate::pointer();
    }

    bool SSLSocket::force_handshake()
    {
      SECStatus status;
      status = SSL_ForceHandshake ( m_fd );
      return status == SECSuccess;
    }

    bool SSLSocket::rehandshake ( bool flush_cache )
    {
      SECStatus status;
      status = SSL_ReHandshake ( m_fd, flush_cache );
      return status == SECSuccess;
    }

    bool SSLSocket::reset_handshake ( bool as_server )
    {
      SECStatus status;
      status = SSL_ResetHandshake ( m_fd, as_server );
      return status == SECSuccess;
    }

    void SSLSocket::set_authenticate_certificate_callback(sigc::slot< bool, bool, bool > slot)
    {
      m_authenticate_certificate_slot = slot;
    }

    void SSLSocket::unset_authenticate_certificate_callback()
    {
      m_authenticate_certificate_slot.disconnect();
    }

    void SSLSocket::set_bad_certificate_callback(sigc::slot< bool > slot)
    {
      SECStatus result;
      m_bad_certificate_slot = slot;
      result = SSL_BadCertHook ( m_fd, SSLSocket::SSL_BadCertHook_proxy, this );
      if ( result != SECSuccess ) throw -1; // Conexus::exception::nss::ssl_badcerthook();
    }

    void SSLSocket::unset_bad_certificate_callback()
    {
      SECStatus result;
      result = SSL_BadCertHook ( m_fd, NULL, NULL );
      if ( result != SECSuccess ) throw -1; // Conexus::exception::nss::ssl_badcerthook();
      m_bad_certificate_slot.disconnect();
    }

    void SSLSocket::set_client_authentication_callback(sigc::slot< bool, struct CERTDistNamesStr*, struct CERTCertificateStr**, struct SECKEYPrivateKeyStr** > slot)
    {
      SECStatus result;
      m_client_authentication_slot = slot;
      result = SSL_GetClientAuthDataHook( m_fd, SSLSocket::SSL_GetClientAuthDataHook_proxy, this );
      if ( result != SECSuccess ) throw -1; // Conexus::exception::nss::ssl_getclientauthdatahook();
    }

    void SSLSocket::set_default_client_authentication_callback(char* nickname)
    {
      SECStatus result;
      result = SSL_GetClientAuthDataHook( m_fd, NSS_GetClientAuthData, nickname );
      if ( result != SECSuccess ) throw -1; // Conexus::exception::nss::ssl_getclientauthdatahook();
    }

    void SSLSocket::unset_client_authentication_callback()
    {
      SECStatus result;
      result = SSL_GetClientAuthDataHook( m_fd, NULL, NULL );
      if ( result != SECSuccess ) throw -1;
      m_client_authentication_slot.disconnect();
    }

    sigc::signal< void > SSLSocket::signal_handshake_complete()
    {
      return m_signal_handshake_complete;
    }

    //     char * SSLSocket::on_ssl_pkcs11_password( PK11SlotInfo * info, PRBool retry )
//     {
//       return default_nss_password_function( info, retry, NULL );
//     }

    bool SSLSocket::on_authenticate_certificate ( bool checksig, bool isServer )
    {
      if ( m_authenticate_certificate_slot )
        return m_authenticate_certificate_slot(checksig, isServer);

      return SSL_AuthCertificate ( CERT_GetDefaultCertDB(), m_fd, checksig, isServer ) == SECSuccess;
    }

    bool SSLSocket::on_bad_certificate( )
    {
      if ( m_bad_certificate_slot ) return m_bad_certificate_slot();
      return false;
    }

    bool SSLSocket::on_client_authentication( struct CERTDistNamesStr * caNames, struct CERTCertificateStr ** pRetCert, struct SECKEYPrivateKeyStr ** pRetKey )
    {
      if ( m_client_authentication_slot )
        return m_client_authentication_slot( caNames, pRetCert, pRetKey );
      return false;
    }

//     char * SSLSocket::SSL_SetPKCS11PinArg_proxy( PK11SlotInfo * info, PRBool retry, void * arg )
//     {
//       SSLSocket * ptr = static_cast<SSLSocket*>( arg );
//       return ptr->on_ssl_pkcs11_password( info, retry );
//     }

    SECStatus SSLSocket::SSL_AuthCertificateHook_proxy ( void * arg, PRFileDesc * socket, PRBool checksig, PRBool isServer )
    {
      SSLSocket * ptr = static_cast<SSLSocket*> ( arg );
      if ( ptr->on_authenticate_certificate ( checksig, isServer ) ) return SECSuccess;
      return SECFailure;
    }

    SECStatus SSLSocket::SSL_BadCertHook_proxy ( void * arg, PRFileDesc * socket )
    {
      SSLSocket * ptr = static_cast<SSLSocket*> ( arg );
      if ( ptr->on_bad_certificate() ) return SECSuccess;
      return SECFailure;
    }

    void SSLSocket::SSL_HandshakeCallback_proxy( PRFileDesc * socket, void * arg )
    {
      SSLSocket * ptr = static_cast<SSLSocket*>( arg );
      ptr->m_signal_handshake_complete.emit();
    }

    SECStatus SSLSocket::SSL_GetClientAuthDataHook_proxy( void * arg, PRFileDesc * socket, struct CERTDistNamesStr * caNames, struct CERTCertificateStr ** pRetCert, struct SECKEYPrivateKeyStr ** pRetKey )
    {
      SSLSocket* ptr = static_cast<SSLSocket*>( arg );
      return (ptr->on_client_authentication( caNames, pRetCert, pRetKey )?SECSuccess:SECFailure);
    }

  }

}
