/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "utility.h"
#include <conexus-nspr/utility.h>
#include <conexus/utility.h>

#include <iostream>
#include <termios.h>

#include <sslproto.h>

namespace Conexus
{
  namespace NSS
  {

    bool __initialized = false;

    sigc::signal<std::string,PK11SlotInfo*,PRBool,void*>::accumulated<password_accumulator> signal_pk11_password_request;

    char* pk11_password_callback( PK11SlotInfo* slot, PRBool retry, void* arg )
    {
      char * pw;
      std::string s = signal_pk11_password_request.emit(slot, retry, arg);
      pw = PORT_Strdup( s.c_str() );
      return pw;
    }

    bool init ( bool use_dispatcher_default,
                    const std::string& nss_config_dir,
                    EXPORT_POLICY export_policy,
                    bool read_write )
    {

      SECStatus init_result;

      if ( ! Conexus::NSPR::initialized() )
        Conexus::NSPR::init ( use_dispatcher_default );

      if ( ! Conexus::initialized() )
        Conexus::init ( use_dispatcher_default );

      if ( __initialized )
        return false;

      PK11_SetPasswordFunc ( pk11_password_callback );

      if ( nss_config_dir.size() > 0 && read_write )
        init_result = NSS_InitReadWrite ( nss_config_dir.c_str() );
      else if ( nss_config_dir.size() > 0 )
        init_result = NSS_Init ( nss_config_dir.c_str() );
      else
        init_result = NSS_NoDB_Init ( NULL );

      // Bail at this point if the initialization failed
      if ( init_result != SECSuccess ) return false;

      return set_export_policy( export_policy );
    }

    bool initialized()
    {
      return __initialized;
    }

    bool set_ssl_default_option(uint32_t option, bool on)
    {
      SECStatus result;
      result = SSL_OptionSetDefault( option, on );
      return result == SECSuccess;
    }

    bool is_ssl_default_option_set(uint32_t option)
    {
      PRBool on;
      SECStatus result;
      result = SSL_OptionGetDefault( option, &on );
      // TODO throw a real error code
      if ( result != SECSuccess )
        throw -1;
      return on == PR_TRUE;
    }

    bool enable_default_cipher(uint32_t cipher, bool enabled)
    {
      SECStatus result;
      result = SSL_CipherPrefSetDefault( cipher, enabled );
      return result == SECSuccess;
    }

    bool is_default_cipher_enabled(uint32_t cipher)
    {
      SECStatus result;
      PRBool enabled;
      result = SSL_CipherPrefGetDefault( cipher, &enabled );
      return enabled == PR_TRUE;
    }

    void clear_session_cache( )
    {
      SSL_ClearSessionCache();
    }

    bool configure_server_session_cache ( int max_cache_entries,
                                          uint32_t timeout,
                                          uint32_t ssl3_timeout,
                                          const char* cache_directory,
                                          bool multi_processor
                                        )
    {
      SECStatus result;

      if ( ! multi_processor )
        result = SSL_ConfigServerSessionIDCache ( max_cache_entries, timeout, ssl3_timeout, cache_directory );
      else
        result = SSL_ConfigMPServerSIDCache( max_cache_entries, timeout, ssl3_timeout, cache_directory );
      return result == SECSuccess;
    }

    bool inherit_server_session_cache(const char * env_string)
    {
      SECStatus result;
      result = SSL_InheritMPServerSIDCache( env_string );
      return result == SECSuccess;
    }

    bool set_export_policy(EXPORT_POLICY export_policy)
    {
      SECStatus result = SECSuccess;

      result = SECSuccess;
      
      switch ( export_policy )
      {
        case US_DOMESTIC:
          result = NSS_SetDomesticPolicy();
          break;
        case US_EXPORTABLE:
          result = NSS_SetExportPolicy();
          break;
        case FRANCE:
          result = NSS_SetFrancePolicy();
          break;
        case EXPORT_NONE:
        default:
          break;
      }

      return result == SECSuccess;
    }

    bool set_cipher_policy(uint32_t cipher, POLICY policy)
    {
      SECStatus result;
      result = SSL_CipherPolicySet( cipher, policy );
      return result == SECSuccess;
    }

    POLICY get_cipher_policy(uint32_t cipher)
    {
      PRInt32 policy;
      SECStatus result;
      result = SSL_CipherPolicyGet( cipher, &policy );
      // TODO throw a real exception
      if ( result != SECSuccess )
        throw -1;
      return static_cast<POLICY>( policy );
    }

    long int send_client_hello_cache_hits()
    {
      return SSL_GetStatistics()->sch_sid_cache_hits;
    }

    long int send_client_hello_cache_misses()
    {
      return SSL_GetStatistics()->sch_sid_cache_misses;
    }

    long int send_client_hello_cache_not_ok()
    {
      return SSL_GetStatistics()->sch_sid_cache_not_ok;
    }

    long int handle_server_hello_cache_hits()
    {
      return SSL_GetStatistics()->hsh_sid_cache_hits;
    }

    long int handle_server_hello_cache_misses()
    {
      return SSL_GetStatistics()->hsh_sid_cache_misses;
    }

    long int handle_server_hello_cache_not_ok()
    {
      return SSL_GetStatistics()->hsh_sid_cache_not_ok;
    }

    long int handle_client_hello_cache_hits()
    {
      return SSL_GetStatistics()->hch_sid_cache_hits;
    }

    long int handle_client_hello_cache_misses()
    {
      return SSL_GetStatistics()->hch_sid_cache_misses;
    }

    long int handle_client_hello_cache_not_ok()
    {
      return SSL_GetStatistics()->hch_sid_cache_not_ok;
    }

    void finalize()
    {
      NSS_Shutdown();
    }

    void set_nss_password_function ( PK11PasswordFunc func )
    {
      PK11_SetPasswordFunc ( func );
    }

    char* default_nss_password_function ( PK11SlotInfo* slot, PRBool retry, void* arg )
    {
      struct termios original_settings, new_settings;
      int c;
      std::string s;
      char* password = NULL;

      tcgetattr ( 0, &original_settings ); // store old settings for restore
      new_settings = original_settings;
      new_settings.c_lflag &= ( ~ECHO ); // turn off echo
      new_settings.c_lflag &= ( ~ICANON ); // turn off canonical settings
      new_settings.c_cc[ VTIME ] = 0;
      new_settings.c_cc[ VMIN ] = 1;
      tcsetattr ( 0, TCSANOW, &new_settings );

      if ( retry )
      {
        printf ( "Incorrect password. Please re-enter the password.\n" );
        fflush ( stdout );
      }

      printf ( "Please enter the password for slot \"%s\": ", PK11_GetTokenName ( slot ) );
      fflush ( stdout );

      while ( ( c = getc ( stdin ) ) != '\n' )
      {
        s += c;
        printf ( "*" );
        fflush ( stdout );
      }

      tcsetattr ( 0, TCSANOW, &original_settings ); // restore original settings

      printf ( "\n" );
      fflush ( stdout );

      password = PL_strdup ( s.c_str() );

      return password;
    }

    void disable_all_ssl_ciphers()
    {
      const PRUint16 *cipherSuites = SSL_ImplementedCiphers;
      int             i            = SSL_NumImplementedCiphers;
      SECStatus       rv;

      /* disable all the SSL3 cipher suites */
      while (--i >= 0) {
        PRUint16 suite = cipherSuites[i];
        rv = SSL_CipherPrefSetDefault(suite, PR_FALSE);
        if (rv != SECSuccess)
          std::cerr << "SSL_CipherPrefSetDefault didn't like value 0x" << std::hex << suite << " (i = " << std::dec << i << ")" << std::endl;
      }
    }


  }


}

bool operator <(const SECItem & a, const SECItem & b)
{
  SECComparison result;
  result = SECITEM_CompareItem( &a, &b );
  return result == SECLessThan;
}

bool operator <=(const SECItem & a, const SECItem & b)
{
  SECComparison result;
  result = SECITEM_CompareItem( &a, &b );
  return result != SECGreaterThan;
}

bool operator ==(const SECItem & a, const SECItem & b)
{
  SECComparison result;
  result = SECITEM_CompareItem( &a, &b );
  return result == SECEqual;
}

bool operator !=(const SECItem & a, const SECItem & b)
{
  SECComparison result;
  result = SECITEM_CompareItem( &a, &b );
  return result != SECEqual;
}

bool operator >=(const SECItem & a, const SECItem & b)
{
  SECComparison result;
  result = SECITEM_CompareItem( &a, &b );
  return result != SECLessThan;
}

bool operator >(const SECItem & a, const SECItem & b)
{
  SECComparison result;
  result = SECITEM_CompareItem( &a, &b );
  return result == SECGreaterThan;
}


