/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "context.h"

#include <conexus-ssl/except.h>

namespace Conexus
{

  namespace SSL
  {

    Context::Context ( Method m ) :
        m_method ( m )
    {
      SSL_METHOD* meth;

      switch ( m )
      {
        case SSLV2:
          meth = SSLv2_method();
          break;
        case SSLV2_SERVER:
          meth = SSLv2_server_method();
          break;
        case SSLV2_CLIENT:
          meth = SSLv2_client_method();
          break;
        case SSLV3:
          meth = SSLv3_method();
          break;
        case SSLV3_SERVER:
          meth = SSLv3_server_method();
          break;
        case SSLV3_CLIENT:
          meth = SSLv3_client_method();
          break;
        case TLSV1:
          meth = TLSv1_method();
          break;
        case TLSV1_SERVER:
          meth = TLSv1_server_method();
          break;
        case TLSV1_CLIENT:
          meth = TLSv1_client_method();
          break;
        case SSLV23:
          meth = SSLv23_method();
          break;
        case SSLV23_SERVER:
          meth = SSLv23_server_method();
          break;
        case SSLV23_CLIENT:
          meth = SSLv23_client_method();
          break;
        case DTLSV1:
          meth = DTLSv1_method();
          break;
        case DTLSV1_SERVER:
          meth = DTLSv1_server_method();
          break;
        case DTLSV1_CLIENT:
          meth = DTLSv1_client_method();
          break;
      }

      m_cobj = SSL_CTX_new ( meth );

      if ( m_cobj == NULL ) throw_ssl_exception();
    
    }

    Context::pointer Conexus::SSL::Context::create ( Method m )
    {
      return pointer ( new Context ( m ) );
    }

    Context::~Context()
    {
      SSL_CTX_free ( m_cobj );
    }

    Method Context::method()
    {
      return m_method;
    }

    SSL_CTX * Context::cobj()
    {
      return m_cobj;
    }

    bool Context::use_certificate ( X509 * x )
    {
      return SSL_CTX_use_certificate ( m_cobj, x );
    }

    bool Context::use_certificate_asn1 ( int len, unsigned char * d )
    {
      return SSL_CTX_use_certificate_ASN1 ( m_cobj, len, d );
    }

    bool Context::use_certificate_file ( const std::string & file, FileType type )
    {
      return SSL_CTX_use_certificate_file ( m_cobj, file.c_str(), type );
    }

    bool Context::use_certificate_chain_file ( const std::string & file )
    {
      return SSL_CTX_use_certificate_chain_file ( m_cobj, file.c_str() );
    }

    bool Context::use_private_key ( EVP_PKEY * pkey )
    {
      return SSL_CTX_use_PrivateKey ( m_cobj, pkey );
    }

    bool Context::use_private_key_asn1 ( int pk, unsigned char * d, long len )
    {
      return SSL_CTX_use_PrivateKey_ASN1 ( pk, m_cobj, d, len );
    }

    bool Context::use_private_key_file ( const std::string& file, FileType type )
    {
      return SSL_CTX_use_PrivateKey_file ( m_cobj, file.c_str(), type );
    }

    bool Context::use_rsa_private_key ( RSA * rsa )
    {
      return SSL_CTX_use_RSAPrivateKey ( m_cobj, rsa );
    }

    bool Context::use_rsa_private_key_asn1 ( unsigned char * d, long len )
    {
      return SSL_CTX_use_RSAPrivateKey_ASN1 ( m_cobj, d, len );
    }

    bool Context::use_rsa_private_key_file ( const std::string & file, FileType type )
    {
      return SSL_CTX_use_RSAPrivateKey_file ( m_cobj, file.c_str(), type );
    }

    bool Context::check_private_key() const
    {
      return SSL_CTX_check_private_key ( m_cobj );
    }

    void Context::set_default_password_slot(PasswordSlot slot)
    {
      SSL_CTX_set_default_passwd_cb_userdata( m_cobj, this );

      m_default_password_slot = slot;

      if ( m_default_password_slot )
        SSL_CTX_set_default_passwd_cb( m_cobj, Context::passwd_cb_proxy );
      else
        SSL_CTX_set_default_passwd_cb( m_cobj, NULL );
    }

    bool Context::load_verify_locations(const std::string & ca_file, const std::string & ca_path)
    {
      if ( ca_file.size() == 0 && ca_path.size() == 0 ) return false;
      if ( ca_file.size() == 0 ) return SSL_CTX_load_verify_locations( m_cobj, NULL, ca_path.c_str() );
      if ( ca_path.size() == 0 ) return SSL_CTX_load_verify_locations( m_cobj, ca_file.c_str(), NULL );
      return SSL_CTX_load_verify_locations( m_cobj, ca_file.c_str(), ca_path.c_str() );
    }

    int Context::passwd_cb_proxy(char * buf, int size, int rwflag, void * userdata)
    {
      // if we don't have a Context object we can't get the slot
      if ( userdata == NULL ) return 0;

      // We must assume userdata is a Context object
      Context* context = (Context*)userdata;

      // Do we have a slot?
      if ( ! context->m_default_password_slot ) return 0;

      // We have a slot, let's call it to get the password
      std::string password;
      context->m_default_password_slot( password, (EncDecFlag)rwflag );

      // The provided buffer is too small to hold the password
      if ( (size_t)size < password.size()+1 ) return 0;
      
      strncpy( buf, password.c_str(), password.size()+1 );

      // pem_passwd_cb requires the size fo the buffer to be returned
      return password.size();
    }

    void Context::set_verify_depth(int depth)
    {
      SSL_CTX_set_verify_depth( m_cobj, depth );
    }

    void Context::set_read_ahead(bool set)
    {
      SSL_CTX_set_read_ahead( m_cobj, set );
    }

  }

}

