/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_SSLCONTEXT_H
#define CONEXUS_SSLCONTEXT_H

#include <string>

#include <openssl/ssl.h>

#include <sigc++/sigc++.h>

#include <conexus/pointer.h>
#include <conexus-ssl/enums.h>

namespace Conexus
{

  namespace SSL
  {

    /**
      @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
    */
    class Context
    {
      protected:
        Context ( Method m );

      public:
        typedef ConexusPointer<Context> pointer;

        static pointer create ( Method m=SSLV23 );

        virtual ~Context();

        Method method();

        /**
         * loads the certificate x
         *
         * The rest of the certificates needed to form the complete certificate
         * chain can be specified using add_extra_chain_cert().
         */
        bool use_certificate ( X509 *x );

        /**
         * loads the ASN1 encoded certificate from the memory location d (with
         * length len)
         */
        bool use_certificate_asn1 ( int len, unsigned char *d );

        /**
         * loads the first certificate stored in file
         *
         * The formatting type of the certificate must be specified
         * from the known types PEM or ASN1.
         */
        bool use_certificate_file ( const std::string& file, FileType type );

        /**
         * loads a certificate chain from file
         *
         * The certificates must be in PEM format and must be sorted starting
         * with the subject’s certificate (actual client or server certificate),
         * followed by intermediate CA certificates if applicable, and ending at
         * the highest level (root) CA.
         */
        bool use_certificate_chain_file ( const std::string& file );

        /**
         * adds pkey as private key to this context.
         *
         * If a certificate has already been set and the private does not belong to
         * the certificate an error is returned. To change a certificate, private
         * key pair the new certificate needs to be set with use_certificate()
         * before setting the private key with use_private_key()
         */
        bool use_private_key ( EVP_PKEY *pkey );

        /**
         * adds the private key of type pk stored at memory location d (length len)
         */
        bool use_private_key_asn1 ( int pk, unsigned char *d, long len );

        /**
         * adds the first private key found in file
         *
         * The formatting type of the certificate must be specified from
         * the known types PEM or ASN1.
         */
        bool use_private_key_file ( const std::string& file, FileType type );

        /**
         * adds pkey as private key to this context.
         *
         * If a certificate has already been set and the private does not belong to
         * the certificate an error is returned. To change a certificate, private
         * key pair the new certificate needs to be set with use_certificate()
         * before setting the private key with use_private_key()
         */
        bool use_rsa_private_key ( RSA *rsa );

        /**
         * adds the private key of type pk stored at memory location d (length len)
         */
        bool use_rsa_private_key_asn1 ( unsigned char *d, long len );

        /**
         * adds the first private key found in file
         *
         * The formatting type of the certificate must be specified from
         * the known types PEM or ASN1.
         */
        bool use_rsa_private_key_file ( const std::string& file, FileType type );

        /**
         * checks the consistency of a private key with the corresponding
         * certificate loaded
         *
         * If more than one key/certificate pair (RSA/DSA) is installed, the
         * last item installed will be checked. If e.g. the last item was a
         * RSA certificate or key, the RSA key/certificate pair will be
         * checked.
         */
        bool check_private_key() const;

        /**
         * A callback slot which must be provided by the application, hands
         * back the password to be used during decryption.
         *
         * The slot must fill the std::string& parameter with the password.
         * The EncDecFlag parameter indicates whether the callback is used
         * for encryption (parameter=ENCRYPTION) or decryption (parameter=
         * DECRYPTION).
         */
        typedef sigc::slot<void,std::string&,EncDecFlag> PasswordSlot;

        /**
         * sets the default password callback slot called when loading/storing
         * a PEM certificate with encryption
         */
        void set_default_password_slot ( PasswordSlot slot=PasswordSlot() );

        /** set default locations for trusted CA certificates */
        bool load_verify_locations ( const std::string& ca_file, const std::string& ca_path=std::string() );

        /**
         * sets the maximum depth for the certificate chain verification that
         * shall be allowed
         */
        void set_verify_depth ( int depth );

        /**
         * Sets the read ahead parameter for the context
         * DTLS should have this set
         */
        void set_read_ahead( bool set=true );

        /** Returns the underlying C object */
        SSL_CTX* cobj();

      protected:
        /** The SSL method set in the constructor */
        Method m_method;

        /** The underlying C object */
        SSL_CTX* m_cobj;

        /** The default password callback slot */
        PasswordSlot m_default_password_slot;

        /** A static proxy used to provide a callback point for the C library */
        static int passwd_cb_proxy ( char* buf, int size, int rwflag, void* userdata );

    };

  }

}

#endif
