/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ipv6_tcp.h"

#include <fcntl.h>
#include <conexus-ssl/except.h>

namespace Conexus
{

  namespace SSL
  {

    namespace IPv6
    {

      TCP::TCP ( Context::pointer context, uint32_t address, uint16_t port ) :
          Conexus::IPv6::TCP ( address, port ),
          SSL ( context ),
          m_want_write(false),
          m_want_read(false)
      {
      }

      TCP::TCP ( Context::pointer context, const struct in6_addr& address, uint16_t port ) :
          Conexus::IPv6::TCP ( address, port ),
          SSL ( context ),
          m_want_write(false),
          m_want_read(false)
      {
      }

      TCP::TCP ( Context::pointer context, const std::string& address, uint16_t port ) :
          Conexus::IPv6::TCP ( address, port ),
          SSL ( context ),
          m_want_write(false),
          m_want_read(false)
      {
      }

      TCP::TCP ( Context::pointer context, int accepted_fd, bool b ) :
          Conexus::IPv6::TCP ( accepted_fd, b ), SSL ( context )
      {
        BIO* bio = BIO_new_socket ( m_fd, BIO_NOCLOSE );
        this->set_bio ( bio, bio );
      }

      TCP::pointer TCP::create ( Context::pointer context, uint16_t port )
      {
        return pointer ( new TCP ( context, in6addr_any, port ) );
      }

      TCP::pointer TCP::create ( Context::pointer context, uint32_t address, uint16_t port )
      {
        return pointer ( new TCP ( context, address, port ) );
      }

      TCP::pointer TCP::create ( Context::pointer context, const struct in6_addr& address, uint16_t port )
      {
        return pointer ( new TCP ( context, address, port ) );
      }

      TCP::pointer TCP::create ( Context::pointer context, const std::string& address, uint16_t port )
      {
        return pointer ( new TCP ( context, address, port ) );
      }

      TCP::pointer TCP::create ( Context::pointer context, int accepted_fd, bool b )
      {
        return pointer ( new TCP ( context, accepted_fd, b ) );
      }

      TCP::~TCP() throw()
      {
      }

      void TCP::open() throw ( open_exception )
      {
        Conexus::IPv6::TCP::open();
        BIO* bio = BIO_new_socket ( m_fd, BIO_NOCLOSE );
        this->set_bio ( bio, bio );
      }

      void TCP::connect() throw ( connect_exception )
      {
        Conexus::IPv6::TCP::connect();
        this->ssl_connect();
      }

      void TCP::connect ( const Conexus::IPv6::Address& a ) throw ( connect_exception )
      {
        Conexus::IPv6::TCP::connect ( a );
        this->ssl_connect();
      }

      size_t TCP::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
      {
        int bytes_written;

        // Do we have valid data?
        if ( ! data ) throw_write_exception ( EINVAL );

        if ( ! is_accepted() && ! is_connected() )
        {
          try
          {
            change_state ( SOCKET_CONNECTED );
          }
          catch ( Conexus::exception::state::failed )
          {
            throw Conexus::exception::write::not_connected();
          }
        }

        // We need to do some prep work based on whether the timeout indicates
        // a blocking write, a non-blocking immediate write, or a blocking with timeout
        if ( timeout.is_null() )
        {
          this->set_blocking(true);
        }
        else
        {
          this->set_blocking(false);
        
          fd_set select_fd_set;
          struct timeval tv_timeout = timeout;

          FD_ZERO ( &select_fd_set );
          FD_SET ( m_fd, &select_fd_set );

          // Check to see if the last SSL operation wanted to read something
          // if so, we won't try writing until the socket is readable
          if ( m_want_read )
          {
            switch ( ::select( m_fd+1, &select_fd_set, NULL, NULL, &tv_timeout ) )
            {
              case  0: throw_write_exception(ETIMEDOUT);
              case -1: throw_write_exception(errno);
              default: m_want_read = false;
            }
          }

          // Now, let's check to see if we're writable
          FD_ZERO ( &select_fd_set );
          FD_SET ( m_fd, &select_fd_set );
          switch ( ::select( m_fd+1, NULL, &select_fd_set, NULL, &tv_timeout ) )
          {
            case  0: throw_write_exception(ETIMEDOUT);
            case -1: throw_write_exception(errno);
            default: break;
          }
        }

        // perform the actual write
        bytes_written = SSL_write ( m_cobj, data, data.size() );

        switch ( SSL_get_error ( m_cobj, bytes_written ) )
        {
          case SSL_ERROR_NONE:
            break;
          case SSL_ERROR_ZERO_RETURN:
            throw Conexus::exception::write::disconnected();
          case SSL_ERROR_WANT_READ:
            m_want_read = true;
            throw_write_exception(ETIMEDOUT);
          case SSL_ERROR_WANT_WRITE:
            throw_write_exception(ETIMEDOUT);
          default:
            throw_ssl_exception();
        }

        return bytes_written;
      }

      Data TCP::read_data ( size_t size, Timeout timeout ) throw ( read_exception )
      {
        int bytes_read;
        Data d;

        if ( ! this->is_bound() )
        {
          try
          {
            this->change_state ( SOCKET_BOUND );
          }
          catch ( Conexus::exception::state::failed )
          {
            throw Conexus::exception::read::not_bound();
          }
        }

        // Before trying a read we'll see if there is anything
        // pending in the SSL buffer that can be delivered
        int pending;
        pending = SSL_pending ( m_cobj );

        if ( pending > 0 )
        {
          d.resize ( pending );
          bytes_read = SSL_read ( m_cobj, d, pending );
          switch ( SSL_get_error ( m_cobj, bytes_read ) )
          {
            case SSL_ERROR_NONE:
              if ( m_timestamp ) d.set_current_time();
              if ( d.size() != ( unsigned ) bytes_read ) d.resize ( bytes_read );
              return d;
            case SSL_ERROR_ZERO_RETURN:
              throw Conexus::exception::read::disconnected();
            case SSL_ERROR_WANT_READ:
              throw_read_exception(ETIMEDOUT);
            case SSL_ERROR_WANT_WRITE:
              m_want_write = true;
              throw_read_exception(ETIMEDOUT);
            default:
              throw_ssl_exception();
          }
        }

        // Since there is nothing pending in the SSL buffer,
        // we need to perform a read from the underlying socket

        // We need to do some prep work based on whether the timeout indicates
        // a blocking or non-blocking read
        if ( timeout.is_null() )
        {
          this->set_blocking(true);
        }
        else
        {
          this->set_blocking(false);

          fd_set select_fd_set;
          struct timeval tv_timeout = timeout;

          FD_ZERO ( &select_fd_set );
          FD_SET ( m_fd, &select_fd_set );

          // Check to see if the last SSL operation wanted to write something
          // if so, we won't try reading until the socket is writable
          if ( m_want_write )
          {
            switch ( ::select( m_fd+1, NULL, &select_fd_set, NULL, &tv_timeout ) )
            {
              case  0: throw_read_exception(ETIMEDOUT);
              case -1: throw_read_exception(errno);
              default: m_want_write = false;
            }
          }

          // Now, let's check to see if we're readable
          FD_ZERO ( &select_fd_set );
          FD_SET ( m_fd, &select_fd_set );
          switch ( ::select( m_fd+1, &select_fd_set, NULL, NULL, &tv_timeout ) )
          {
            case  0: throw_read_exception(ETIMEDOUT);
            case -1: throw_read_exception(errno);
            default: break;
          }
        }
        
        // We need to try and find out what the read size will be

        // If read size is zero then we will need to find out how
        // many bytes are available in the SSL record by peeking
        // at one byte to advance the SSL state
        if ( size == 0 )
        {
          int peek_result;
          uint8_t peek_byte;
          peek_result = SSL_peek ( m_cobj, &peek_byte, 1 );
          switch ( SSL_get_error ( m_cobj, peek_result ) )
          {
            case SSL_ERROR_NONE:
              size = SSL_pending ( m_cobj );
              break;
            case SSL_ERROR_ZERO_RETURN:
              throw Conexus::exception::read::disconnected();
            case SSL_ERROR_WANT_READ:
              throw_read_exception(ETIMEDOUT);
            case SSL_ERROR_WANT_WRITE:
              m_want_write = true;
              throw_read_exception(ETIMEDOUT);
            default:
              throw_ssl_exception();
          }
        }

        // if size is still zero then we will use the default read size
        if ( size == 0 ) size = this->fallback_read_size();

        // resize the data object to hold the entire read buffer
        d.resize ( size );

        // perform the actual read
        bytes_read = SSL_read ( m_cobj, d, size );

        switch ( SSL_get_error ( m_cobj, bytes_read ) )
        {
          case SSL_ERROR_NONE:
            break;
          case SSL_ERROR_ZERO_RETURN:
            throw Conexus::exception::read::disconnected();
          case SSL_ERROR_WANT_READ:
            throw Conexus::exception::read::timeout();
          case SSL_ERROR_WANT_WRITE:
            m_want_write = true;
            throw Conexus::exception::read::timeout();
          default:
            throw_ssl_exception();
        }

        if ( m_timestamp ) d.set_current_time();
        if ( d.size() != ( unsigned ) bytes_read ) d.resize ( bytes_read );
        return d;
      }

    }

  }

}

