/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ssl.h"

#include <iostream>
#include <conexus-ssl/except.h>

namespace Conexus
{

  namespace SSL
  {

    SSL::SSL ( Context::pointer context ) :
        m_cobj ( NULL ),
        m_context ( context )
    {
      if ( m_context ) m_cobj = SSL_new ( m_context->cobj() );
    }

    SSL::pointer SSL::create ( Context::pointer context )
    {
      return pointer ( new SSL ( context ) );
    }

    SSL::~SSL()
    {
      if ( m_cobj ) SSL_free ( m_cobj );
    }

    ssize_t SSL::ssl_write ( const Data data )
    {
      return this->ssl_write( data, data.size() );
    }

    ssize_t SSL::ssl_write ( const char* data, size_t datalen )
    {
      int result;

      if ( datalen == 0 ) return 0;
      result = SSL_write ( m_cobj, data, datalen );
      if ( result <= 0 ) throw_ssl_exception();
      return result;
    }

    Data SSL::ssl_read ( size_t s )
    {
      int bytes = 0;
      int error;
      Data d;

      // if read size is zero then we will read all available
      if ( s == 0 ) s = SSL_pending(m_cobj);

      // if read size is still zero then return no data
      if ( s == 0 ) return d;

      // resize the data object to hold the entire read buffer
      d.resize ( s );

      // perform the actual read
      bytes = SSL_read( m_cobj, d, s );

      error = SSL_get_error(m_cobj, bytes);

      switch(error)
      {
        case SSL_ERROR_NONE:
          if ( d.size() != ( unsigned ) bytes ) d.resize ( bytes );
          return d;
        case SSL_ERROR_ZERO_RETURN:
          throw Conexus::exception::read::disconnected();
        default:
          throw_ssl_exception();
      }

    }

    size_t SSL::ssl_input_available()
    {
      return SSL_pending(m_cobj);
    }
    
    void SSL::ssl_connect()
    {
      int result;
      result = SSL_connect(m_cobj);
      if ( result <= 0 ) throw_ssl_exception();
    }

    void SSL::ssl_accept()
    {
      int result;
      result = SSL_accept(m_cobj);
      if ( result <= 0 ) throw_ssl_exception();
    }

    void SSL::set_client_mode()
    {
      SSL_set_connect_state(m_cobj);
    }

    void SSL::set_server_mode()
    {
      SSL_set_accept_state(m_cobj);
    }

    Context::pointer SSL::context()
    {
      return m_context;
    }

    void SSL::set_context ( Context::pointer context )
    {
      if ( m_cobj ) SSL_free ( m_cobj );
      m_cobj = NULL;

      m_context = context;
      if ( m_context ) m_cobj = SSL_new ( m_context->cobj() );
    }

    bool SSL::clear()
    {
      if ( m_cobj == NULL ) return false;
      return SSL_clear ( m_cobj );
    }

    bool SSL::ssl_shutdown ( bool bidirectional )
    {
      int result, result2;

      if ( m_cobj == NULL ) return false;

      result = SSL_shutdown ( m_cobj );

      switch ( result )
      {
        case 1:
          return true;
        case 0:
          if ( ! bidirectional ) return true;
          result2 = SSL_shutdown ( m_cobj );
          if ( result2 == 1 ) return true;
        case -1:
          break;
      }

      throw_ssl_exception();
    }

    bool SSL::use_certificate ( X509 * x )
    {
      return SSL_use_certificate ( m_cobj, x );
    }

    bool SSL::use_certificate_asn1 ( unsigned char * d, int len )
    {
      return SSL_use_certificate_ASN1 ( m_cobj, d, len );
    }

    bool SSL::use_certificate_file ( const std::string & file, FileType type )
    {
      return SSL_use_certificate_file ( m_cobj, file.c_str(), type );
    }

    bool SSL::use_private_key ( EVP_PKEY * pkey )
    {
      return SSL_use_PrivateKey ( m_cobj, pkey );
    }

    bool SSL::use_private_key_asn1 ( int pk, unsigned char * d, long len )
    {
      return SSL_use_PrivateKey_ASN1 ( pk, m_cobj, d, len );
    }

    bool SSL::use_private_key_file ( const std::string & file, FileType type )
    {
      return SSL_use_PrivateKey_file ( m_cobj, file.c_str(), type );
    }

    bool SSL::use_rsa_private_key ( RSA * rsa )
    {
      return SSL_use_RSAPrivateKey ( m_cobj, rsa );
    }

    bool SSL::use_rsa_private_key_asn1 ( unsigned char * d, long len )
    {
      return SSL_use_RSAPrivateKey_ASN1 ( m_cobj, d, len );
    }

    bool SSL::use_rsa_private_key_file ( const std::string & file, FileType type )
    {
      return SSL_use_RSAPrivateKey_file ( m_cobj, file.c_str(), type );
    }

    bool SSL::check_private_key() const
    {
      return SSL_check_private_key ( m_cobj );
    }

    void SSL::set_verify_depth ( int depth )
    {
      SSL_set_verify_depth ( m_cobj, depth );
    }

    long SSL::verify_result()
    {
      if ( m_cobj == NULL ) return X509_V_OK;
      return SSL_get_verify_result ( m_cobj );
    }

    X509 * SSL::peer_certificate()
    {
      if ( m_cobj == NULL ) return NULL;
      return SSL_get_peer_certificate ( m_cobj );
    }

    bool SSL::set_bio ( BIO * read_bio, BIO * write_bio )
    {
      if ( m_cobj == NULL ) return false;
      SSL_set_bio ( m_cobj, read_bio, write_bio );
      return true;
    }

    BIO * SSL::read_bio()
    {
      if ( m_cobj == NULL ) return NULL;
      return SSL_get_rbio ( m_cobj );
    }

    BIO * SSL::write_bio()
    {
      if ( m_cobj == NULL ) return NULL;
      return SSL_get_wbio ( m_cobj );
    }

    void SSL::set_read_ahead(bool set)
    {
      SSL_set_read_ahead(m_cobj, set);
    }

    ::SSL* SSL::cobj()
    {
      return m_cobj;
    }


  }

}
