/***************************************************************************
 *   Copyright (C) 2008 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_SSLSSL_H
#define CONEXUS_SSLSSL_H

#include <string>

#include <openssl/ssl.h>

#include <conexus/pointer.h>
#include <conexus/data.h>
#include <conexus/except.h>
#include <conexus-ssl/enums.h>
#include <conexus-ssl/context.h>

namespace Conexus
{

  namespace SSL
  {

    /**
      @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
    */
    class SSL
    {
      protected:
        SSL(Context::pointer context);

      public:
        typedef ConexusPointer<SSL> pointer;

        static pointer create(Context::pointer context = Context::pointer());

        ~SSL();

        ssize_t ssl_write( const Data data );

        ssize_t ssl_write( const char* data, size_t datalen );

        Data ssl_read( size_t s = 0 );

        size_t ssl_input_available();

        void ssl_connect();

        void ssl_accept();

        /**
         * Sets the SSL engine into a client(connect) mode when performing
         * new handshakes.
         *
         * If ssl_connect() is not used in a client application, this
         * method must be explicitly called before handshaking occurs.
         */
        void set_client_mode();

        /**
         * Sets the SSL engine into a server(accept) mode when performing
         * new handshakes.
         *
         * If ssl_accept() is not used in a server application, this
         * method must be explicitly called before handshaking occurs.
         */
        void set_server_mode();

        /** returns the SSL context associated with this SSL object */
        Context::pointer context();

        /** sets the SSL context */
        void set_context( Context::pointer context );

        /**
         * reset SSL object to allow another connection
         * @return true if successful, false otherwise
         */
        bool clear();

        /**
         * shuts down an active TLS/SSL connection
         *
         * It sends the "close notify" shutdown alert to the peer.
         * 
         * @param bidirectional if true, calls the underlying SSL_shutdown()
         * twice to shut down the bidirectional shutdown
         */
        bool ssl_shutdown( bool bidirectional=false );

        
        /**
         * loads the certificate x
         *
         * The rest of the certificates needed to form the complete certificate
         * chain can be specified using add_extra_chain_cert().
         */
        bool use_certificate ( X509 *x );

        /**
         * loads the ASN1 encoded certificate from the memory location d (with
         * length len)
         */
        bool use_certificate_asn1 ( unsigned char *d, int len );

        /**
         * loads the first certificate stored in file
         *
         * The formatting type of the certificate must be specified
         * from the known types PEM or ASN1.
         */
        bool use_certificate_file ( const std::string& file, FileType type );

        /**
         * adds pkey as private key to this context.
         *
         * If a certificate has already been set and the private does not belong to
         * the certificate an error is returned. To change a certificate, private
         * key pair the new certificate needs to be set with use_certificate()
         * before setting the private key with use_private_key()
         */
        bool use_private_key ( EVP_PKEY *pkey );

        /**
         * adds the private key of type pk stored at memory location d (length len)
         */
        bool use_private_key_asn1 ( int pk, unsigned char *d, long len );

        /**
          * adds the first private key found in file
          *
          * The formatting type of the certificate must be specified from
          * the known types PEM or ASN1.
         */
        bool use_private_key_file ( const std::string& file, FileType type );

        /**
         * adds pkey as private key to this context.
         *
         * If a certificate has already been set and the private does not belong to
         * the certificate an error is returned. To change a certificate, private
         * key pair the new certificate needs to be set with use_certificate()
         * before setting the private key with use_private_key()
         */
        bool use_rsa_private_key ( RSA *rsa );

        /**
         * adds the private key of type pk stored at memory location d (length len)
         */
        bool use_rsa_private_key_asn1 ( unsigned char *d, long len );

        /**
         * adds the first private key found in file
         *
         * The formatting type of the certificate must be specified from
         * the known types PEM or ASN1.
         */
        bool use_rsa_private_key_file ( const std::string& file, FileType type );

        /**
         * checks the consistency of a private key with the corresponding
         * certificate loaded
         *
         * If more than one key/certificate pair (RSA/DSA) is installed, the
         * last item installed will be checked. If e.g. the last item was a
         * RSA certificate or key, the RSA key/certificate pair will be
         * checked.
         *
         * If no key/certificate was explicitly added for this ssl, the last
         * item added into ctx will be checked.
         */
        bool check_private_key() const;

        /**
         * sets the maximum depth for the certificate chain verification that
         * shall be allowed
         */
        void set_verify_depth ( int depth );

        /**
         * get result of peer certificate verification
         *
         * returns X509_V_OK if the underlying C object does not yet
         * exist, to follow the same logic with the case where no
         * peer certificate was presented... namely that no error
         * has occurred.
         */
        long verify_result();

        /** returns the X509 certificate of the peer */
        X509* peer_certificate();

        /** connect the SSL object with a BIO */
        bool set_bio( BIO* read_bio=NULL, BIO* write_bio=NULL );

        /** sets the read ahead parameter of the SSL object */
        void set_read_ahead( bool set=true );

        /** returns a pointer to the read channel bio */
        BIO* read_bio();

        /** returns a pointer to the write channel bio */
        BIO* write_bio();
        
        ::SSL* cobj();

      protected:
        ::SSL* m_cobj;
        Context::pointer m_context;

    };

  }

}

#endif
