/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSDATA_H
#define CONEXUSDATA_H

#include <cstdlib>
#include <stdexcept>
#include <string>

#include <sigc++/sigc++.h>

#include <conexus/pointer.h>

#include <glibmm/timeval.h>
#include <glibmm/thread.h>

namespace Conexus
{

  typedef enum DataMode {
    COPY,
    MANAGED,
    UNMANAGED,
  } DataMode;

  /**
   * This structure represents a data chunk.
   *
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   */
  class Data
  {
    public:

      /** Constructor initializes buffer to a specific size */
      Data( size_t s = 0, unsigned priority = 0 ) throw (std::bad_alloc);

      /** Constructor initializes buffer with a copy of specific data */
      Data( const void* d, size_t s, unsigned priority = 0, DataMode mode=COPY ) throw (std::bad_alloc);

      /** Copy constructor */
      Data( const Data& other );

      /** Destructor */
      ~Data();

      /** Pointer to the actual data */
      uint8_t* data();

      /** Constant pointer to the actual data */
      const uint8_t* data() const;

      /**
       * Sets the buffer to specific data.
       * newdata + newsize must not be within the current storage region if a copy is performed
       * @return false if a realloc is required and the realloc fails
       */
      bool set_data( const void* newdata, size_t newsize, DataMode mode=COPY ) throw (std::bad_alloc);

      /** Current buffer size */
      size_t size() const;

      /**
       * Resize buffer to size s
       * @return true on success, false on failure
       *
       *  - If s < current size, then buffer is shrunk
       *  - If s > current size, then buffer is expanded and zero-filled
       *  - If s = current size, no changes occur
       *  - If currently set to an external buffer, this method will fail
       */
      bool resize( size_t s ) throw (std::bad_alloc);

      /**
       * Performs a deep copy of this data object.
       * This method allocates a new memory chunk, copies the existing data into that data chunk,
       * and returns an object containing a smart pointer to the underlying data.
       */
      Data clone() const;

      operator bool();

      operator bool() const;

      /** Cast to a 1-octet pointer */
      operator uint8_t*();

      /** Cast to a 1-octet const pointer */
      operator const uint8_t*() const;

      operator char*();

      operator const char*() const;

      operator void*();

      operator const void*() const;

      uint8_t& operator[](const ssize_t offset);

      const uint8_t& operator[](const ssize_t offset) const;

      /** Returns a hex string representation of the data */
      std::string hex_string( std::string separator = std::string() ) const;

      /** Releases any allocated memory and resets size to 0 */
      void clear();

      /** Returns the timeval associated with this data item */
      const Glib::TimeVal& time() const;

      /** Sets the timeval associated with this data item */
      void set_time( const Glib::TimeVal& );

      /** Sets the timeval associated with this data item to the current time */
      void set_current_time();

      /** Get the priority, where 0 is the lowest priority */
      unsigned priority() const;

      /** Set the priority, where 0 is the lowest priority */
      void set_priority( unsigned p );

      bool operator<( const Data& other ) const;
      bool operator<=( const Data& other ) const;
      bool operator==( const Data& other ) const;
      bool operator!=( const Data& other ) const;
      bool operator>=( const Data& other ) const;
      bool operator>( const Data& other ) const;

      /**
       * Similar to memcmp.
       * Compares the memory areas of two data items. If the two data
       * items are of different sizes (s1 and s2) only the first n
       * bytes will be compared, where n is the lesser of s1 and s2.
       *
       * Unlike memcmp, 0 will be returned only if the memory areas
       * are identical and s1 is equal to s2.
       *
       * If memcmp would return 0 for the first n bytes and s1 != s2,
       * the an integer less than 0 is returned if s1 \< s2 and an
       * integer greater than 0 is returned if s1 \> s2.
       */
      int compare( const Data& other ) const;

      /**
       * Appends a copy of the other data object to this data object
       * @return true if successful, false otherwise
       */
      bool append( const Data& other );

      Data operator+( const Data& other ) const;

      Data& operator+=( const Data& other );

    protected:

      class Storage {
        public:

          Storage(): data(NULL), size(0), manage_data(false), priority(0) { }

          ~Storage() {
            if ( data && manage_data ) {
                ::free( data );
                data = NULL;
            }
          }

          typedef ConexusPointer<Storage> pointer;

          /** Pointer to data buffer */
          uint8_t* data;

          /** Size of the data in bytes. */
          size_t size;

          /** True if the data is external and cannot be resized */
          bool manage_data;

          /** Priority, where 0 is the lowest priority */
          unsigned int priority;

          /** An associated time value */
          Glib::TimeVal time;

      };

      Storage::pointer m_storage;

  };


}

#endif
