/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "endpoint.h"
#include "utility.h"

namespace Conexus
{

  Endpoint::Endpoint ( bool close_on_destruct ) :
      m_read_thread ( NULL ),
      m_read_delivery_thread ( NULL ),
      m_read_terminate ( false ),
      m_read_terminate_immediate ( false ),
      m_default_read_timeout ( 0, 500000000 ),
      m_fallback_read_size ( 0xFFFF ),
      m_default_write_timeout ( 0, 0 ),
      m_timestamp ( true ),
      m_dispatcher ( NULL ),
      m_use_dispatcher ( false ),
      m_close_on_destruct ( close_on_destruct ),
      m_state ( ENDPOINT_CLOSED ),
      m_readable ( true ),
      m_writable ( true )
  {
  }

  Endpoint::~Endpoint()
  {
    if ( this->is_running() )
      this->stop();
  }

  bool Endpoint::is_running()
  {
    return m_read_thread;
  }

  void Endpoint::start( )
  {
    this->start ( Conexus::use_dispatcher_default() );
  }

  void Endpoint::start ( bool use_dispatcher )
  {
    if ( m_read_thread )
      return ;

    if ( this->is_closed() )
      this->open();

    m_use_dispatcher = use_dispatcher;

    m_read_terminate = false;

    if ( m_use_dispatcher )
    {
      m_dispatcher = new Glib::Dispatcher();
      m_dispatcher->connect ( sigc::mem_fun ( *this, &Endpoint::emit_received_data ) );
    }
    else
    {
      m_read_delivery_thread = Glib::Thread::create ( sigc::mem_fun ( *this, &Endpoint::read_delivery_thread_main ), true );
    }

    m_read_thread = Glib::Thread::create ( sigc::mem_fun ( *this, &Endpoint::read_thread_main_proxy ), true );

    m_signal_start_stop.emit ( START );
  }

  void Endpoint::stop( )
  {
    if ( ! m_read_thread )
      return ;

    // Set the variable to terminate the read
    m_read_terminate = true;

    // Wait for the read thread to finish
    m_read_thread->join();
    m_read_thread = NULL;

    if ( m_dispatcher )
    {
      delete m_dispatcher;
      m_dispatcher = NULL;
    }
    else
    {
      // Signal the read delivery thread to unlock
      // so that it can check m_read_terminate and end
      m_read_delivery_conditional.signal();

      // And wait for the read delivery thread to finish
      m_read_delivery_thread->join();
      m_read_delivery_thread = NULL;
    }

    m_read_terminate = false;

    m_signal_start_stop.emit ( STOP );
  }

  void Endpoint::read_thread_main_proxy()
  {
    this->read_thread_main();
  }

  void Endpoint::read_thread_main()
  {
    Data d;
    bool disconnected = false;
    bool eof = false;

    while ( ! m_read_terminate )
    {
      try
      {
        d = this->read_data ( 0, m_default_read_timeout );
        this->queue_received_data ( d );
      }
      catch ( exception::read::retry )
      {
        continue;
      }
      catch ( exception::read::disconnected )
      {
        disconnected = true;
        break;
      }
      catch ( exception::read::eof )
      {
        eof = true;
        break;
      }
    }

    // if we broke out due to a read disconnect or EOF we need to do
    // an orderly shutdown which means making sure that all reads
    // are delivered and the delivery thread is shutdown and joined
    if ( ! m_read_terminate )
    {
      m_read_terminate = true;
      m_read_delivery_conditional.signal();
      if ( m_read_delivery_thread )
        m_read_delivery_thread->join();
      m_read_terminate = false;
      m_read_thread = NULL;
      m_read_delivery_thread = NULL;
    }

    // Only after all reads have been delivered will we emit the disconnected or EOF signals
    if ( disconnected ) m_signal_disconnected.emit();
    if ( eof ) m_signal_eof.emit();
  }

  void Endpoint::read_delivery_thread_main()
  {
    Data current;

    // Grab the read queue lock
    m_read_queue_lock.lock();

    // Repeat main loop until terminate is set
    while ( ! m_read_terminate )
    {

      // loop in the conditional until there is something to deliver or terminate is set
      while ( !m_read_terminate && m_read_queue.empty() )
        m_read_delivery_conditional.wait ( m_read_queue_lock );

      // if we have a terminate and are set for immediate terminates then break out of the loop
      if ( m_read_terminate && m_read_terminate_immediate )
      {
        m_read_queue_lock.unlock();
        break;
      }

      // Otherwise it's time to deliver the data
      while ( ! m_read_queue.empty() )
      {
        current = m_read_queue.front();
        m_read_queue.pop();
        m_read_queue_lock.unlock();
        m_signal_data.emit ( current );
        m_signal_data_received_size.emit ( current.size() );
        m_read_queue_lock.lock();
      }
    }

    m_read_queue_lock.unlock();
  }

  sigc::signal<void, const Data> Endpoint::signal_data()
  {
    return m_signal_data;
  }

  void Endpoint::queue_received_data ( const Data d )
  {
    m_read_queue_lock.lock();
    m_read_queue.push ( d );
    if ( m_use_dispatcher )
    {
      m_read_queue_lock.unlock();
      m_dispatcher->emit();
    }
    else
    {
      m_read_delivery_conditional.signal();
      m_read_queue_lock.unlock();
    }
  }

  void Endpoint::emit_received_data( )
  {
    Data current;
    m_read_queue_lock.lock();
    while ( ! m_read_queue.empty() )
    {
      current = m_read_queue.front();
      m_read_queue.pop();
      m_read_queue_lock.unlock();
      m_signal_data.emit ( current );
      m_signal_data_received_size.emit ( current.size() );
      m_read_queue_lock.lock();
    }
    m_read_queue_lock.unlock();
  }

  sigc::signal< void, size_t > Endpoint::signal_data_received_size( )
  {
    return m_signal_data_received_size;
  }

  sigc::signal< void, size_t > Endpoint::signal_data_transmitted_size( )
  {
    return m_signal_data_transmitted_size;
  }

  const Timeout& Endpoint::default_read_timeout() const
  {
    return m_default_read_timeout;
  }

  void Endpoint::set_default_read_timeout ( Timeout t )
  {
    if ( t.is_null() ) t = Timeout ( 0,0 );
    m_default_read_timeout = t;
  }

  size_t Endpoint::fallback_read_size()
  {
    return m_fallback_read_size;
  }

  void Endpoint::set_fallback_read_size ( size_t s )
  {
    m_fallback_read_size = s;
  }

  const Timeout& Endpoint::default_write_timeout() const
  {
    return m_default_write_timeout;
  }

  void Endpoint::set_default_write_timeout ( Timeout t )
  {
    if ( t.is_null() ) t = Timeout ( 0,0 );
    m_default_write_timeout = t;
  }

  bool Endpoint::read_terminate_immediate()
  {
    return m_read_terminate_immediate;
  }

  void Endpoint::set_read_terminate_immediate ( bool i )
  {
    m_read_terminate_immediate = i;
  }

  bool Endpoint::is_timestamping_received_data()
  {
    return m_timestamp;
  }

  void Endpoint::set_timestamp_received_data ( bool t )
  {
    m_timestamp = t;
  }

  ssize_t Endpoint::write ( const void* data, size_t size ) throw ( write_exception )
  {
    try
    {
      return this->write_data( Data(data, size, UNMANAGED ), m_default_write_timeout );
    }
    catch ( exception::write::disconnected& e )
    {
      m_signal_disconnected.emit();
      throw e;
    }
  }

  ssize_t Endpoint::write ( const void* data, size_t size, Timeout timeout ) throw ( write_exception )
  {
    try
    {
      return this->write_data( Data(data, size, UNMANAGED), timeout );
    }
    catch ( exception::write::disconnected& e )
    {
      m_signal_disconnected.emit();
      throw e;
    }
  }

  ssize_t Endpoint::write ( const Data data ) throw ( write_exception )
  {
    return this->write_data( data, m_default_write_timeout );
  }

  ssize_t Endpoint::write ( const Data data, Timeout to ) throw ( write_exception )
  {
    return this->write_data( data, to );
  }

  ssize_t Endpoint::write ( Endpoint::pointer source, size_t block_size, bool keep_open )
  {
    Conexus::Data data;
    ssize_t written = 0;

    if ( ! source ) return 0;

    try
    {
      do
      {
        data = source->read ( block_size );
        if ( data.size() > 0 )
        {
          written += this->write ( data );
        }
      }
      while ( data.size() > 0 );
    }
    catch ( read_exception& e )
    {
    }

    if ( ! keep_open )
      source->close();

    return written;
  }

  Data Endpoint::read(size_t size) throw( read_exception )
  {
    return this->read( size, m_default_read_timeout );
  }

  Data Endpoint::read(size_t size, Timeout timeout) throw( read_exception )
  {
    try
    {
      Data d = this->read_data ( size, timeout );

      if ( m_read_thread ) this->queue_received_data ( d );

      return d;
    }
    catch ( exception::read::disconnected& e )
    {
      m_signal_disconnected.emit();
      throw e;
    }
    catch ( exception::read::eof& e )
    {
      m_signal_eof.emit();
      throw e;
    }
  }

  Data Endpoint::read(Timeout timeout) throw( read_exception )
  {
    return this->read( 0, timeout );
  }

  size_t Endpoint::input_available() throw ()
  {
    return -1;
  }

  void Endpoint::change_state ( long states ) throw ( state_exception )
  {
    if ( states == ENDPOINT_UNCHANGED )
      return ;

    // check for inconsistent states
    if ( ( states & ENDPOINT_CLOSED ) && ( states & ENDPOINT_OPENED ) )
      throw exception::state::inconsistent();

    // perform the proper actions
    if ( is_open() && states & ENDPOINT_CLOSED )
    {
      this->close();
      if ( this->is_open() )
        throw exception::state::failed();
    }
    else if ( this->is_closed() && states & ENDPOINT_OPENED )
    {
      this->open();
      if ( this->is_closed() )
        throw exception::state::failed();
    }
  }

  void Endpoint::set_state_opened()
  {
    if ( is_open() )
      return ;
    m_state &= ~ENDPOINT_CLOSED;
    m_state |= ENDPOINT_OPENED;
    m_signal_opened.emit();
  }

  void Endpoint::set_state_closed()
  {
    if ( is_closed() )
      return ;
    m_state &= ~ENDPOINT_OPENED;
    m_state |= ENDPOINT_CLOSED;
    m_signal_closed.emit();
  }

  bool Endpoint::close_on_destruct() const
  {
    return m_close_on_destruct;
  }

  void Endpoint::set_close_on_destruct ( bool value )
  {
    m_close_on_destruct = value;
  }

  void Endpoint::close_and_reopen ( long new_state )
  {
    long prev_state = state();
    change_state ( ENDPOINT_CLOSED );
    if ( new_state == ENDPOINT_UNCHANGED )
      change_state ( prev_state );
    else
      change_state ( new_state );
  }

  sigc::signal<void> Endpoint::signal_opened()
  {
    return m_signal_opened;
  }

  sigc::signal<void> Endpoint::signal_closed()
  {
    return m_signal_closed;
  }

  sigc::signal< void, StartStop > Endpoint::signal_start_stop()
  {
    return m_signal_start_stop;
  }

  sigc::signal< void > Endpoint::signal_eof()
  {
    return m_signal_eof;
  }

  sigc::signal< void > Endpoint::signal_disconnected()
  {
    return m_signal_disconnected;
  }

  bool Endpoint::is_open()
  {
    return m_state & ENDPOINT_OPENED;
  }

  bool Endpoint::is_closed()
  {
    return m_state & ENDPOINT_CLOSED;
  }

  long Endpoint::state()
  {
    return m_state;
  }


} // namespace Conexus

Conexus::Endpoint& operator<< ( Conexus::Endpoint & io, const Conexus::Data d )
{
  io.write ( d );
  return io;
}

Conexus::Endpoint::pointer operator<< ( Conexus::Endpoint::pointer io, const Conexus::Data d )
{
  if ( ! io )
    throw std::invalid_argument ( "Conexus: << operator with pointer argument, but pointer is NULL." );
  io->write ( d );
  return io;
}

Conexus::Endpoint& operator>> ( Conexus::Endpoint & io, Conexus::Data d )
{
  d = io.read();
  return io;
}

Conexus::Endpoint::pointer operator>> ( Conexus::Endpoint::pointer io, Conexus::Data d )
{
  if ( ! io )
    throw std::invalid_argument ( "Conexus: >> operator with pointer argument, but pointer is NULL." );
  d = io->read();
  return io;
}



