/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSIO_H
#define CONEXUSIO_H

#include <set>
#include <queue>

#include <sys/select.h>

#include <glibmm/dispatcher.h>
#include <glibmm/thread.h>

#include <conexus/pointer.h>
#include <conexus/enums.h>
#include <conexus/except.h>
#include <conexus/data.h>
#include <conexus/object.h>

/**
 * \mainpage Conexus Library Reference
 *
 * \section main_toc Table of Contents
 *
 *  - Articles
 *    - \ref quick_start
 * 
 *  - Important Reference Documentation Sections
 *    - <a href="hierarchy.html">Class Hierarchy</a>
 *    - <a href="annotated.html">Class List</a>
 *    - <a href="group__Endpoint.html">I/O Endpoints</a>
 *      - <a href="group__IPv4.html">IPv4 Endpoints</a>
 *      - <a href="group__IPv6.html">IPv6 Endpoints</a>
 *    - <a href="group__Exceptions.html">Exceptions</a>
 */

/**
 * \page quick_start Quick Start Guide to Conexus
 *
 * This is intended to be a very succinct introduction to conexus, and is \b not
 * intended as a general tutorial. This quick start guide will use IPv6/UDP sockets
 * to demonstrate basic conexus concepts.
 *
 * \section quick_start_sections Sections:
 *
 *  -# \ref quick_start_concepts
 *  -# \ref quick_start_examples
 *     -# \ref quick_start_client
 *        -# \ref quick_start_client_code
 *        -# \ref quick_start_client_discussion
 *     -# \ref quick_start_server
 *        -# \ref quick_start_server_code
 *        -# \ref quick_start_server_discussion
 *
 * \section quick_start_concepts Concepts
 * 
 * \subsection quick_start_concept_1 Concept #1: Endpoints
 *
 * \par
 * Endpoints are the points where the \e I and \e O in \e I/O (Input/Output) occur. If you
 * perform a write operation on an endpoint then data will be sent via the endpoint's
 * communication mechanism (which is dependent upon the specific endpoint type). If you
 * perform a read operation then data will be received via the endpoint's communication mechanism
 * (again, dependent upon the specific endpoint type, but only if data is available).
 *
 * \par
 * Endpoints derived from Conexus::Endpoint include endpoints that employ serial communications,
 * IPv4/TCP, IPv4/UDP, IPv6/TCP, IPv6/UDP, kernel message queues, pipes, et. al.
 *
 * \subsection quick_start_concept_2 Concept #2: Data
 *
 * \par
 * The Conexus::Data class is essentially an encapsulation of a data buffer ( a pointer to a
 * block of bytes) and the buffer size (in bytes). This class is analogous to the \c buf and
 * \c count parameters in the POSIX \c write() function:
 * \code ssize_t write(int fd, const void *buf, size_t count);\endcode
 *
 * \par
 * In addition to storing the buffer and buffer size Conexus::Data provides a mechanism for
 * storing a priority value (useful when creating priority queues) and a single time value
 * as well as several other utility methods.
 *
 * \subsection quick_start_concept_3 Concept #3: Starting and stopping endpoints (or moving to an event driven model)
 * 
 * \par
 * Data can be read from a Conexus::Endpoint with the \c read() method which is similar in concept to the
 * POSIX \c read() operation. However, a Conexus::Endpoint also provides \c start() and
 * \c stop() methods. When a Conexus::Endpoint is started it will create a separate service
 * thread. This thread will watch for input data and emit \c signal_data when data is received
 * by the endpoint.
 *
 * \par
 * To receive the data as it is received by an endpoint you need to connect \c signal_data to
 * the function or class method you want to receive the data. Since data is delivered via a
 * sigc++ signal you can connect multiple functions and/or class methods and all connected
 * callbacks will receive the endpoint's data.
 *
 * \subsection quick_start_concept_4 Concept #4: sigc++ signal and slot library
 * 
 * \par
 * As mentioned in <i>Concept #3: Starting and stopping endpoints</i> a running endpoint delivers
 * received data to one or more of your functions or classes via the \c signal_data signal.
 * 
 * \par
 * To receive data from an endpoint you need to:
 *  -# Create the function or class method that will receive the data from the endpoint.
 *    - Your function or method must have the following signature:
 *      -# \c void \e function_or_method_name \c (const \c Conexus::Data)
 *      - This means that the return type of the function or method must be \c void and the first parameter must be
 *        \c const \c Conexus::Data
 *      - If you want other parameters you must use \c sigc::bind , which is discussed in the Gtkmm tutorial and the sigc++ documentation.
 *  -# Create the sigc++ slot that will receive the data from the endpoint.
 *    - Generally, \c sigc::mem_fun and \c sigc::ptr_fun will suffice
 *    - The required signature of the sigc::slot is:
 *      -# \c sigc::slot<void, \c Conexus::Data>
 *  -# Connect the slot to the endpoint using the \c signal_data() method to access the data signal. You will use the signal's \c connect() method to accomplish this.
 *
 * \subsection quick_start_concept_567 Concepts #5, #6 and #7: Smart Pointers, pointer and create()
 * 
 * \par
 * An early design decision was made to focus on the use of smart pointers as defined in the ANSI C++
 * committee's \c tr1 (Technical Report 1), which as of this writing has been accepted by the committee
 * and recommended to ISO for incorporation as the first revision to the C++ standard. A reference
 * implementation can be found in the boost library or in the \c std::tr1 namespace of gcc 4.0 or
 * greater.
 *
 * \par
 * Two key concepts are the \c pointer typedef and the \c create method defined in each conexus class.
 *
 * \par
 * The \c pointer typedef is defined \e within \e each \e class and is typedefed to a smart pointer
 * to that \e specific class. Thus, \c Conexus::IPv6::UDP::pointer is a smart pointer to an IPv6/UDP
 * endpoint, and \c Conexus::TTY::pointer is a smart pointer to a TTY endpoint.
 *
 * \par
 * Each class also has static \c create() methods with parameters that are identical to
 * the class' constructors. The \c create() method dynamically creates an object and returns a smart
 * pointer to that object (the typedefed \c pointer ), similar to the way the \c new operator dynamically
 * creates an object and returns a raw pointer to the newly created object. Since the \c create() method is
 * static, you do not need an instance of the class to call it; \c Classname::create() is sufficient.
 *
 * \par
 * Naturally, the smart pointer returned from \c create() can be assigned to the class' \c pointer
 * type, which allows you to use a syntax such as:
 * \code Classname::pointer my_variable = Classname::create(); \endcode
 *
 * \section quick_start_examples Quick start example IPv6/UDP client and server
 * 
 * \par
 * This example will demonstrate the creation of an IPv6 UDP client and server pair.
 * The client will send a simple data pattern to the server on port 1500.
 *
 * \subsection quick_start_client An IPv6/UDP client
 *
 * \subsubsection quick_start_client_code IPv6/UDP client code
 * 
 * \par
 * \include quickstart_client.cpp
 *
 * \subsubsection quick_start_client_discussion IPv6/UDP client code discussion
 * 
 * \par
 * First, we need to include the main conexus library header:
 * \code
 * #include <conexus.h>
 * 
 * int main(int argc, char* argv[]) {
 * \endcode
 *
 * \par
 * Next, we will declare the variable containing the data we will send. For this we will send a simple C string:
 * \code const char* data = "0123456789"; \endcode
 *
 * \par
 * Before we do anything in conexus we must call the library's \c init() method
 * which will take care of various initializations including the threading system.
 * \code Conexus::init(); \endcode
 *
 * \par
 * With the conexus library initialized we will create the IPv6/UDP endpoint that
 * will transmit our character string.
 * \code Conexus::IPv6::UDP::pointer udp = Conexus::IPv6::UDP::create(); \endcode
 * \note The statement above declares a smart pointer to the IPv6/UDP endpoint
 * through the \c pointer typedef in the \c Conexus::IPv6::UDP class, and also
 * calls upon the static method \c create() in the same class to actually
 * create the endpoint. We use the \c pointer typedef since the return value
 * from the \c create() method is always a smart pointer to \e that \e class \e type.
 *
 * \par
 * There are several ways a UDP socket can write to an address. The Conexus::Socket
 * class (derived from Conexus::Endpoint and ancestor of Conexus::IPv6::UDP)
 * introduces the \c writeto() method. However, for this example we will use the
 * basic \c write() method present in Conexus::Endpoint (ancestor of our
 * udp class).
 *
 * \par
 * While the \c writeto() method accepts the address as a parameter, the \c write()
 * method requires the remote address to already be set, so we will do that now.
 * \code
 *   udp->set_remote_address("::1", 1500);
 * \endcode
 *
 * \par
 * Now that we have created the IPv6/UDP endpoint and set the remote address we can
 * simply write the data to the endpoint to send it via IPv6/UDP to the remote address.
 * \code
 *   udp->write( data, 11 );
 * \endcode
 *
 * \par
 * That's it. All that is left of our example program is to provide a return value
 * from \c main() and the closing brace.
 * \code
 *   return 0;
 * }
 * \endcode
 * 
 * \subsection quick_start_server An IPv6/UDP server
 *
 * \par
 * This section provides an example of an IPv6/UDP server. The server will listen for
 * all traffic on port 1500 and print the data to \c stdout as a C string. It will run
 * for 20 seconds then automatically shut-down.
 * 
 * \subsubsection quick_start_server_code IPv6/UDP server code
 * 
 * \par
 * \include quickstart_server.cpp
 *
 * \subsubsection quick_start_server_discussion IPv6/UDP server code discussion
 * 
 * \par
 * Like the client example, we will start by including the main conexus header.
 * Since we will also print the received data to \c cout we will need to include
 * \c iostream . Finally, since we call the \c sleep() function we will also
 * need to include \c unistd.h .
 * \code
 * #include <conexus.h>
 * #include <iostream>
 * #include <unistd.h>
 * \endcode
 *
 * \par
 * We will declare the prototype of the function that will receive all input data. As mentioned
 * above in the \ref quick_start_concepts section our function must have a signature with a \c void
 * return type and the first parameter must be a \c const \c Conexus::Data type.
 * \code
 * void print_data(const Conexus::Data d);
 * \endcode
 *
 * \par
 * As in the client example, before calling any other functions in the conexus library we must ensure that the
 * conexus library is initialized by calling the \c init() method.
 * \code
 * int main(int argc, char* argv[]) {
 *   Conexus::init();
 * \endcode
 *
 * \par
 * With the library initialized we can create an IPv6/UDP endpoint. This will
 * dynamically create an IPv6/UDP endpoint with the port set to 1500.
 * \code
 *   Conexus::IPv6::UDP::pointer udp = Conexus::IPv6::UDP::create(1500);
 * \endcode
 *
 * \par
 * With the endpoint created we can now connect our callback function \c print_data() to
 * the endpoint's data signal. We use the \c signal_data() accessor to get the signal,
 * and then use the signal's \c connect() method. The \c connect() method in the sigc++
 * library takes a sigc++ slot as its parameter. We turn the \c print_data() function
 * into a slot by using the \c sigc::ptr_fun() class which creates a slot from a
 * pointer to a function (hence \c ptr_fun ). Thus, our function name is preceeded by
 * the address operator \c & .
 * \code
 *   udp->signal_data().connect( sigc::ptr_fun(&print_data) );
 * \endcode
 *
 * \par
 * Now our endpoint is created and connected to our \c print_data()
 * function. We will start the endpoint using the \c start() method. This will create
 * a service thread that will call our \c print_data() function as our endpoint receives
 * data.
 * \code
 *   udp->start();
 * \endcode
 *
 * \note The callback will occur <i><b>in the service thread</b></i>, and not in the
 * main thread.
 * 
 * \par
 * Since all data will be delivered to \c print_data() in the service thread we will
 * have this main thread sleep for 20 seconds. When you run the example server and client
 * you should start the server, then run the client several times. You will see that the
 * service thread is responding to your client while the main thread is sleeping at this
 * point.
 * \code
 *   sleep(20);
 * \endcode
 *
 * \par
 * Now that we have run for 20 seconds we will shut down the endpoint and allow the endpoint
 * to perform any necessary cleanup, where the term "necessary cleanup" varies by endpoint
 * type.
 * \code
 *   udp->stop();
 * \endcode
 *
 * \par
 * With the endpoint stopped, all that remains is to return a proper value from \c main and
 * to include the closing brace.
 * \code
 *   return 0;
 * }
 * \endcode
 *
 * \par
 * Now we need to provide the body of our \c print_data() function. This is a one line function
 * that adds a little helper text and prints out the number of bytes received and the actual
 * data as a C style string.
 * \code
 * void print_data(const Conexus::Data d) {
 *   std::cout << "Received " << d.size() << " bytes of data: " << (const char*)(d.data()) << "\n";
 * }
 * \endcode
 *
 * \note The bytes received are available through the Conexus::Data
 * class' \c size() member and the data is available through the \c data() member.
 *
 * \note The typecast is necessary since the native data type of the Conexus::Data class
 * is an unsigned 8-bit integer (\c uint8_t ) which is probably (but not necessarily) a
 * C \c unsigned \c char type.
 *
 * \section quick_start_end Conclusion and more information
 * 
 * \par
 * That concludes this quick-start tutorial. Additional information may be found in the conexus library
 * documentation as well as the sigc++ library documentation. An additional source of information on
 * how to use the sigc++ classes can be found in the Gtkmm tutorial.
 * 
 */


/**
 * A convenience macro to create a static generic_create() method that
 * returns an Endpoint::pointer rather than the specific type smart pointer.
 *
 * To use this, just put:
 *   CONEXUS_ENDPOINT_GENERIC_CREATE();
 * somewhere in the public section of your class derived from Endpoint.
 */
#define CONEXUS_ENDPOINT_GENERIC_CREATE() static Endpoint::pointer generic_create() { return create(); }

namespace Conexus
  {

  /**
   * @defgroup Endpoint Conexus I/O Endpoints
   */

    struct Timeout
    {
      Timeout( long int us=0 ): sec(0), nsec(us*1000) { }

      Timeout( long int s, long int ns ): sec(s), nsec(ns) { }

      operator struct timeval()
      {
        struct timeval tv;
        tv.tv_sec = sec;
        tv.tv_usec = nsec / 1000 + ((nsec%1000>=500)?1:0);
        return tv;
      }

      operator struct timespec()
      {
        struct timespec ts;
        ts.tv_sec = sec;
        ts.tv_nsec = nsec;
        return ts;
      }

      operator bool() { return sec > 0 && nsec >= 0; }

      operator double() { return sec + nsec/1000000000.0; }

      Timeout& operator=(const struct timeval& tv) { sec = tv.tv_sec; nsec = tv.tv_usec * 1000; return *this; }

      Timeout& operator=(const struct timespec& tv) { sec = tv.tv_sec; nsec = tv.tv_nsec; return *this; }

      bool operator==( const Timeout& other ) const { return sec == other.sec && nsec == other.nsec; }

      bool is_zero() const { return sec == 0 && nsec == 0; }

      bool is_null() const { return sec < 0 || nsec < 0; }

      long int sec;
      long int nsec;
    };

  /**
   * This class is the top level object for all subclasses performing I/O (Input/Output)
   * operations.
   *
   * read/write methods are pure virtual methods in this base class and it is the
   * responsibility of all children to provide proper implementations that may go beyond
   * the traditional read(2) and write(2) functions standard in Linux/Unix.
   *
   * Children are also responsible for maintenance of the file descriptor member @em m_fd
   * and the integer state value @em m_state.
   *
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   * @ingroup Endpoint
   */
  class Endpoint: public Object
    {
      protected:
        
      /**
       * Default constructor setting state to 0 and file descriptor to -1 (which should be
       * an invalid fd on just about every POSIX system.
       *
       * This constructor simply sets the values according to the parameters and performs no
       * real actions with the file descriptor.
       */
        Endpoint(bool close_on_destruct=true);

      public:

      /** Class scope smart pointer typedef */
      typedef ConexusPointer<Endpoint> pointer;

      /**
       * Destructor does nothing. It will not call @em close() upon the file descriptor or perform
       * any other cleanup; these are the responsibility of the children.
       */
      virtual ~Endpoint();

      /**
       * Start the endpoint. If the endpoint is configured for multi-threaded
       * operation this will create and start the read thread.
       *
       * Whether a dispatcher is used or not is dependent upon the global setting
       * which defaults to false, but may be set with Conexus::init() or
       * Conexus::set_use_dispatcher_default().
       *
       * If you want to handle it manually, endpoints are capable of reading/writing
       * without ever calling start/stop.
       */
      virtual void start( );

      /**
       * Starts the endpoint, and forces it to use a specific dispatcher value possibly
       * overriding the global default.
       */
      virtual void start( bool use_dispatcher );

      /**
       * Stops any running threads. Stopping is not immediate. If a read operation is
       * being performed the current read will complete before the thread is stopped.
       */
      virtual void stop( );

      /** True if the endpoint is running with threaded reads */
      virtual bool is_running();

      /** True if received data is timestamped */
      virtual bool is_timestamping_received_data();

      /** Set to true to timestamp received data */
      virtual void set_timestamp_received_data(bool t=true);

      /**
       * Signal emitted when data is read. Connect to this signal if you want to receive all
       * data read by the endpoint. The data object received is unique for each callback
       * point, allowing a callback to modify the data set without effect subsequent
       * callbacks.
       */
      sigc::signal<void, const Data> signal_data();

      /**
       * Signal emitted when data is received notifying the callback of the number of bytes
       * read.
       * 
       * Connect to this signal if you need to know when data is received, and how much,
       * but don't want the data contents.
       */
      sigc::signal<void, size_t> signal_data_received_size();

      /**
       * Signal emitted when data is transmitted, notifying the callback of the number of
       * bytes transmitted.
       * 
       * Connect to this signal if you need to know when data is transmitted, and how much,
       * but don't want the transmitted data contents.
       */
      sigc::signal<void, size_t> signal_data_transmitted_size();

      /** Signal emitted when this endpoint is opened */
      sigc::signal<void> signal_opened();

      /** Signal emitted when this endpoint is closed */
      sigc::signal<void> signal_closed();

      /**
       * Signal emitted when this endpoint is started
       *
       * The first signal parameter indicates the state
       */
      sigc::signal<void,StartStop> signal_start_stop();

      /**
       * Signal emitted when an EOF (End Of File) condition is observed
       */
      sigc::signal<void> signal_eof();

      /**
       * Signal emitted when a disconnect condition is observed
       */
      sigc::signal<void> signal_disconnected();

      /** The default timeout used when read(size_t) is called */
      const Timeout& default_read_timeout() const;

      /**
       * Set the default read timeout
       * A null timeout causes a blocking read to occur, while a non-null timeout causes
       * a non-blocking read to occur with the timeout parameter as specified
       */
      void set_default_read_timeout( Timeout t );

      /**
       * When a read operation occurs, the read_data() method relies upon the
       * endpoint's input_available() method. When this method returns 0, the
       * default read_size is used for the read operation.
       *
       * Default setting is 65535 (2^16 - 1 or 0xFFFF)
       */
      size_t fallback_read_size();

      /** Sets the default read size used when input_available() returns 0 */
      void set_fallback_read_size(size_t s);

      /** The default timeout used when a write(uint8_t*,size_t) or write(Data) is called */
      const Timeout& default_write_timeout() const;

      /**
       * Set the default write timeout
       * A null timeout causes a blocking write to occur, while a non-null timeout causes
       * a non-blocking write to occur with the timeout parameter as specified
       */
      void set_default_write_timeout( Timeout t );

      bool read_terminate_immediate();

      void set_read_terminate_immediate( bool i=true );

      /**
       * Will attempt to open the object without any additional information.
       */
      virtual void open() throw (open_exception) = 0;

      /**
       */
      virtual void close(bool force=false) throw (close_exception) = 0;

      /**
       * Write data to the endpoint using the default write timeout parameter
       *
       * If the default write timeout parameter is null (sec<0 or nsec<0) then
       * a blocking write will occur. Otherwise a non-blocking write will occur
       * using the default write timeout value.
       *
       * @return The number of bytes actually written
       * @param data Pointer to the raw data block to be written.
       * @param size Size in bytes of the raw data block to write.
       */
      ssize_t write(const void* data, size_t size) throw (write_exception);

      /**
       * Write data to the endpoint using the specified write timeout
       * 
       * @return The number of bytes actually written or accepted by the write queue
       * @param data Pointer to the raw data block to be written.
       * @param size Size in bytes of the raw data block to write.
       * @param timeout If timeout is null a blocking write will occur. Otherwise a non-blocking write will occur using the timeout parameter.
       */
      ssize_t write(const void* data, size_t size, Timeout timeout ) throw (write_exception);

      /**
       * Reads source until a zero-length read, timeout or EOF occurs and writes into this endpoint.
       * Inspired by the NSPR PR_TransmitFile() function
       */
      ssize_t write(Endpoint::pointer source, size_t block_size=65535, bool keep_open=true);
      
      /**
       * Write a Conexus::Data object to the endpoint using the default write timeout parameter
       *
       * If the default write timeout parameter is null (sec<0 or nsec<0) then
       * a blocking write will occur. Otherwise a non-blocking write will occur
       * using the default write timeout value.
       *
       * @return The number of bytes actually written
       * @param data The Conexus::Data object to be written
       */
      ssize_t write(const Data data) throw (write_exception);

      /**
       * Write a Conexus::Data object to the endpoint using the specified write timeout
       *
       * @return The number of bytes actually written or accepted by the write queue
       * @param data The Conexus::Data object to be written
       * @param timeout If timeout is null a blocking write will occur. Otherwise a non-blocking write will occur using the timeout parameter.
       */
      ssize_t write(const Data data, Timeout timeout ) throw (write_exception);

      /**
       * Read into a dynamically created buffer
       * 
       * Read up to size bytes from the endpoint or all available data if s=0 using the default read timeout parameter
       * @return the data read from the endpoint
       * 
       * @note The returned data object stores the data as a smart pointer. Therefore, if you want to keep the
       * data from being automatically deallocated, keep the returned data object (or a
       * copy of the returned data object) around as long as you need it. Call by value
       * is sufficient to keep the object around, but extracting the pointer to the actual
       * data via the dereference operator or through the smart pointer's get() method is
       * not sufficient to keep the data block from being deallocated (again, keeping a
       * copy around is easier).
       */
      Data read( size_t size=0 ) throw (read_exception);

      /**
       * Read into a dynamically created buffer
       *
       * Read up to size bytes from the endpoint or all available data if s=0 using the provided read timeout parameter
       * @return the data read from the endpoint
       *
       * @note The returned data object stores the data as a smart pointer. Therefore, if you want to keep the
       * data from being automatically deallocated, keep the returned data object (or a
       * copy of the returned data object) around as long as you need it. Call by value
       * is sufficient to keep the object around, but extracting the pointer to the actual
       * data via the dereference operator or through the smart pointer's get() method is
       * not sufficient to keep the data block from being deallocated (again, keeping a
       * copy around is easier).
       */
      Data read( size_t size, Timeout timeout ) throw (read_exception);

      /**
       * Read into a dynamically created buffer
       *
       * Read all available data from the endpoint using the provided read timeout parameter
       * @return the data read from the endpoint
       *
       * @note The returned data object stores the data as a smart pointer. Therefore, if you want to keep the
       * data from being automatically deallocated, keep the returned data object (or a
       * copy of the returned data object) around as long as you need it. Call by value
       * is sufficient to keep the object around, but extracting the pointer to the actual
       * data via the dereference operator or through the smart pointer's get() method is
       * not sufficient to keep the data block from being deallocated (again, keeping a
       * copy around is easier).
       */
      Data read( Timeout timeout ) throw (read_exception);

      /**
       * Returns the number of bytes available for immediate reading, 0
       * if no bytes are available, or -1 if the method is invalid for a derived class.
       */
      virtual size_t input_available() throw ();

      /**
       * It is intended that children will provide their own implementation of
       * this method to take into account the actions necessary for any additional
       * states they may introduce.
       *
       * This method as implemented recognizes the OPEN, CLOSE and UNCHANGED states
       * and calls the virtual open, close (or no call) as is appropriate.
       *
       * This method does not actually modify the m_state member. It is the
       * responsibility of the called virtual methods open() and close() to
       * actually perform the modification.
       *
       * @return true if all requested state changes occurred, false otherwise
       */
      virtual void change_state(long new_state) throw (state_exception);

      /**
       * Return this endpoint's current numeric state.
       * The meaning of this value is specific to the endpoint.
       */
      long state();

      /** If set, calls this endpoint's virtual close() method on destruction */
      void set_close_on_destruct(bool value);

      /** True if this endpoint's virtual close() method will be called on destruction */
      bool close_on_destruct() const;

      /**
       * Close the I/O point and reopen to a new (or same state).
       *
       * If state is UNCHANGED will reopen to the previous state.
       *
       * close and reopen is accomplished by calling change_state(CLOSED) and then
       * change_state(new_state), not by calling close directly.
       */
      void close_and_reopen(long state = ENDPOINT_UNCHANGED);

      /** True if this endpoint is in an open state */
      bool is_open();

      /** True if this endpoint is in a closed state */
      bool is_closed();

    protected:
      bool m_use_blocking_reads;
      bool m_use_blocking_writes;
      
      Glib::Thread *m_read_thread;
      Glib::Thread *m_read_delivery_thread;
      bool m_read_terminate;
      bool m_read_terminate_immediate;

      Timeout m_default_read_timeout;
      size_t m_fallback_read_size;
      Timeout m_default_write_timeout;
      bool m_timestamp;

      Glib::Mutex m_read_queue_lock;

      Glib::Cond m_read_delivery_conditional;

      Glib::Dispatcher* m_dispatcher;
      bool m_use_dispatcher;

      sigc::signal<void, const Data> m_signal_data;
      sigc::signal<void, size_t> m_signal_data_received_size;
      sigc::signal<void, size_t> m_signal_data_transmitted_size;
      std::queue<Data> m_read_queue;

      virtual void read_thread_main();
      void read_thread_main_proxy();

      virtual void read_delivery_thread_main();

      void queue_received_data( const Data d );

      void emit_received_data( );

      
      bool m_close_on_destruct;
      long m_state;
      bool m_readable;
      bool m_writable;

      virtual void set_state_opened();
      virtual void set_state_closed();

      sigc::signal<void> m_signal_opened;
      sigc::signal<void> m_signal_closed;
      sigc::signal<void,StartStop> m_signal_start_stop;
      sigc::signal<void> m_signal_eof;
      sigc::signal<void> m_signal_disconnected;

      virtual size_t write_data(const Data data, Timeout timeout) throw (write_exception) = 0;

      /**
       * This is the primary mechanism of reading from an endpoint and must be
       * reimplemented by children.
       *
       * @return The data object read. If no data was available for reading an
       * instance may still be returned, but it will have a zero data size.
       *
       * @param timeout If > 0, an upper bound on the amount of time (in microseconds)
       * elapsed before read_data() returns.
       * If = 0, then the read will be attempted immediately. If < 0, read will block
       * without timeout.
       *
       * @param s The upper bound on the number of bytes to read. If s = 0, all available
       * data will be read.
       *
       * @throw Conexus::exception::read::timeout Thrown when a read timeout occurs
       *
       * @throw Conexus::exception::read::eof Thrown when a read encounters an EOF (End Of
       * File) condition
       *
       * @throw Conexus::exception::read::disconnect Thrown when a read encounters a
       * disconnect condition
       *
       * @throw Conexus::read_exception Thrown (or a descendant is thrown) when a read
       * encounters an exceptional condition
       */
      virtual Data read_data( size_t size, Timeout timeout) throw (read_exception) = 0;
};

}

Conexus::Endpoint& operator<<(Conexus::Endpoint& io, const Conexus::Data d);
Conexus::Endpoint::pointer operator<<(Conexus::Endpoint::pointer io, const Conexus::Data d);
Conexus::Endpoint& operator>>(Conexus::Endpoint& io, Conexus::Data d);
Conexus::Endpoint::pointer operator>>(Conexus::Endpoint::pointer io, Conexus::Data d);

#endif
