/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSENUMS_H
#define CONEXUSENUMS_H

#include <netinet/ip.h>
#include <termios.h>

namespace Conexus {

  typedef enum ReadWrite {
    READ_WRITE_NONE=0,
    READ=1,
    WRITE=2,
    READ_WRITE=3,
    READ_WRITE_UNCHANGED=4,
    READ_WRITE_LAST_STATE=READ_WRITE_UNCHANGED
  } ReadWrite;

  /**
   * @relates Conexus::Endpoint
   * These enumerations are used in the socket class methods, and use is also encouraged in children.
   */
  typedef enum EndpointState {
    ENDPOINT_NOSTATE   = 0,
    ENDPOINT_CLOSED    = 1<<0,    /**< The Endpoint object is in a closed state */
    ENDPOINT_OPENED    = 1<<1,   /**< The Endpoint object is in an opened state */
    ENDPOINT_LAST_STATE  = ENDPOINT_OPENED,
    ENDPOINT_UNCHANGED = ~0, /**< setting this mode will keep the mode unchanged */
  } EndpointState;

  /**
   * @relates Conexus::Endpoint
   * Enumerates start/stop states
   *
   * STOP was selected as 0, so that a variable of this enum can boolean test
   * START as true and STOP as false.
   */
  typedef enum StartStop {
    STOP=0, 
    START=1,
  } StartStop;
  
  /**
   * @relates Conexus::Socket
   * These enumerations are used in the socket class methods, and use is also encouraged in children.
   */
  typedef enum SocketState {
    SOCKET_BOUND     = ENDPOINT_LAST_STATE << 1,  /**< The socket is bound to an interface */
    SOCKET_CONNECTED = ENDPOINT_LAST_STATE << 2,  /**< The socket is connected (TCP) */
    SOCKET_LISTENING = ENDPOINT_LAST_STATE << 3,  /**< The socket is listening for connections (TCP) */
    SOCKET_ACCEPTED  = ENDPOINT_LAST_STATE << 4,  /**< The socket is an accepted connection (TCP) */
    SOCKET_LAST_STATE  = SOCKET_ACCEPTED,
  } SocketState;

  /**
   * @relates Conexus::Pipe
   */
  typedef enum PipeMode {
    PIPE_READ=ENDPOINT_LAST_STATE<<1,           /**< opened for reading */
    PIPE_WRITE=ENDPOINT_LAST_STATE<<2,            /**< opened for writing */
  } PipeMode;

  /**
   * @relates Conexus::File
   */
  typedef enum FileMode {
    FILE_READ=1<<0,
    FILE_WRITE=1<<1,
    FILE_APPEND=1<<2,
    FILE_CREATE=1<<3,
    FILE_DIRECT=1<<4,
    FILE_LARGE_FILE=1<<5,
    FILE_NO_ACCESS_TIME=1<<6,
    FILE_NO_FOLLOW=1<<7,
    FILE_TRUNCATE=1<<8,
    FILE_EXCLUSIVE_CREATE=1<<9,
    FILE_UNCHANGED=1<<10,
    FILE_LAST_STATE=FILE_UNCHANGED,
  } FileMode;

  /**
   * @relates Conexus::MessageQueue
   */
  typedef enum QueueMode {
    QUEUE_READ=1<<0,
    QUEUE_WRITE=1<<1,
    QUEUE_CREATE=1<<2,
    QUEUE_EXCLUSIVE_CREATE=1<<3,
    QUEUE_NONBLOCKING=1<<4,
    QUEUE_UNCHANGED=1<<5,
  } QueueMode;

  /**
   * @relates Conexus::TTY
   */
  typedef enum TOS {
    TOS_LOWDELAY = IPTOS_LOWDELAY,       /**< minimize delays for interactive traffic */
    TOS_THROUGHPUT = IPTOS_THROUGHPUT,   /**< optimize throughput */
    TOS_RELIABILITY = IPTOS_RELIABILITY, /**< optimize for reliability */
    TOS_LOWCOST = IPTOS_LOWCOST,         /**< for "filler data" where slow transmission doesn’t matter */
  } TOS;

  /**
   * @relates Conexus::TTY
   * Aliases for the tty function options relating to when an option change takes effect.
   */
  typedef enum SetOption {
    /** the setting change occurs immediately */
    SET_NOW = TCSANOW,

    /**
     * the setting change occurs after all output written to the object
     * has been transmitted, and all input that has been received but not
     * read will be discarded before the change is made.
     */
    SET_FLUSH = TCSAFLUSH,

    /**
     * the setting change occurs after all output has been transmitted;
     * This function should be used when changing parameters that affect output.
     */
    SET_DRAIN = TCSADRAIN
  } SetOption;

  /**
   * @relates Conexus::TTY
   */
  typedef enum Parity {
    PARITY_NONE,  /** No parity check */
    PARITY_ODD,   /** Odd parity checks */
    PARITY_EVEN,  /** Even parity checks */
  } Parity;

  /**
   * @relates Conexus::TTY
   */
  typedef enum ParityError {
    PARITY_ERROR_DISCARD,  /** Discards all parity error bytes */
    PARITY_ERROR_MARK,     /** Delivers three bytes (0xFF, 0x00 and c; where c is the error byte */
    PARITY_ERROR_ZERO,     /** Delivers 0x00 when a parity error is detected */
    PARITY_ERROR_IGNORE,   /** Ignores all parity errors and delivers all data as is (this is not the same as IGNPAR) */
  } ParityError;

  /**
   * @relates Conexus::TTY
   */
  typedef enum FlowControl {
    FLOW_NONE,  /** CRTSCTS off, IXON off, IXOFF off */
    FLOW_HARDWARE,  /** CRTSCTS on, IXON off, IXOFF off */
    FLOW_SOFTWARE,  /** CRTSCTS off, IXON on, IXOFF on */
    FLOW_CUSTOM,  /** Combinations of CRTSCTS, IXON, IXOFF other than FLOW_NONE, FLOW_HARDWARE, or FLOW_SOFTWARE */
  } FlowControl;

}

#endif
