/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "except.h"

namespace Conexus
{

  void throw_exception ( int exception_num, ExceptionType type ) throw ( conexus_exception )
  {
    switch ( type )
    {
      case EXCEPTION_BIND:         throw_bind_exception ( exception_num );
      case EXCEPTION_CLOSE:        throw_close_exception ( exception_num );
      case EXCEPTION_CONNECT:      throw_connect_exception ( exception_num );
      case EXCEPTION_LISTEN:       throw_listen_exception ( exception_num );
      case EXCEPTION_OPEN:         throw_open_exception ( exception_num );
      case EXCEPTION_READ:         throw_read_exception ( exception_num );
      case EXCEPTION_WRITE:        throw_write_exception ( exception_num );
      case EXCEPTION_ADDRESS:      throw_address_exception ( exception_num );
      case EXCEPTION_SOCKET:       throw_socket_exception ( exception_num );
    }

    throw conexus_exception();
  }

  void throw_open_exception ( int exception_num ) throw ( open_exception )
  {
    switch ( exception_num )
    {
      case EPROTONOSUPPORT: throw exception::open::protocol_not_supported();
      case EAFNOSUPPORT:    throw exception::open::af_not_supported();
      case ENFILE:          throw exception::open::system_file();
      case EMFILE:          throw exception::open::process_file();
      case EACCES:          throw exception::open::access();
      case ENOBUFS:
      case ENOMEM:          throw exception::open::no_memory();
      case EINVAL:          throw exception::open::invalid_protocol();
      case EEXIST:          throw exception::open::exists();
      case EFAULT:          throw exception::open::fault();
      case EISDIR:          throw exception::open::is_directory();
      case ELOOP:           throw exception::open::loop();
      case ENAMETOOLONG:    throw exception::open::name_too_long();
      case ENODEV:          throw exception::open::no_device();
      case ENOENT:          throw exception::open::no_exist();
      case ENOSPC:          throw exception::open::no_space();
      case ENOTDIR:         throw exception::open::not_directory();
      case ENXIO:           throw exception::open::no_io();
      case EOVERFLOW:       throw exception::open::file_too_large();
      case EPERM:           throw exception::open::no_permission();
      case EROFS:           throw exception::open::read_only();
      case ETXTBSY:         throw exception::open::executable_busy();
      case EWOULDBLOCK:     throw exception::open::would_block();
      default:              throw open_exception();
    }
  }

  void throw_close_exception ( int exception_num ) throw ( close_exception )
  {
    switch ( exception_num )
    {
      case EBADF: throw exception::close::bad_fd();
      case EINTR: throw exception::close::interrupted();
      default:    throw close_exception();
    }
  }

  void throw_read_exception ( int exception_num ) throw ( read_exception )
  {
    switch ( exception_num )
    {
      case EBADF:        throw exception::read::bad_fd();
      case ECONNREFUSED: throw exception::read::connection_refused();
      case ENOTCONN:     throw exception::read::not_connected();
      case ENOTSOCK:     throw exception::read::not_socket();
      case EAGAIN:       throw exception::read::would_block();
      case EINTR:        throw exception::read::interrupted();
      case EFAULT:       throw exception::read::invalid_user_space_address();
      case EINVAL:       throw exception::read::invalid_argument();
      case ENOMEM:       throw exception::read::no_memory();
      case ETIMEDOUT:    throw exception::read::timeout();
      default:           throw read_exception();
    }
  }

  void throw_write_exception ( int exception_num ) throw ( write_exception )
  {
    switch ( exception_num )
    {
      case EWOULDBLOCK:  throw exception::write::would_block();
      case EBADF:        throw exception::write::bad_fd();
      case ECONNRESET:   throw exception::write::connection_reset();
      case EDESTADDRREQ: throw exception::write::destination_address_required();
      case EFAULT:       throw exception::write::invalid_user_space_address();
      case EINTR:        throw exception::write::interrupted();
      case EINVAL:       throw exception::write::invalid_argument();
      case EISCONN:      throw exception::write::is_connected();
      case EMSGSIZE:     throw exception::write::message_size();
      case ENOBUFS:      throw exception::write::no_buffers();
      case ENOMEM:       throw exception::write::no_memory();
      case ENOTCONN:     throw exception::write::not_connected();
      case ENOTSOCK:     throw exception::write::not_socket();
      case EOPNOTSUPP:   throw exception::write::operation_not_supported();
      case EPIPE:        throw exception::write::pipe();
      case ECONNREFUSED: throw exception::write::connection_refused();
      case EACCES:       throw exception::write::permission_denied();
      case ETIMEDOUT:    throw exception::write::timeout();
    }
    
    throw write_exception();
  }

  void throw_bind_exception ( int exception_num ) throw ( bind_exception )
  {
    switch ( exception_num )
    {
      case EBADF:        throw exception::bind::bad_fd();
      case EINVAL:       throw exception::bind::bound_socket();
      case EACCES:       throw exception::bind::access();
      case ENOTSOCK:     throw exception::bind::not_socket();
      case EROFS:        throw exception::bind::read_only_filesystem();
      case EFAULT:       throw exception::bind::segfault();
      case ENAMETOOLONG: throw exception::bind::name_too_long();
      case ENOENT:       throw exception::bind::not_exist();
      case ENOMEM:       throw exception::bind::no_kernel_memory();
      case ENOTDIR:      throw exception::bind::not_directory();
      case ELOOP:        throw exception::bind::loop();
    }
  
    throw bind_exception();
  }

  void throw_connect_exception ( int exception_num ) throw ( connect_exception )
  {
    switch ( exception_num )
    {
      case EBADF:        throw exception::connect::bad_fd();
      case EFAULT:       throw exception::connect::segfault();
      case ENOTSOCK:     throw exception::connect::not_socket();
      case EISCONN:      throw exception::connect::connected_socket();
      case ECONNREFUSED: throw exception::connect::connection_refused();
      case ETIMEDOUT:    throw exception::connect::timeout();
      case ENETUNREACH:  throw exception::connect::network_unreachable();
      case EADDRINUSE:   throw exception::connect::address_in_use();
      case EINPROGRESS:  throw exception::connect::not_immediate();
      case EALREADY:     throw exception::connect::in_progress();
      case EAGAIN:       throw exception::connect::try_again();
      case EAFNOSUPPORT: throw exception::connect::af_not_supported();
      case EACCES:
      case EPERM:        throw exception::connect::access();
    }
  
    throw connect_exception();
  }

  void throw_listen_exception ( int exception_num ) throw ( listen_exception )
  {
    switch ( exception_num )
    {
      case EADDRINUSE: throw exception::listen::address_in_use();
      case EBADF:      throw exception::listen::bad_fd();
      case ENOTSOCK:   throw exception::listen::not_socket();
      case EOPNOTSUPP: throw exception::listen::operation_not_supported();
    }
  
    throw listen_exception();
  }

  void throw_address_exception ( int exception_num ) throw ( address_exception )
  {
    switch ( exception_num )
    {
      case  EAI_AGAIN:      throw exception::address::again();
      case  EAI_BADFLAGS:   throw exception::address::bad_flags();
      case  EAI_FAIL:       throw exception::address::fail();
      case  EAI_MEMORY:     throw exception::address::memory();
      case  EAI_NONAME:     throw exception::address::no_name();
      case  EAI_SYSTEM:     throw exception::address::system();
      case  EAI_FAMILY:     throw exception::address::family();
      case  EAI_SOCKTYPE:   throw exception::address::socket_type();
      case  EAI_SERVICE:    throw exception::address::service();
      case  EAI_ADDRFAMILY: throw exception::address::address_family();
      case  EAI_NODATA:     throw exception::address::no_data();
    }
  
    throw address_exception();
  }

  void throw_socket_exception ( int exception_num ) throw ( socket_exception )
  {
    switch ( exception_num )
    {
      case EBADF:       throw exception::socket::bad_fd();
      case EFAULT:      throw exception::socket::segfault();
      case EINVAL:      throw exception::socket::optlen_invalid();
      case ENOPROTOOPT: throw exception::socket::no_option();
      case ENOTSOCK:    throw exception::socket::not_socket();
    }
  
    throw socket_exception();
  }

}
