/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSEXCEPT_H
#define CONEXUSEXCEPT_H

/**
 * @defgroup Exceptions Exceptions
 */

#include <stdexcept>
#include <errno.h>
#include <sstream>

namespace Conexus
{

  /**
   * Base class for all conexus library exceptions.
   * @author Rick L Vinyard Jr
   * @ingroup Exceptions
   */
  class conexus_exception: public std::runtime_error
  {
    public:
      conexus_exception ( const std::string& s="Unknown exception.", int e=0, const std::string& c="" ) :
          std::runtime_error ( s ),
          m_exception_number ( e ),
          m_exception_string ( s ),
          m_class ( c )
      {
        if ( m_exception_number == 0 )
          m_exception_number = errno;
        std::ostringstream ostr;
        ostr << "conexus:" << m_class << ":";
        if ( m_exception_number != 0 )
          ostr << "errno[" << m_exception_number << "]: ";
        else
          ostr << " ";
        ostr << m_exception_string;
        m_return_string = ostr.str();
      }

      ~conexus_exception() throw () { }

      virtual const char * what () const throw ()
      {
        return m_return_string.c_str();
      }

    protected:
      int m_exception_number;
      std::string m_return_string;
      std::string m_exception_string;
      std::string m_class;
  };

}

#include <conexus/except_address.h>
#include <conexus/except_bind.h>
#include <conexus/except_close.h>
#include <conexus/except_connect.h>
#include <conexus/except_listen.h>
#include <conexus/except_open.h>
#include <conexus/except_read.h>
#include <conexus/except_socket.h>
#include <conexus/except_state.h>
#include <conexus/except_write.h>

namespace Conexus
{

  typedef enum ExceptionType
  {
    EXCEPTION_ADDRESS,
    EXCEPTION_OPEN,
    EXCEPTION_BIND,
    EXCEPTION_CLOSE,
    EXCEPTION_CONNECT,
    EXCEPTION_LISTEN,
    EXCEPTION_READ,
    EXCEPTION_SOCKET,
    EXCEPTION_STATE,
    EXCEPTION_WRITE,
  } ExceptionType;

  void throw_exception ( int exception_num, ExceptionType type ) throw ( conexus_exception );

  void throw_bind_exception ( int exception_num ) throw ( bind_exception );
  void throw_close_exception ( int exception_num ) throw ( close_exception );
  void throw_connect_exception ( int exception_num ) throw ( connect_exception );
  void throw_listen_exception ( int exception_num ) throw ( listen_exception );
  void throw_open_exception ( int exception_num ) throw ( open_exception );
  void throw_read_exception ( int exception_num ) throw ( read_exception );
  void throw_write_exception ( int exception_num ) throw ( write_exception );
  void throw_address_exception ( int exception_num ) throw ( address_exception );
  void throw_socket_exception ( int exception_num ) throw ( socket_exception );

  /**
   * Interface index error
   * @author Rick L Vinyard Jr
   * @ingroup Exceptions
   */
  class interface_index_exception: public conexus_exception
  {
    public:
      interface_index_exception() : conexus_exception ( "Interface index error; name or number does not exist" ) { }
  };



} // namespace Conexus

#endif

