/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSEXCEPT_BIND_H
#define CONEXUSEXCEPT_BIND_H

#include <conexus/except.h>
#include <cstring>

namespace Conexus
  {

    /**
     * Base class for all binding exceptions.
     * @author Rick L Vinyard Jr
     * @ingroup Exceptions
     */
    class bind_exception: public conexus_exception
      {
      public:
      bind_exception(int e=0): conexus_exception(::strerror(e), e,"bind")
        { }
      bind_exception(const std::string s, int e=0): conexus_exception(s, e,"bind")
        { }
      }
    ;

  namespace exception
  {

  namespace bind
      {

      /**
       * EBADF: Thrown when equivalent BSD socket API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class bad_fd: public bind_exception
        {
        public:
          bad_fd(): bind_exception("sockfd is not a valid descriptor.", EBADF)
          { }
        }
      ;

      /**
       * EINVAL: Thrown when equivalent BSD socket API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class bound_socket: public bind_exception
        {
        public:
          bound_socket(): bind_exception("The socket is already bound to an address.", EINVAL)
          { }
        }
      ;

      /**
       * EACCES: Thrown when equivalent BSD socket API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class access: public bind_exception
        {
        public:
          access(): bind_exception("The address is protected, and the user is not the super-user.", EACCES)
          { }
        }
      ;

      /**
       * ENOTSOCK: Thrown when equivalent BSD socket API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_socket: public bind_exception
        {
        public:
          not_socket(): bind_exception("Argument is a descriptor for a file, not a socket.", ENOTSOCK)
          { }
        }
      ;

      /**
       * EINVAL: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class invalid_address: public bind_exception
        {
        public:
          invalid_address(): bind_exception("The addrlen is wrong, or the socket was not in the AF_UNIX family.", EINVAL)
          { }
        }
      ;

      /**
       * EROFS: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class read_only_filesystem: public bind_exception
        {
        public:
          read_only_filesystem(): bind_exception("The socket inode would reside on a read-only file system.", EROFS)
          { }
        }
      ;

      /**
       * EFAULT: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class segfault: public bind_exception
        {
        public:
          segfault(): bind_exception("addr points outside the user’s accessible address space.", EFAULT)
          { }
        }
      ;

      /**
       * ENAMETOOLONG: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class name_too_long: public bind_exception
        {
        public:
          name_too_long(): bind_exception("addr is too long.", ENAMETOOLONG)
          { }
        }
      ;

      /**
       * ENOENT: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_exist: public bind_exception
        {
        public:
          not_exist(): bind_exception("The file does not exist.", ENOENT)
          { }
        }
      ;

      /**
       * ENOMEM: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_kernel_memory: public bind_exception
        {
        public:
          no_kernel_memory(): bind_exception("Insufficient kernel memory was available.", ENOMEM)
          { }
        }
      ;

      /**
       * ENOTDIR: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_directory: public bind_exception
        {
        public:
          not_directory(): bind_exception("A component of the path prefix is not a directory.", ENOTDIR)
          { }
        }
      ;

      /**
       * ELOOP: Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class loop: public bind_exception
        {
        public:
          loop(): bind_exception("Too many symbolic links were encountered in resolving addr.", ELOOP)
          { }
        }
      ;

      /**
       * Thrown when bind is requested and udp socket doesn't have an address
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_address: public bind_exception
        {
        public:
          no_address(): bind_exception("cannot bind socket without an address.", -1)
          { }
        };

      /**
       * Thrown when bind is requested of the Socket class
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_socket_bind: public bind_exception
      {
      public:
      no_socket_bind(): bind_exception("The Socket class bind method cannot be called without an address.", -1)
        { }
      };

      /**
       * Thrown when bind is requested and a socket cannot be opened
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_open: public bind_exception
      {
      public:
      not_open(): bind_exception("Socket could not be opened for bind method.", -1)
        { }
      };

      }

  }

}

#endif // CONEXUSBIND_ERROR_H
