/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSEXCEPT_CONNECT_H
#define CONEXUSEXCEPT_CONNECT_H

#include <conexus/except.h>

namespace Conexus
{

  /**
   * Base class for all connect exceptions.
   * @author Rick L Vinyard Jr
   * @ingroup Exceptions
   */
  class connect_exception: public conexus_exception
  {
    public:
      connect_exception ( int e=0 ) : conexus_exception ( "Unknown exception, socket not connected.",e,"connect" ) { }
      connect_exception ( const std::string s, int e=0 ) : conexus_exception ( s,e,"connect" ) { }
  };

  namespace exception
  {


    namespace connect
    {


      /**
       * EBADF
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class bad_fd: public connect_exception
      {
        public:
          bad_fd() : connect_exception ( ::strerror ( EBADF ), EBADF ) { }
      };

      /**
       * EFAULT
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class segfault: public connect_exception
      {
        public:
          segfault() : connect_exception ( ::strerror ( EFAULT ), EFAULT ) { }
      };

      /**
       * ENOTSOCK
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_socket: public connect_exception
      {
        public:
          not_socket() : connect_exception ( ::strerror ( ENOTSOCK ), ENOTSOCK ) { }
      };

      /**
       * EISCONN
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class connected_socket: public connect_exception
      {
        public:
          connected_socket() : connect_exception ( ::strerror ( EISCONN ), EISCONN ) { }
      };

      /**
       * ECONNREFUSED
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class connection_refused: public connect_exception
      {
        public:
          connection_refused() : connect_exception ( ::strerror(ECONNREFUSED), ECONNREFUSED ) { }
      };

      /**
       * ETIMEDOUT
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class timeout: public connect_exception
      {
        public:
          timeout() : connect_exception ( ::strerror(ETIMEDOUT), ETIMEDOUT ) { }
      };

      /**
       * ENETUNREACH
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class network_unreachable: public connect_exception
      {
        public:
          network_unreachable() : connect_exception ( ::strerror(ENETUNREACH), ENETUNREACH ) { }
      };

      /**
       * EADDRINUSE
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class address_in_use: public connect_exception
      {
        public:
          address_in_use() : connect_exception ( ::strerror(EADDRINUSE), EADDRINUSE ) { }
      };

      /**
       * EINPROGRESS
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_immediate: public connect_exception
      {
        public:
          not_immediate() : connect_exception ( ::strerror(EINPROGRESS), EINPROGRESS ) { }
      };

      /**
       * EALREADY
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class in_progress: public connect_exception
      {
        public:
          in_progress() : connect_exception ( ::strerror(EALREADY), EALREADY ) { }
      };

      /**
       * EAGAIN
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class try_again: public connect_exception
      {
        public:
          try_again() : connect_exception ( ::strerror(EAGAIN), EAGAIN ) { }
      };

      /**
       * EAFNOSUPPORT
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class af_not_supported: public connect_exception
      {
        public:
          af_not_supported() : connect_exception ( ::strerror(EAFNOSUPPORT), EAFNOSUPPORT ) { }
      };

      /**
       * EACCES or EPERM
       * Thrown when equivalent UNIX domain (AF_UNIX) API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class access: public connect_exception
      {
        public:
          access() : connect_exception ( ::strerror(EACCES), EACCES ) { }
      };

      /**
       * Thrown when attempting to autoconnect to a socket that is not capable of
       * autoconnecting or does not have autoconnection enabled.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_remote_address: public connect_exception
      {
        public:
          no_remote_address() : connect_exception ( "Socket cannot auto connect because a default remote address has not been set.", -1 ) { }
      };

      /**
       * Thrown when connect is requested of the Socket class
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_socket_connect: public connect_exception
      {
        public:
          no_socket_connect() : connect_exception ( "The Socket class connect() method cannot be called without an address.", -1 )
          { }
      };

      /**
       * Thrown when connect is requested and a socket cannot be opened
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_open: public connect_exception
      {
        public:
          not_open() : connect_exception ( "Socket could not be opened for connect() method.", -1 )
          { }
      };

      /**
       * Thrown when connect is requested of the UDPPoset class
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_udpposet_connect: public connect_exception
      {
        public:
          no_udpposet_connect() : connect_exception ( "The UDPPoset class connect() method cannot be called.", -1 )
          { }
      };



    }
  }
}

#endif // CONEXUSCONNECT_ERROR_H
