/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSEXCEPT_READ_H
#define CONEXUSEXCEPT_READ_H

#include <conexus/except.h>

namespace Conexus
{

  /**
   * Base class for all receiving exceptions.
   * @author Rick L Vinyard Jr
   * @ingroup Exceptions
   */
  class read_exception: public conexus_exception
  {
    public:
      read_exception ( ) : conexus_exception ( "Unknown exception, read failed.",0,"read" )
      { }
      read_exception ( int e ) : conexus_exception ( strerror ( e ),e,"read" )
      { }
      read_exception ( const std::string s ) : conexus_exception ( s,0,"read" )
      { }
  };

  namespace exception
  {

    namespace read
    {

      /**
       * Base class for all read exceptions that can be retried without intervention
       *
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class retry: public read_exception
      {
        public:
          retry() : read_exception ( "Unknown exception, read failed but should be retried." )
          { }
          retry ( int e ) : read_exception ( e )
          { }
          retry ( const std::string s ) : read_exception ( s )
          { }
      };

      /**
       * EBADF: Thrown when equivalent BSD socket API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class bad_fd: public read_exception
      {
        public:
          bad_fd() : read_exception ( EBADF )
          { }
      };

      /**
       * ECONNREFUSED
       * A remote host refused to allow the network connection (typically
       * because it is not running the requested service).
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class connection_refused: public read_exception
      {
        public:
          connection_refused() : read_exception ( ECONNREFUSED )
          { }
      };

      /**
       * ENOTCONN
       * The socket is associated with a connection-oriented protocol and
       * has not been connected (see connect(2) and accept(2)).
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_connected: public read_exception
      {
        public:
          not_connected() : read_exception ( ENOTCONN )
          { }
      };

      /**
       * ENOTSOCK
       * The argument s is not a socket.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_socket: public read_exception
      {
        public:
          not_socket() : read_exception ( ENOTSOCK )
          { }
      };

      /**
       * EAGAIN
       * The  socket  is  marked  non-blocking  and the receive operation
       * would block, or a receive timeout had been set and  the  timeout
       * expired before data was received.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class would_block: public retry
      {
        public:
          would_block() : retry ( EAGAIN )
          { }
      };

      /**
       * EINTR
       * The  receive  was interrupted by delivery of a signal before any
       * data were available.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class interrupted: public retry
      {
        public:
          interrupted() : retry ( EINTR )
          { }
      };

      /**
       * EFAULT
       * An invalid user space address was specified for a parameter.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class invalid_user_space_address: public read_exception
      {
        public:
          invalid_user_space_address() : read_exception ( EFAULT )
          { }
      };

      /**
       * EINVAL
       * Invalid argument passed.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class invalid_argument: public read_exception
      {
        public:
          invalid_argument() : read_exception ( EINVAL )
          { }
      };

      /**
       * ENOMEM
       * No memory available.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_memory: public read_exception
      {
        public:
          no_memory() : read_exception ( ENOMEM )
          { }
      };

      /**
       * Not opened.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_opened: public read_exception
      {
        public:
          not_opened() : read_exception ( "Trying to read from a device that can't be opened." )
          { }
      };

      /**
       * Not bound.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_bound: public read_exception
      {
        public:
          not_bound() : read_exception ( "Trying to read from a socket that can't be bound." )
          { }
      };

      /**
       * Disconnected.
       *
       * Trying to read on a socket where the peer has performed an orderly shutdown.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class disconnected: public read_exception
      {
        public:
          disconnected() : read_exception ( "Trying to read from a socket where the peer has performed an orderly shutdown." )
          { }
      };

      /**
       * Timeout.
       *
       * The read operation timed out.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class timeout: public retry
      {
        public:
          timeout() : retry ( ETIMEDOUT )
          { }
      };

      /**
       * End of file.
       *
       * An end of file condition occurred.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class eof: public read_exception
      {
        public:
          eof() : read_exception ( "End of file." )
          { }
      };


    }

  }

}

#endif // CONEXUSSEND_ERROR_H
