/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSEXCEPT_WRITE_H
#define CONEXUSEXCEPT_WRITE_H

#include <conexus/except.h>
#include <cstring>

namespace Conexus
{

  /**
   * Base class for all write exceptions.
   * @author Rick L Vinyard Jr
   * @ingroup Exceptions
   */
  class write_exception: public conexus_exception
  {
    public:
      write_exception() : conexus_exception ( "Unknown exception, write failed.",0,"write" )
      { }
      write_exception ( int e ) : conexus_exception ( strerror ( e ),e,"write" )
      { }
      write_exception ( const std::string s ) : conexus_exception ( s,0,"write" )
      { }
  };

  namespace exception
  {

    namespace write
    {

      /**
       * Base class for all write exceptions that can be retried without intervention
       *
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class retry: public write_exception
      {
        public:
          retry() : write_exception ( "Unknown exception, write failed but should be retried." )
          { }
          retry ( int e ) : write_exception ( e )
          { }
          retry ( const std::string s ) : write_exception ( s )
          { }
      };

      /**
       * EAGAIN or EWOULDBLOCK
       * The socket is marked non-blocking and  the  requested  operation
       * would block.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class would_block: public retry
      {
        public:
          would_block() : retry ( EAGAIN )
          { }
      }
      ;

      /**
       * EBADF: Thrown when equivalent BSD socket API error is encountered. May be caught alone
       * or via parent class.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class bad_fd: public write_exception
      {
        public:
          bad_fd() : write_exception ( EBADF )
          { }
      };

      /**
       * ECONNRESET
       * Connection reset by peer.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class connection_reset: public write_exception
      {
        public:
          connection_reset() : write_exception ( ECONNRESET )
          { }
      };

      /**
       * EDESTADDRREQ
       * The socket is not connection-mode, and no peer address is set.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class destination_address_required: public write_exception
      {
        public:
          destination_address_required() : write_exception ( EDESTADDRREQ )
          { }
      };


      /**
       * EFAULT
       * An invalid user space address was specified for a parameter.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class invalid_user_space_address: public write_exception
      {
        public:
          invalid_user_space_address() : write_exception ( EFAULT )
          { }
      };

      /**
       * EINTR
       * A signal occurred before any data was transmitted.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class interrupted: public retry
      {
        public:
          interrupted() : retry ( EINTR )
          { }
      };


      /**
       * EINVAL
       * Invalid argument passed.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class invalid_argument: public write_exception
      {
        public:
          invalid_argument() : write_exception ( EINVAL )
          { }
      };

      /**
       * EISCONN
       * The connection-mode socket was connected already but a recipient
       * was specified.  (Now either  this  error  is  returned,  or  the
       * recipient specification is ignored.)
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class is_connected: public write_exception
      {
        public:
          is_connected() : write_exception ( EISCONN )
          { }
      };


      /**
       * EMSGSIZE
       * The  socket  type  requires that message be sent atomically, and
       * the size of the message to be sent made this impossible.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class message_size: public write_exception
      {
        public:
          message_size() : write_exception ( EMSGSIZE )
          { }
      };

      /**
       * ENOBUFS
       * The output queue for a network interface was full.  This  gener-
       * ally  indicates  that the interface has stopped writing, but may
       * be caused by transient congestion.   (Normally,  this  does  not
       * occur  in Linux. Packets are just silently dropped when a device
       * queue overflows.)
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_buffers: public retry
      {
        public:
          no_buffers() : retry ( ENOBUFS )
          { }
      };


      /**
       * ENOMEM
       * No memory available.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_memory: public write_exception
      {
        public:
          no_memory() : write_exception ( ENOMEM )
          { }
      };

      /**
       * EACCES
       * Permission denied.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class permission_denied: public write_exception
      {
        public:
          permission_denied() : write_exception ( EACCES )
          { }
      };

      /**
       * ENOTCONN
       * The socket is not connected, and no target has been given.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_connected: public write_exception
      {
        public:
          not_connected() : write_exception ( ENOTCONN )
          { }
      };

      /**
       * ENOTSOCK
       * The argument s is not a socket.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_socket: public write_exception
      {
        public:
          not_socket() : write_exception ( ENOTSOCK )
          { }
      };


      /**
       * EOPNOTSUPP
       * Some bit in the flags argument is inappropriate for  the  socket
       * type.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class operation_not_supported: public write_exception
      {
        public:
          operation_not_supported() : write_exception ( EOPNOTSUPP )
          { }
      };


      /**
       * EPIPE
       * The  local  end  has  been  shut  down  on a connection oriented
       * socket.  In this case the process will also  receive  a  SIGPIPE
       * unless MSG_NOSIGNAL is set.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class pipe: public write_exception
      {
        public:
          pipe() : write_exception ( EPIPE )
          { }
      };

      /**
       * ECONNREFUSED
       * With writes, this is most likely triggered by an ICMP UNREACHABLE message.
       * Therefore, the first one (or even several) messages may transmit before
       * the ICMP mechanism reports UNREACHABLE.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class connection_refused: public write_exception
      {
        public:
          connection_refused() : write_exception ( ECONNREFUSED )
          { }
      };

      /**
       * Thrown when attempting to send without specifying a default address.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_default_remote_address: public write_exception
      {
        public:
          no_default_remote_address() : write_exception ( "Socket cannot send because a default remote address has not been set." )
          { }
      };

      /**
       * Thrown when sendto without an address is requested of the Socket class
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class no_sendto_without_address: public write_exception
      {
        public:
          no_sendto_without_address() : write_exception ( "Socket is not capable of sendto without an address." )
          { }
      };

      /**
       * Writing on an I/O than can't open
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class not_opened: public write_exception
      {
        public:
          not_opened() : write_exception ( "Trying to write on I/O that cannot open." )
          { }
      };

      /**
       * Writing to a read-only I/O
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class read_only: public write_exception
      {
        public:
          read_only() : write_exception ( "Trying to write to a read only I/O." )
          { }
      };

      /**
       * Timeout.
       *
       * The write operation timed out.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class timeout: public retry
      {
        public:
          timeout() : retry ( ETIMEDOUT )
          { }
      };

      /**
       * Disconnected.
       *
       * Trying to write on a socket where the peer has performed an orderly shutdown.
       * @author Rick L Vinyard Jr
       * @ingroup Exceptions
       */
      class disconnected: public write_exception
      {
        public:
          disconnected() : write_exception ( "Trying to write from a socket where the peer has performed an orderly shutdown." )
          { }
      };




    }

  }

}

#endif // CONEXUSSEND_ERROR_H
