/***************************************************************************
 *   Copyright (C) 2007 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "factory.h"
#include "utility.h"

namespace Conexus
{

  Factory::Factory()
  {}


  Factory::pointer Factory::create( )
  {
    return pointer ( new Factory() );
  }

  Factory::~Factory()
  {}

  const Factory::Names & Factory::names( )
  {
    return m_names;
  }

  const Factory::IndexNameMap & Factory::index_name_map( )
  {
    return m_index_name;
  }

  const Factory::NameIndexMap & Factory::name_index_map( )
  {
    return m_name_index;
  }

  unsigned long Factory::index ( const std::string & name )
  {
    NameIndexMap::iterator iter = m_name_index.find ( name );
    if ( iter == m_name_index.end() )
      return 0;
    return iter->second;
  }

  unsigned long Factory::index ( EndpointCreator cr )
  {
    IndexCreatorMap::iterator iter;
    for ( iter = m_index_creator.begin(); iter != m_index_creator.end(); iter++ )
    {
      if ( iter->second == cr )
        return iter->first;
    }
    return 0;
  }

  unsigned long Factory::index ( EndpointServerCreator cr )
  {
    IndexServerCreatorMap::iterator iter;
    for ( iter = m_index_servercreator.begin(); iter != m_index_servercreator.end(); iter++ )
    {
      if ( iter->second == cr )
        return iter->first;
    }
    return 0;
  }

  std::string Factory::name ( unsigned long index )
  {
    IndexNameMap::iterator iter = m_index_name.find ( index );
    if ( iter == m_index_name.end() )
      return std::string();
    return iter->second;
  }

  std::string Factory::name ( EndpointCreator cr )
  {
    NameCreatorMap::iterator iter;
    for ( iter = m_name_creator.begin(); iter != m_name_creator.end(); iter++ )
    {
      if ( iter->second == cr )
        return iter->first;
    }
    return std::string();
  }

  std::string Factory::name ( EndpointServerCreator cr )
  {
    NameServerCreatorMap::iterator iter;
    for ( iter = m_name_servercreator.begin(); iter != m_name_servercreator.end(); iter++ )
    {
      if ( iter->second == cr )
        return iter->first;
    }
    return std::string();
  }

  unsigned long Factory::add ( const std::string & name, EndpointCreator cr, unsigned long index )
  {
    if ( name == std::string() )
      return 0;

    if ( m_name_index.find ( name ) != m_name_index.end() )
      return 0;

    if ( index != 0 && m_index_name.find ( index ) != m_index_name.end() )
      return 0;

    // if index is 0 we need to assign an unused index
    if ( index == 0 )
    {
      // if this is the first item registered, then let's just use 1 as the index
      if ( m_index_name.size() == 0 )
      {
        index = 1;
      }
      else
      {
        // let's try and use the last value + 1
        index = m_index_name.rbegin()->first + 1;
        // and check for rollover of the unsigned value
        if ( index == 0 )
        {
          // if we had a rollover we need to search for the first unused value
          IndexNameMap::iterator iter;
          for ( iter = m_index_name.begin(); iter != m_index_name.end() && iter->first == ++index; iter++ );
          // did we roll though all the numbers? That's more items that I ever expected would be registered
          if ( index == 0 )
            return 0;
        }
      }
    }

    // At this point we have an unused name and index
    m_names.insert ( name );
    m_name_index[name] = index;
    m_index_name[index] = name;
    m_index_creator[index] = cr;
    m_name_creator[name] = cr;

    m_signal_added.emit ( name );

    return index;
  }

  unsigned long Factory::add ( const std::string & name, EndpointServerCreator cr, unsigned long index )
  {
    if ( name == std::string() )
      return 0;

    if ( m_name_index.find ( name ) != m_name_index.end() )
      return 0;

    if ( index != 0 && m_index_name.find ( index ) != m_index_name.end() )
      return 0;

    // if index is 0 we need to assign an unused index
    if ( index == 0 )
    {
      // if this is the first item registered, then let's just use 1 as the index
      if ( m_index_name.size() == 0 )
      {
        index = 1;
      }
      else
      {
        // let's try and use the last value + 1
        index = m_index_name.rbegin()->first + 1;
        // and check for rollover of the unsigned value
        if ( index == 0 )
        {
          // if we had a rollover we need to search for the first unused value
          IndexNameMap::iterator iter;
          for ( iter = m_index_name.begin(); iter != m_index_name.end() && iter->first == ++index; iter++ );
          // did we roll though all the numbers? That's more items that I ever expected would be registered
          if ( index == 0 )
            return 0;
        }
      }
    }

    // At this point we have an unused name and index
    m_names.insert ( name );
    m_name_index[name] = index;
    m_index_name[index] = name;
    m_index_servercreator[index] = cr;
    m_name_servercreator[name] = cr;

    m_signal_added.emit ( name );

    return index;
  }

  bool Factory::remove ( unsigned long index )
  {
    if ( m_index_name.find ( index ) == m_index_name.end() )
      return false;

    std::string name = m_index_name[index];

    m_names.erase ( name );
    m_index_name.erase ( index );
    m_index_creator.erase ( index );
    m_index_servercreator.erase ( index );
    m_name_index.erase ( name );
    m_name_creator.erase ( name );
    m_name_servercreator.erase ( name );

    m_signal_removed.emit ( name );

    return true;
  }

  bool Factory::remove ( const std::string & name )
  {
    if ( m_name_index.find ( name ) == m_name_index.end() )
      return false;

    unsigned long index = m_name_index[name];

    m_names.erase ( name );
    m_index_name.erase ( index );
    m_index_creator.erase ( index );
    m_index_servercreator.erase ( index );
    m_name_index.erase ( name );
    m_name_creator.erase ( name );
    m_name_servercreator.erase ( name );

    m_signal_removed.emit ( name );

    return true;
  }

  bool Factory::remove ( EndpointCreator cr )
  {
    unsigned long index = this->index ( cr );
    if ( index == 0 )
      return false;
    return this->remove ( index );
  }

  bool Factory::remove ( EndpointServerCreator cr )
  {
    unsigned long index = this->index ( cr );
    if ( index == 0 )
      return false;
    return this->remove ( index );
  }

  Endpoint::pointer Factory::create_endpoint ( unsigned long index, const std::string& instance_name )
  {
    IndexCreatorMap::iterator iter = m_index_creator.find ( index );
    if ( iter == m_index_creator.end() )
      return Endpoint::pointer();
    Endpoint::pointer created = iter->second();
    if ( instance_name.size() > 0 )
      this->registry().add( instance_name, created );
    return created;
  }

  Endpoint::pointer Factory::create_endpoint ( const std::string & class_name, const std::string& instance_name )
  {
    NameCreatorMap::iterator iter = m_name_creator.find ( class_name );
    if ( iter == m_name_creator.end() )
      return Endpoint::pointer();
    Endpoint::pointer created = iter->second();
    if ( instance_name.size() > 0 )
      this->registry().add( instance_name, created );
    return created;
  }

  EndpointServer::pointer Factory::create_server ( unsigned long index, const std::string& instance_name )
  {
    IndexServerCreatorMap::iterator iter = m_index_servercreator.find ( index );
    if ( iter == m_index_servercreator.end() )
      return EndpointServer::pointer();
    EndpointServer::pointer created = iter->second();
    if ( instance_name.size() > 0 )
      this->registry().add( instance_name, created );
    return created;
  }

  EndpointServer::pointer Factory::create_server ( const std::string & class_name, const std::string& instance_name )
  {
    NameServerCreatorMap::iterator iter = m_name_servercreator.find ( class_name );
    if ( iter == m_name_servercreator.end() )
      return EndpointServer::pointer();
    return iter->second();
  }

  Object::pointer Factory::create ( unsigned long index, const std::string& instance_name )
  {
    Endpoint::pointer ep = this->create_endpoint ( index, instance_name );
    if ( ep )
      return ep;
    return this->create_server ( index, instance_name );
  }

  Object::pointer Factory::create ( const std::string & class_name, const std::string& instance_name )
  {
    Endpoint::pointer ep = this->create_endpoint ( class_name, instance_name );
    if ( ep )
      return ep;
    return this->create_server ( class_name, instance_name );
  }

  sigc::signal< void, const std::string& > Factory::signal_added( )
  {
    return m_signal_added;
  }

  sigc::signal< void, const std::string& > Factory::signal_removed( )
  {
    return m_signal_removed;
  }

  Registry& Factory::registry( )
  {
    if ( m_registry )
      return *m_registry;
    return default_registry();
  }

  void Factory::set_registry ( Registry::pointer registry )
  {
    m_registry = registry;
  }

}

