/***************************************************************************
 *   Copyright (C) 2007 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSFACTORY_H
#define CONEXUSFACTORY_H

#include <set>
#include <map>

#include <sigc++/sigc++.h>

#include <conexus/endpoint.h>
#include <conexus/endpointserver.h>
#include <conexus/pointer.h>
#include <conexus/registry.h>

/**
 * \example factory_client.cpp
 * Conexus::Factory client example
 *
 * \example factory_server.cpp
 * Conexus::Factory server example
 */

namespace Conexus
{

  /**
    @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
  */
  class Factory : public sigc::trackable
  {
    public:

      typedef ConexusPointer<Factory> pointer;

      typedef Endpoint::pointer(*EndpointCreator)();

      typedef EndpointServer::pointer(*EndpointServerCreator)();

      typedef std::set<std::string> Names;

      typedef std::map<unsigned long,std::string> IndexNameMap;

      typedef std::map<std::string,unsigned long> NameIndexMap;

      Factory();

      static pointer create();

      ~Factory();

      const Names& names();

      const IndexNameMap& index_name_map();

      const NameIndexMap& name_index_map();

      unsigned long index ( const std::string& name );

      unsigned long index ( EndpointCreator cr );

      unsigned long index( EndpointServerCreator cr );

      std::string name ( unsigned long index );

      std::string name ( EndpointCreator cr );

      std::string name( EndpointServerCreator cr );

      /**
       * Adds the endpoint creator by name to the factory.
       * If index is 0, an unused index will be assigned.
       * @return The index assigned to this creator or 0 if the add failed.
       */
      unsigned long add ( const std::string& name, EndpointCreator cr, unsigned long index=0 );

      /**
       * Adds the endpoint server creator by name to the factory.
       * If index is 0, an unused index will be assigned.
       * @return The index assigned to this creator or 0 if the add failed.
       */
      unsigned long add ( const std::string& name, EndpointServerCreator cr, unsigned long index=0 );

      bool remove ( unsigned long index );

      bool remove ( const std::string& name );

      bool remove ( EndpointCreator cr );

      bool remove ( EndpointServerCreator cr );

      Endpoint::pointer create_endpoint( unsigned long index, const std::string& instance_name=std::string() );

      Endpoint::pointer create_endpoint( const std::string& class_name, const std::string& instance_name=std::string() );

      EndpointServer::pointer create_server( unsigned long index, const std::string& instance_name=std::string() );

      EndpointServer::pointer create_server( const std::string& class_name, const std::string& instance_name=std::string() );

      Object::pointer create( unsigned long index, const std::string& instance_name=std::string() );

      Object::pointer create( const std::string& class_name, const std::string& instance_name=std::string() );

      sigc::signal<void,const std::string&> signal_added();

      sigc::signal<void,const std::string&> signal_removed();

      Registry& registry();

      void set_registry( Registry::pointer registry );

    protected:

      typedef std::map<unsigned long,EndpointCreator> IndexCreatorMap;
      typedef std::map<std::string,EndpointCreator> NameCreatorMap;
      typedef std::map<unsigned long,EndpointServerCreator> IndexServerCreatorMap;
      typedef std::map<std::string,EndpointServerCreator> NameServerCreatorMap;

      Names m_names;
      IndexCreatorMap m_index_creator;
      NameCreatorMap m_name_creator;
      IndexServerCreatorMap m_index_servercreator;
      NameServerCreatorMap m_name_servercreator;
      IndexNameMap m_index_name;
      NameIndexMap m_name_index;
      Registry::pointer m_registry;

      sigc::signal<void,const std::string&> m_signal_added;

      sigc::signal<void,const std::string&> m_signal_removed;

  };

}

#endif
