/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "file.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

namespace Conexus
{

  File::File ( const std::string& filename, long unsigned mode ) :
      m_eof ( true ), m_throw_eof ( false )
  {
    if ( mode & FILE_UNCHANGED )
      mode = mode & ~FILE_UNCHANGED;
    this->set_filename ( filename, mode );
  }

  File::pointer File::create ( const std::string& filename, long unsigned mode )
  {
    return pointer ( new File ( filename, mode ) );
  }

  File::~File()
  {}

  void File::open ( const std::string name, long unsigned mode ) throw ( open_exception )
  {
    int flags=0;

    if ( is_open() )
      change_state ( ENDPOINT_CLOSED );

    if ( ! ( mode & FILE_UNCHANGED ) )
      m_mode = mode;

    if ( m_mode == 0 ) throw exception::open::invalid_file_mode();

    if ( m_mode & ( FILE_READ|FILE_WRITE ) )
      flags |= O_RDWR;
    else if ( m_mode & FILE_READ )
      flags |= O_RDONLY;
    else if ( m_mode & FILE_WRITE )
      flags |= O_WRONLY;

    if ( m_mode & FILE_APPEND )
      flags |= O_APPEND;
    if ( m_mode & FILE_CREATE )
      flags |= O_CREAT;
    if ( m_mode & FILE_DIRECT )
      flags |= O_DIRECT;
    if ( m_mode & FILE_LARGE_FILE )
      flags |= O_LARGEFILE;
    if ( m_mode & FILE_NO_ACCESS_TIME )
      flags |= O_NOATIME;
    if ( m_mode & FILE_NO_FOLLOW )
      flags |= O_NOFOLLOW;
    if ( m_mode & FILE_TRUNCATE )
      flags |= O_TRUNC;
    if ( m_mode & FILE_EXCLUSIVE_CREATE )
      flags |= O_EXCL;

    if ( m_mode & FILE_CREATE )
      m_fd = ::open ( name.c_str(), flags, S_IREAD|S_IWRITE );
    else
      m_fd = ::open ( name.c_str(), flags );

    if ( m_fd == -1 )
      throw open_exception ( errno );

    m_eof = false;

    this->set_state_opened();
  }

  void File::open( ) throw ( open_exception )
  {
    if ( m_filename == std::string() )
      throw ( exception::open::no_filename() );
    this->open ( m_filename );
  }

  std::string File::filename( )
  {
    return m_filename;
  }

  long unsigned File::mode( )
  {
    return m_mode;
  }

  void File::set_filename ( std::string d, long unsigned mode ) throw ( open_exception )
  {
    long tempstate;

    // if the filename is the same and the state is the same, then leave
    if ( m_filename == d && m_mode == mode )
      return;

    tempstate = m_state;

    if ( this->is_open() )
      this->close();

    m_filename = d;
    if ( ! ( m_mode & FILE_UNCHANGED ) )
      m_mode = mode;

    this->change_state ( tempstate );

    m_signal_filename_changed.emit();
  }

  sigc::signal<void> File::signal_filename_changed()
  {
    return m_signal_filename_changed;
  }

  Data File::read_data ( size_t size, Timeout timeout ) throw ( read_exception )
  {
    Data d = FileDescriptor::read_data ( size, timeout );
    if ( ! d )
    {
      m_eof = true;
      if ( m_throw_eof ) throw exception::read::eof();
      else m_signal_eof.emit();
      return Data();
    }
    return d;
  }

  bool File::eof( )
  {
    return m_eof;
  }

  void File::set_eof(bool state)
  {
    m_eof = state;
  }

  bool File::is_throw_eof_set( )
  {
    return m_throw_eof;
  }

  void File::set_throw_eof ( bool teof )
  {
    m_throw_eof = teof;
  }

}

