/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSFILE_H
#define CONEXUSFILE_H

#include <conexus/enums.h>
#include <conexus/filedescriptor.h>

/**
 * \example file_client.cpp
 * Conexus::File write example
 *
 * \example file_server.cpp
 * Conexus::File read example
 *
 * \example file_eof.cpp
 * Conexus::File EOF handling example
 */

namespace Conexus
{

  /**
   * This class encapsulates a file.
   *
   * TODO add support for stat(2)
   *
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   * @ingroup Endpoint
   */
  class File : public FileDescriptor
  {
    protected:

      File( const std::string& filename,
            long unsigned mode );

    public:
      typedef ConexusPointer<File> pointer;

      CONEXUS_ENDPOINT_GENERIC_CREATE();

      static File::pointer create( const std::string& filename = std::string(),
                                   long unsigned mode = 0 );

      virtual ~File();

      /**
       * Opens the named file in the specified mode.
       */
      virtual void open( const std::string name,
                         long unsigned mode = FILE_UNCHANGED )
          throw ( open_exception);

      /**
       * Will attempt to open the file if a filename has been specified. Otherwise an
       * exception will be thrown.
       */
      virtual void open() throw ( open_exception );

      /** Will return the filename. */
      std::string filename();

      /** Returns the POSIX file mode of the endpoint */
      long unsigned mode();

      /** True if file has encountered an EOF condition, false otherwise */
      bool eof();

      /**
       * Forces the eof state to a specific value
       *
       * Setting this to false sets the internal eof state to false, even
       * if an EOF condition has been encountered. This effectively resets
       * the internal state of the endpoint but has no actual effect on the
       * file. If the file is truly EOF it will remain EOF.
       */
      void set_eof(bool state=true);

      /**
       * If true, file endpoint will throw an eof exception when an EOF condition is encountered
       *
       * The default setting for any endpoint is to \b not throw an EOF exception
       */
      bool is_throw_eof_set();

      /**
       * Sets the endpoint behavior when an EOF condition is encountered
       *
       * @param teof If set to true, endpoint will throw an exception when an EOF is encountered
       */
      void set_throw_eof( bool teof = true );

      /**
       * Sets the filename.
       *
       * If the file is currently open and d is not equal to the current filename,
       * the current file will be closed.
       *
       * Also, if auto_open is set, the new file will be immediately opened.
       *
       * Additionally, this method checks whether the parameter d is the same as the current filename,
       * and parameter mode is the same as the current mode.
       * If both are equal, the file will not be closed and reopened unless force=true.
       */
      void set_filename( std::string d, long unsigned mode = FILE_UNCHANGED ) throw ( open_exception );

      sigc::signal<void> signal_filename_changed();

    protected:
      std::string m_filename;
      long unsigned m_mode;
      bool m_eof;
      bool m_throw_eof;

      sigc::signal<void> m_signal_filename_changed;

      /**
       * Reimplements Endpoint::read_data() using FileDescriptor::read_data()
       * to catch the exception where data read size is zero indicating EOF.
       * When encountered throws eof exception.
       * @throw Conexus::exception::read::eof When file EOF is encounter on a zero length read
       */
      virtual Data read_data( size_t size, Timeout timeout ) throw ( read_exception );

  };

}

#endif
