/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "filedescriptor.h"

#include <fcntl.h>

namespace Conexus
{

  FileDescriptor::FileDescriptor ( int fd, bool close_on_destruct ) :
      Endpoint ( close_on_destruct ),
      m_fd ( fd )
  {}


  FileDescriptor::~FileDescriptor()
  {
    if ( m_close_on_destruct && this->is_open() )
      this->close();
  }

  int FileDescriptor::fd() throw ()
  {
    return m_fd;
  }

  void FileDescriptor::close ( bool force ) throw ( close_exception )
  {
    if ( this->is_closed() && !force )
      return ;

    int i = ::close ( m_fd );
    if ( i == -1 )
      throw_close_exception ( errno );

    m_fd = -1;

    this->set_state_closed();
  }

  bool FileDescriptor::is_blocking()
  {
    int fcntl_flags;

    if ( ! this->is_open() ) return true;
    
    fcntl_flags = ::fcntl ( m_fd, F_GETFL );
    if ( fcntl_flags == -1 ) throw -1;
    return ! (fcntl_flags & O_NONBLOCK);
  }

  bool FileDescriptor::set_blocking(bool block)
  {
    int fcntl_flags;
    int fcntl_result;

    if ( ! this->is_open() ) return false;
    
    fcntl_flags = ::fcntl ( m_fd, F_GETFL );
    if ( fcntl_flags == -1 ) throw -1;

    if ( block )
    {
      if ( ! fcntl_flags & O_NONBLOCK ) return true;
      fcntl_flags &= ~O_NONBLOCK;
      fcntl_result = ::fcntl ( m_fd, F_SETFL, fcntl_flags );
      if ( fcntl_result == -1 ) throw_write_exception ( errno );
    }
    else
    {
      if ( fcntl_flags & O_NONBLOCK ) return true;
      fcntl_flags |= O_NONBLOCK;
      fcntl_result = ::fcntl ( m_fd, F_SETFL, fcntl_flags );
      if ( fcntl_result == -1 ) throw -1;
    }

    return true;
  }

  size_t FileDescriptor::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
  {
    int write_result;

    // Do we have valid data?
    if ( ! data ) throw_write_exception ( EINVAL );

    // Is the file descriptor open?
    if ( this->is_closed() ) this->open();

    // We need to do some prep work based on whether the timeout indicates
    // a blocking write, a non-blocking immediate write, or a blocking with timeout
    if ( timeout.is_null() )
    {
      this->set_blocking(true);
    }
    else
    {
      this->set_blocking(false);
      
      fd_set writefd;
      struct timeval tv_timeout = timeout;

      FD_ZERO ( &writefd );
      FD_SET ( m_fd, &writefd );
      switch ( ::select ( m_fd + 1, NULL, &writefd, NULL, &tv_timeout ) )
      {
        case  0: throw_write_exception(ETIMEDOUT);
        case -1: throw_write_exception(errno);
        default: break;
      }
    }

    write_result = ::write( m_fd, data, data.size() );

    if ( write_result == -1 ) throw_write_exception(errno);

    return write_result;
  }

  Data FileDescriptor::read_data( size_t size, Timeout timeout ) throw ( read_exception )
  {
    int read_result;

    // Is the file descriptor open?
    if ( this->is_closed() ) this->open();

    // We need to do some prep work based on whether the timeout indicates
    // a blocking read, a non-blocking immediate read, or a blocking with timeout
    if ( timeout.is_null() )
    {
      this->set_blocking(true);
    }
    else
    {
      this->set_blocking(false);
        
      fd_set readfd;
      struct timeval tv_timeout = timeout;
        
      FD_ZERO ( &readfd );
      FD_SET ( m_fd, &readfd );
      switch ( ::select ( m_fd + 1, &readfd, NULL, NULL, &tv_timeout ) )
      {
        case  0: throw_read_exception(ETIMEDOUT);
        case -1: throw_read_exception(errno);
        default: break;
      }
    }

    // We need to size the read
    // if read size is zero then we will read all available
    if ( size == 0 ) size = this->input_available();

    // if size is <= zero then we will use the default read size
    if ( size <= 0 ) size = this->fallback_read_size();

    // resize the data object to hold the entire read buffer
    Data data(size);

    // perform the actual read
    read_result = ::read ( m_fd, data, size );

    if ( read_result == -1 ) throw_read_exception(errno);

    if ( m_timestamp ) data.set_current_time();

    if ( data.size() != ( unsigned ) read_result ) data.resize ( read_result );

    return data;
  }

}

