/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_IPADDRESS_H
#define CONEXUS_IPADDRESS_H

#include <string>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <sigc++/sigc++.h>

#include <conexus/enums.h>
#include <conexus/address.h>
#include <conexus/except.h>

namespace Conexus
  {

    /**
     * IP network address interface.
     *
     * This class represents an IP address string based interface to both IPv4 and IPv6 addresses
     * which are further implemented in IPv4::Address and IPv6::Address.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     */
    class IPAddress: public Address
      {
      public:
        typedef ConexusPointer<IPAddress> pointer;

        IPAddress():Address() { }

        IPAddress(const struct sockaddr_storage& addr): Address(addr) { }

        virtual ~IPAddress() { }

        /** Returns the address as a string. */
        virtual std::string address_string() const = 0;

        /** Returns the netmask associated with this address as a string. */
        virtual std::string subnet_mask_string() const = 0;

        /** Returns the prefix length of the subnet mask associated with this address. */
        virtual unsigned prefix_length() const = 0;

        /** Returns the prefix length of the subnet mask associated with this address as a string. */
        virtual std::string prefix_length_string() const = 0;

        /** Returns the address as an address/subnetmask string. */
        virtual std::string address_subnet_mask_string() const = 0;

        /** Returns the address as a CIDR address/prefixlen string. */
        virtual std::string cidr_address_string() const = 0;

        /** Returns the prefix part of this address as a string. */
        virtual std::string prefix_string() const = 0;

        /** Returns the local part of this address as a string. */
        virtual std::string local_address_string() const = 0;

        /** Returns the port number associated with this address. */
        virtual uint16_t port() const = 0;

        /** Returns the port number associated with this address as a string. */
        virtual std::string port_string() const = 0;

        /** Sets the address to the specified value. */
        virtual void set_address(const std::string& address) = 0;

        /** Sets the address and port to the specified value. */
        virtual void set_address(const std::string& address, uint16_t port) = 0;

        /** Sets the prefix length to the specified value. */
        virtual void set_prefix_length(unsigned prefix_length) = 0;

        /** Sets the prefix length to the specified value. */
        virtual void set_prefix_length(const std::string& prefix_length) = 0;

        /** Sets the port portion of this address. */
        virtual void set_port(uint16_t port) = 0;

        /** Sets the port to the specified value. */
        virtual void set_port(const std::string& port) = 0;

        /** Performs a DNS query on the currently set address and returns the hostname string. */
        virtual std::string hostname() const = 0;

        /** Performs a service name lookup on the currently set port and returns the servicename string */
        virtual std::string servicename() const = 0;

        /** True if this is a multicast address */
        virtual bool is_multicast() const = 0;

        /** True if this is the any address */
        virtual bool is_any() const = 0;

        /** True if this is the loopback address */
        virtual bool is_loopback() const = 0;

        virtual sigc::signal<void> signal_changed() = 0;
        virtual sigc::signal<void> signal_address_changed( ) = 0;
        virtual sigc::signal<void> signal_port_changed( ) = 0;
        virtual sigc::signal<void> signal_prefix_changed( ) = 0;

        virtual bool is_ipv4() = 0;

        virtual bool is_ipv6() = 0;

      };

}

#endif
