/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ipv4_address.h"
#include "ipv6_address.h"

#include <iostream>

#include <sstream>
#include <cstdlib>
#include <netdb.h>

namespace Conexus
{

  namespace IPv4
  {

    Address::Address( uint32_t address, uint16_t port ) :
        m_address( address ),
        m_port( port ),
        m_prefix_length( 32 )
    {
      m_sockaddr.ipv4.sin_family = AF_INET;
      m_sockaddr.ipv4.sin_addr.s_addr = htonl( address );
      m_sockaddr.ipv4.sin_port = htons( port );
    }

    Address::Address( const std::string & host, uint16_t port ) :
        m_address( INADDR_ANY ),
        m_port( port ),
        m_prefix_length( 32 )
    {
      m_sockaddr.ipv4.sin_family = AF_INET;
      this->set_address( host );
      m_sockaddr.ipv4.sin_port = htons( port );
    }

    Address::Address( const struct sockaddr_in & addr ) :
        m_address( INADDR_ANY ),
        m_port( 0 ),
        m_prefix_length( 32 )
    {
      m_sockaddr.ipv4 = addr;
      m_address = htonl( m_sockaddr.ipv4.sin_addr.s_addr );
      m_port = htons( m_sockaddr.ipv4.sin_port );
    }

    Address::Address( const struct sockaddr_storage & addr ) :
        IPAddress( addr ),
        m_address( INADDR_ANY ),
        m_port( 0 ),
        m_prefix_length( 32 )
    {
      m_sockaddr.storage = addr;
      m_address = htonl( m_sockaddr.ipv4.sin_addr.s_addr );
      m_port = htons( m_sockaddr.ipv4.sin_port );
    }

    Address::Address( const struct sockaddr & addr ) :
        m_address( INADDR_ANY ),
        m_port( 0 ),
        m_prefix_length( 32 )
    {
      m_address = htonl( m_sockaddr.ipv4.sin_addr.s_addr );
      m_port = htons( m_sockaddr.ipv4.sin_port );
    }

    Address::Address( const IPv6::Address& ipv6addr ) :
        m_address( INADDR_ANY ),
        m_port( 0 ),
        m_prefix_length( 32 )
    {
      uint32_t ipv4addr;
      if ( ipv6addr.is_ipv4_compatible() ) {
        ipv4addr = ipv6addr.address().s6_addr32[3];
        this->set_address( ipv4addr );
        m_sockaddr.ipv4.sin_port = htons( ipv6addr.port() );
      } else
        throw( exception::address::ipv6_not_ipv4_compatible() );
    }

    Address::pointer Address::create( uint32_t address, uint16_t port )
    {
      return Address::pointer( new Address( address, port ) );
    }

    Address::pointer Address::create( const std::string & host, uint16_t port )
    {
      return Address::pointer( new Address( host, port ) );
    }

    Address::pointer Address::create( const struct sockaddr_in& addr )
    {
      return Address::pointer( new Address( addr ) );
    }

    Address::pointer Address::create( const struct sockaddr_storage& addr )
    {
      return Address::pointer( new Address( addr ) );
    }

    Address::pointer Address::create( const struct sockaddr& addr )
    {
      return Address::pointer( new Address( addr ) );
    }

    Address::pointer Address::create( const IPv6::Address& ipv6addr )
    {
      return Address::pointer( new Address( ipv6addr ) );
    }

    Address::~Address()
    {}

    uint32_t Address::address() const
    {
      return m_address;
    }

    std::string Address::address_string() const
    {
      return address_to_string( m_address );
    }

    uint32_t Address::subnet_mask() const
    {
      return create_subnet_mask( m_prefix_length );
    }

    std::string Address::subnet_mask_string() const
    {
      return address_to_string( this->subnet_mask() );
    }

    unsigned Address::prefix_length() const
    {
      return m_prefix_length;
    }

    std::string Address::prefix_length_string() const
    {
      std::ostringstream sout;
      sout << this->prefix_length();
      return sout.str();
    }

    std::string Address::address_subnet_mask_string() const
    {
      return this->address_string() + std::string( "/" ) + this->subnet_mask_string();
    }

    std::string Address::cidr_address_string() const
    {
      return this->address_string() + std::string( "/" ) + this->prefix_length_string();
    }

    uint32_t Address::prefix() const
    {
      return ( this->address() ) & ( this->subnet_mask() );
    }

    std::string Address::prefix_string() const
    {
      return address_to_string( this->prefix() );
    }

    uint32_t Address::broadcast_address() const
    {
      return ( ( this->address() ) & ( this->subnet_mask() ) ) | ~( this->subnet_mask() );
    }

    std::string Address::broadcast_address_string() const
    {
      return address_to_string( this->broadcast_address() );
    }

    uint32_t Address::local_address() const
    {
      return ( this->address() ) & ~( this->subnet_mask() );
    }

    std::string Address::local_address_string() const
    {
      return address_to_string( this->local_address() );
    }

    uint16_t Address::port() const
    {
      return m_port;
    }

    std::string Address::port_string() const
    {
      std::ostringstream sout;
      sout << m_port;
      return sout.str();
    }

    void Address::set_address( const std::string& s )
    {
      uint32_t addr;
      int prefix;

      string_to_address_prefix( s, addr, prefix );

      if ( prefix >= 0 )
        m_prefix_length = prefix;

      m_address = addr;
      m_sockaddr.ipv4.sin_addr.s_addr = htonl( m_address );

      m_signal_address_changed.emit();
      if ( prefix >= 0 )
        m_signal_prefix_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_address( const std::string& address, uint16_t port )
    {
      uint32_t addr;
      int prefix;

      string_to_address_prefix( address, addr, prefix );

      if ( prefix >= 0 )
        m_prefix_length = prefix;

      m_address = addr;
      m_sockaddr.ipv4.sin_addr.s_addr = htonl( m_address );

      m_port = port;
      m_sockaddr.ipv4.sin_port = htons( m_port );

      m_signal_address_changed.emit();
      m_signal_port_changed.emit();
      if ( prefix >= 0 )
        m_signal_prefix_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_address( uint32_t a )
    {
      m_address = a;
      m_sockaddr.ipv4.sin_addr.s_addr = htonl( m_address );

      m_signal_address_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_address_subnet_mask( uint32_t address, uint32_t subnet_mask )
    {
      m_address = address;
      m_sockaddr.ipv4.sin_addr.s_addr = htonl( m_address );

      m_prefix_length = subnet_mask_length( subnet_mask );

      m_signal_address_changed.emit();
      m_signal_prefix_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_address_prefix_length( uint32_t address, unsigned prefix_length )
    {
      m_address = address;
      m_sockaddr.ipv4.sin_addr.s_addr = htonl( m_address );

      m_prefix_length = prefix_length;

      m_signal_address_changed.emit();
      m_signal_prefix_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_subnet_mask( uint32_t subnet_mask )
    {
      this->set_prefix_length( subnet_mask_length( subnet_mask ) );
    }

    void Address::set_subnet_mask( const std::string& subnet_mask )
    {
      uint32_t snm = string_to_address( subnet_mask );
      this->set_prefix_length( subnet_mask_length( snm ) );
    }

    void Address::set_prefix_length( unsigned prefix_length )
    {
      m_prefix_length = std::min(prefix_length, 32U);

      m_signal_prefix_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_prefix_length( const std::string& prefix_length )
    {
      std::istringstream sin( prefix_length );
      uint32_t pl;
      sin >> pl;
      m_prefix_length = pl;

      m_signal_prefix_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_port( uint16_t p )
    {
      m_port = p;
      m_sockaddr.ipv4.sin_port = htons( m_port );

      m_signal_port_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_port( const std::string& port )
    {
      unsigned i = 0;
      while ( i < port.size() && isspace( port[ i ] ) )
        i++;
      if ( isdigit( port[ i ] ) ) {
        std::istringstream sin( port );
        unsigned p;
        sin >> p;
        m_port = p;
        m_sockaddr.ipv4.sin_port = htons( m_port );
        m_signal_port_changed.emit();
        m_signal_changed.emit();
      } else
        m_port = servicename_to_service( port );
        m_sockaddr.ipv4.sin_port = htons( m_port );
        m_signal_port_changed.emit();
        m_signal_changed.emit();
    }

    bool Address::is_valid_hostname( const std::string& name ) const
    {
      try {
        hostname_to_address( name );
        return true;
      } catch ( address_exception ) {
        return false;
      }
    }

    std::string Address::hostname() const
    {
      return address_to_hostname( m_address );
    }

    std::string Address::servicename() const
    {
      return service_to_servicename( m_port );
    }

    struct sockaddr_in& Address::sockaddr_in()
    {
      return m_sockaddr.ipv4;
    }

    struct sockaddr_in* Address::sockaddr_in_ptr()
    {
      return &( m_sockaddr.ipv4 );
    }

    Address::operator struct sockaddr_in*()
    {
      return &( m_sockaddr.ipv4 );
    }

    socklen_t Address::sockaddr_size() const
    {
      return sizeof( struct sockaddr_in );
    }

    bool Address::is_any() const
    {
      return m_address == INADDR_ANY;
    }

    bool Address::is_private() const
    {
      if ( ( m_address & 0xFF000000 ) == 0x10000000 )
        return true;
      if ( ( m_address & 0xFFF00000 ) == 0xAC100000 )
        return true;
      if ( ( m_address & 0xFFFF0000 ) == 0xC0A80000 )
        return true;
      return false;
    }

    bool Address::is_reserved() const
    {
      if ( ( m_address & 0xFFFF0000 ) == 0x00000000 )
        return true;
      if ( ( m_address & 0xF8000000 ) == 0xF0000000 )
        return true;
      return false;
    }

    bool Address::is_loopback() const
    {
      if ( ( m_address & 0xFF000000 ) == ( IN_LOOPBACKNET << 24 ) )
        return true;
      return false;
    }

    bool Address::is_broadcast() const
    {
      if ( m_address == INADDR_BROADCAST )
        return true;
      if ( ( m_address & this->subnet_mask() ) == 0x00000000 )
        return true;
      return false;
    }

    bool Address::is_multicast() const
    {
      if ( IN_MULTICAST( m_address ) )
        return true;
      return false;
    }

    Address& Address::operator=( const Address& other )
    {
      this->m_sockaddr.storage = other.m_sockaddr.storage;
      m_prefix_length = other.m_prefix_length;

      m_address = other.m_address;
      m_sockaddr.ipv4.sin_addr.s_addr = htonl( m_address );

      m_port = other.m_port;
      m_sockaddr.ipv4.sin_port = htons( m_port );

      m_signal_address_changed.emit();
      m_signal_port_changed.emit();
      m_signal_prefix_changed.emit();
      m_signal_changed.emit();
      return *this;
    }

    Address & Address::operator =( const Conexus::Address & other )
    {
      const struct sockaddr_storage* sa = other.sockaddr_storage_ptr();
      const struct sockaddr_in* sain = (const struct sockaddr_in*) sa;

      if ( sa->ss_family != AF_INET )
        // probably should throw an exception
        return *this;

      m_address = sain->sin_addr.s_addr;
      m_sockaddr.ipv4.sin_addr.s_addr = m_address;
      m_address = htonl( m_address );

      m_port = sain->sin_port;
      m_sockaddr.ipv4.sin_port = m_port;
      m_port = htons( m_port );

      m_signal_address_changed.emit();
      m_signal_port_changed.emit();
      m_signal_changed.emit();
      return *this;
    }

    bool Address::operator==( const Address & other ) const
    {
      return ( m_port == other.m_port && m_address == other.m_address );
    }

    sigc::signal<void> Address::signal_changed( )
    {
      return m_signal_changed;
    }

    sigc::signal<void> Address::signal_address_changed( )
    {
      return m_signal_address_changed;
    }

    sigc::signal<void> Address::signal_port_changed( )
    {
      return m_signal_port_changed;
    }

    sigc::signal<void> Address::signal_prefix_changed( )
    {
      return m_signal_prefix_changed;
    }

    bool Address::is_ipv4()
    {
      return true;
    }

    bool Address::is_ipv6()
    {
      return false;
    }

    std::string address_to_string( uint32_t addr )
    {
      char buffer[ INET_ADDRSTRLEN ];
      struct in_addr a;
      a.s_addr = htonl( addr );
      if ( inet_ntop( AF_INET, &a, buffer, INET_ADDRSTRLEN ) == NULL )
        throw exception::address::bad_string();
      return std::string( buffer );
    }

    uint32_t string_to_address( const std::string& saddr )
    {
      struct in_addr a;
      if ( inet_pton( AF_INET, saddr.c_str(), &a ) <= 0 )
        throw exception::address::bad_string();
      return ntohl( a.s_addr );
    }

    void string_to_address_prefix( const std::string& s, uint32_t& address, int& prefix )
    {
      std::string addrstr, prefixstr;

      size_t split;

      prefix = -1;

      split = s.find( '/', 0 );
      if ( split == std::string::npos ) {
        addrstr = s;
      } else {
        addrstr = s.substr( 0, split );
        prefixstr = s.substr( split + 1, std::string::npos );
      }

      try {
        address = string_to_address( addrstr );
      } catch ( exception::address::bad_string ) {
        address = hostname_to_address( addrstr );
      }

      if ( prefixstr.size() > 0 ) {
        try {
          uint32_t snm = string_to_address( prefixstr );
          prefix = subnet_mask_length( snm );
        } catch ( exception::address::bad_string ) {
          std::istringstream sin( prefixstr );
          sin >> prefix;
        }
        if ( prefix > 32 )
          prefix = 32;
      }
    }

    bool is_valid_hostname( const std::string& name )
    {
      try {
        hostname_to_address( name );
        return true;
      } catch ( address_exception ) {
        return false;
      }
    }

    bool validate_subnet_mask( uint32_t subnet_mask ) throw ( )
    {
      static const uint32_t mask = 0x80000000;
      int count = 0;
      // left shift off leading ones until a zero is encountered
      while ( count++ < 32 && ( mask & subnet_mask ) )
        subnet_mask <<= 1;
      // if this is really a netmask the rest should be zeros
      return subnet_mask == 0x00000000;
    }

    std::string address_to_hostname( uint32_t address ) throw ( address_exception )
    {
      struct sockaddr_in sa;
      char ctmp[ NI_MAXHOST ];
      int result;
      sa.sin_family = AF_INET;
      sa.sin_port = 0x00;
      sa.sin_addr.s_addr = htonl( address );
      result = getnameinfo( ( const struct sockaddr* ) & sa, sizeof( struct sockaddr_in ), ctmp, NI_MAXHOST, NULL, 0x00, 0x00 );
      if ( result != 0 )
        throw_exception( result, EXCEPTION_ADDRESS );
      return std::string( ctmp );
    }

    std::string service_to_servicename( uint16_t service ) throw ( address_exception )
    {
      struct sockaddr_in sa;
      char ctmp[ NI_MAXSERV ];
      int result;
      sa.sin_family = AF_INET;
      sa.sin_port = htons( service );
      sa.sin_addr.s_addr = 0x00;
      result = getnameinfo( ( const struct sockaddr* ) & sa, sizeof( struct sockaddr_in ), NULL, 0x00, ctmp, NI_MAXSERV, 0x00 );
      if ( result != 0 )
        throw_exception( result, EXCEPTION_ADDRESS );
      return std::string( ctmp );
    }

    uint32_t hostname_to_address( const std::string& hostname ) throw ( address_exception )
    {
      struct addrinfo hints, *info;
      int result;
      uint32_t address;
      struct sockaddr_in* sa;
      hints.ai_flags = 0x00;
      hints.ai_family = PF_INET;
      hints.ai_socktype = 0x00;
      hints.ai_protocol = 0x00;
      hints.ai_addrlen = 0x00;
      hints.ai_addr = NULL;
      hints.ai_canonname = NULL;
      hints.ai_next = NULL;
      result = getaddrinfo( hostname.c_str(), NULL, &hints, &info );
      if ( result != 0 )
        throw_exception( result, EXCEPTION_ADDRESS );
      while ( info->ai_next != NULL && info->ai_family != PF_INET )
        info = info->ai_next;
      if ( info->ai_family != PF_INET )
        address = 0x00;
      else {
        sa = ( struct sockaddr_in* ) info->ai_addr;
        address = htonl( sa->sin_addr.s_addr );
      }
      freeaddrinfo( info );
      return address;
    }

    uint16_t servicename_to_service( const std::string& servicename ) throw ( address_exception )
    {
      struct addrinfo hints, *info;
      int result;
      uint16_t port;
      struct sockaddr_in* sa;
      hints.ai_flags = 0x00;
      hints.ai_family = PF_UNSPEC;
      hints.ai_socktype = 0x00;
      hints.ai_protocol = 0x00;
      hints.ai_addrlen = 0x00;
      hints.ai_addr = NULL;
      hints.ai_canonname = NULL;
      hints.ai_next = NULL;
      result = getaddrinfo( NULL, servicename.c_str(), &hints, &info );
      if ( result != 0 )
        throw_exception( result, EXCEPTION_ADDRESS );
      sa = ( struct sockaddr_in* ) info->ai_addr;
      port = htons( sa->sin_port );
      freeaddrinfo( info );
      return port;
    }

    uint32_t create_subnet_mask( unsigned prefix_length )
    {
      uint32_t subnet_mask = 0xFFFFFFFF;
      if ( prefix_length > 32 )
        prefix_length = 32;
      subnet_mask <<= 32 - prefix_length;
      return subnet_mask;
    }

    unsigned subnet_mask_length( uint32_t subnet_mask ) throw()
    {
      unsigned count = 0;
      while ( subnet_mask & 0x80000000 ) {
        count++;
        subnet_mask <<= 1;
      }
      return count;
    }

  }
}

