/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_IPV4ADDRESS_H
#define CONEXUS_IPV4ADDRESS_H

#include <string>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <conexus/ipaddress.h>
#include <conexus/except.h>

namespace Conexus
{

  namespace IPv6
  {
    class Address;
  }

  /**
   * @namespace Conexus::IPv4
   * The conexus::IPv4 namespace contains classes capable of communicating at
   * OSI Layers 3 (Network) and 4 (Transport) which includes support for
   * raw IP, TCP/IP and UDP/IP communications [ For more information see IP(7) ]
   */
  namespace IPv4
  {

    /**
     * IPv4 network address.
     *
     * This class represents an IPv4 address and includes support for accessing the address
     * as a traditional sockaddr_in (a BSD socket API IPv4 network address C structure).
     *
     * Host byte order and network byte order: all methods expect parameters and return values
     * in host byte order with the following exception; if you request a sockaddr_in representation
     * of this address, the sockaddr_in C structure will be filled with values that are already in
     * network byte order and ready for use with raw BSD API socket calls.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup IPv4
     */
    class Address: public IPAddress
    {
      public:

        /** Class scope smart pointer typedef */
        typedef ConexusPointer<Address> pointer;

        /**
         * Default constructor, or if parameters are used constructs an address from
         * a numeric address and an optional port number.
         *
         * Sets default family to AF_INET.
         *
         * @param address Numeric address in local host byte order.
         * @param port Port number in local host byte order.
         */
        Address( uint32_t address = INADDR_ANY, uint16_t port = 0 );

        Address( const std::string& host, uint16_t port = 0 );

        /** Constructs from an IPv4 sockaddr */
        Address( const struct sockaddr_in& addr );

        /** Constructs from a generic sockaddr_storage */
        Address( const struct sockaddr_storage& addr );

        /** Constructs from a generic sockaddr */
        Address( const struct sockaddr& addr );

        /** Constructs from an IPv6 address if the IPv6 address is IPv4 compatible */
        Address( const IPv6::Address& addr );

        /** Creates a smart pointer from an address and port number */
        static pointer create( uint32_t address = INADDR_ANY, uint16_t port = 0 );

        static pointer create( const std::string& host, uint16_t port );

        /** Creates a smart pointer from an IPv4 sockaddr_in */
        static pointer create( const struct sockaddr_in& addr );

        /** Creates a smart pointer from a generic sockaddr_storage */
        static pointer create( const struct sockaddr_storage& addr );

        /** Creates a smart pointer from a generic sockaddr */
        static pointer create( const struct sockaddr& addr );

        /** Creates a smart pointer from an IPv6 address if the IPv6 address is IPv4 compatible */
        static pointer create( const IPv6::Address& ipv6addr );

        /** Destructor */
        virtual ~Address();

        /** Returns the IPv4 address. */
        uint32_t address() const;

        /** Returns the IPv4 address as a string. */
        std::string address_string() const;

        /** Returns the subnet mask associated with this address. */
        uint32_t subnet_mask() const;

        /** Returns the netmask associated with this address as a string. */
        std::string subnet_mask_string() const;

        /** Returns the prefix length of the subnet mask associated with this address. */
        unsigned prefix_length() const;

        /** Returns the prefix length of the subnet mask associated with this address as a string. */
        std::string prefix_length_string() const;

        /** Returns the IPv4 address as an address/subnetmask string. */
        std::string address_subnet_mask_string() const;

        /** Returns the IPv4 address as a CIDR address/prefixlen string. */
        std::string cidr_address_string() const;

        /** Returns the prefix part of this address. */
        uint32_t prefix() const;

        /** Returns the prefix part of this address as a string. */
        std::string prefix_string() const;

        /** Returns the broadcast address for this address. */
        uint32_t broadcast_address() const;

        /** Returns the broadcast address for this address as a string. */
        std::string broadcast_address_string() const;

        /** Returns the local part of this address. */
        uint32_t local_address() const;

        /** Returns the local part of this address as a string. */
        std::string local_address_string() const;

        /** Returns the port number associated with this address. */
        uint16_t port() const;

        /** Returns the port number associated with this address as a string. */
        std::string port_string() const;

        /** Set the address to the specified value. */
        void set_address( uint32_t address );

        /**
         * Sets the address to the specified value.
         * If this is not a numeric specification, a DNS query will be performed.
         * Accepts addresses of the following forms:
         *   # aaa.aaa.aaa.aaa[/nnn.nnn.nnn.nnn | /bb][:ppppp | :servicename]
         *   # hostname[/nnn.nnn.nnn.nnn | /bb][:ppppp | :servicename]
         * Where:
         *   - aaa.aaa.aaa.aaa is an address in dotted decimal form
         *   - hostname is a host name
         *   - nnn.nnn.nnn.nnn is an optional subnet mask in dotted decimal form
         *   - bb is an optional prefix length
         *   - ppppp is an optional port specification
         *   - servicename is an optional service name specification
         */
        void set_address( const std::string& address );

        /** Sets the address and port to the specified value. */
        virtual void set_address( const std::string& address, uint16_t port );

        /** Sets the address to an absolute value with a specified subnet mask. */
        void set_address_subnet_mask( uint32_t address, uint32_t subnet_mask );

        /** Sets the address to an absolute value with a specified CIDR prefix length. */
        void set_address_prefix_length( uint32_t address, unsigned prefix_length );

        /** Sets the subnet mask to the specified value. */
        void set_subnet_mask( uint32_t subnet_mask );

        /** Sets the subnet mask to the specified value */
        void set_subnet_mask( const std::string& subnet_mask );

        /** Sets the prefix length to the specified value. */
        void set_prefix_length( unsigned prefix_length );

        /** Sets the prefix length to the specified value. */
        void set_prefix_length( const std::string& prefix_length );

        /** Sets the port portion of this address. */
        void set_port( uint16_t port );

        /** Sets the port to the specified value. */
        void set_port( const std::string& port );

        /**
         * Determine whether the provided string is a valid hostname by performing
         * a DNS query.
         */
        bool is_valid_hostname( const std::string& ) const;

        /** Performs a DNS query on the currently set address and returns the hostname string. */
        std::string hostname() const;

        /** Performs a service name lookup on the currently set port and returns the servicename string */
        std::string servicename() const;

        /**
         * True if this address specifies a private network.
         *
         * For reference, private networks are defined as:
         *  - 10.0.0.0        to   10.255.255.255  (10/8 prefix)
         *  - 172.16.0.0      to   172.31.255.255  (172.16/12 prefix)
         *  - 192.168.0.0     to   192.168.255.255 (192.168/16 prefix)
         */
        bool is_private() const;

        /** True if this address is a reserved address specified as either 0.0.0.0/16 or class E (11110). */
        bool is_reserved() const;

        /** True of this address is the loopback address with prefix 127.0.0.1/24. */
        bool is_loopback() const;

        /** True if this is a valid broadcast address. */
        bool is_broadcast() const;

        /** True if this is a valid multicast address. */
        bool is_multicast() const;

        /** True if this is the any address */
        bool is_any() const;

        /** Overrides the virtual parent method and provides the size in bytes of the underlying sockaddr_in structure. */
        socklen_t sockaddr_size() const;

        /** Provides a non-constant reference to the underlying sockaddr_in structure. */
        struct sockaddr_in& sockaddr_in();

        /** Provides a non-constant pointer to the underlying sockaddr_in structure. */
        struct sockaddr_in* sockaddr_in_ptr();

        operator struct sockaddr_in*();

        Address& operator=( const Address& other );

        Address& operator=( const Conexus::Address& other );

        /** True if the port and address are the same */
        bool operator==( const Address& other ) const;

        virtual sigc::signal<void> signal_changed( );

        virtual sigc::signal<void> signal_address_changed( );

        virtual sigc::signal<void> signal_port_changed( );

        virtual sigc::signal<void> signal_prefix_changed( );

        /** Always true for IPv4::Address */
        virtual bool is_ipv4();

        /** Never true */
        virtual bool is_ipv6();

      protected:
        uint32_t m_address;

        uint16_t m_port;

        /** Netmask prefix length */
        unsigned m_prefix_length;

        sigc::signal<void> m_signal_address_changed;

        sigc::signal<void> m_signal_port_changed;

        sigc::signal<void> m_signal_prefix_changed;

        sigc::signal<void> m_signal_changed;

    };

    /**
     * Convert a host byte order 32-bit address to a string
     * @relates Conexus::IPv4::Address
     */
    std::string address_to_string( uint32_t );

    /**
     * Convert a string to a host byte order 32-bit address
     * @relates Conexus::IPv4::Address
     */
    uint32_t string_to_address( const std::string& );

    /**
     * Convert a string to an address and a prefix length
     * @relates Conexus::IPv4::Address
     */
    void string_to_address_prefix( const std::string&, uint32_t& address, int& prefix );

    /**
     * Determine whether the provided string is a valid hostname by performing
     * a DNS query.
     * @relates Conexus::IPv4::Address
     */
    bool is_valid_hostname( const std::string& );

    /**
     * Convert a host byte order 32-bit address to a hostname string
     * @relates Conexus::IPv4::Address
     */
    std::string address_to_hostname( uint32_t address ) throw ( address_exception );

    /**
     * Convert a host byte order 16-bit port number to a service name string
     * @relates Conexus::IPv4::Address
     */
    std::string service_to_servicename( uint16_t service ) throw ( address_exception );

    /**
     * Convert a hostname string to a 32-bit host byte order address
     * @relates Conexus::IPv4::Address
     */
    uint32_t hostname_to_address( const std::string& hostname ) throw ( address_exception );

    /**
     * Convert a service name string to a 16-bit host byte order port number
     * @relates Conexus::IPv4::Address
     */
    uint16_t servicename_to_service( const std::string& servicename ) throw ( address_exception );

    /**
     * Checks to make sure this is a valid subnet mask; i.e. all 1's followed by all 0's
     * @relates Conexus::IPv4::Address
     */
    bool validate_subnet_mask( uint32_t subnet_mask ) throw();

    /**
     * Given a prefix length, return a subnet mask.
     * @relates Conexus::IPv4::Address
     */
    uint32_t create_subnet_mask( unsigned prefix_length );

    /**
     * Given a CIDR subnet mask, return the number of prefix bits
     * @relates Conexus::IPv4::Address
     */
    unsigned subnet_mask_length( uint32_t subnet_mask ) throw();

  }

}

#endif
