/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ipv4_tcp.h"

#include <iostream>

#include <sys/ioctl.h>
#include <linux/sockios.h>

namespace Conexus
{
  namespace IPv4
  {

    TCP::TCP ( uint32_t address, uint16_t port ) :
        IP ( SOCK_STREAM, IPPROTO_TCP )
    {
      m_remote_address.set_address ( address );
      m_remote_address.set_port ( port );
    }

    TCP::TCP ( const std::string& address, uint16_t port ) :
        IP ( SOCK_STREAM, IPPROTO_TCP )
    {
      m_remote_address.set_address ( address );
      m_remote_address.set_port ( port );
    }

    TCP::TCP ( int accepted_fd, bool ) :
        IP ( SOCK_STREAM, IPPROTO_TCP )
    {
      m_fd = accepted_fd;
      this->set_state_opened();
      m_state |= SOCKET_BOUND | SOCKET_CONNECTED | SOCKET_ACCEPTED;
    }

    TCP::pointer TCP::create ( uint16_t port )
    {
      return pointer ( new TCP ( INADDR_ANY, port ) );
    }

    TCP::pointer TCP::create ( uint32_t address, uint16_t port )
    {
      return pointer ( new TCP ( address, port ) );
    }

    TCP::pointer TCP::create ( const std::string& address, uint16_t port )
    {
      return pointer ( new TCP ( address, port ) );
    }

    TCP::pointer TCP::create ( int accepted_fd, bool b )
    {
      return pointer ( new TCP ( accepted_fd, b ) );
    }

    TCP::~TCP() throw()
    {}

    bool TCP::shutdown ( ReadWrite read_write )
    {
      int result;

      /* The socket must be connected before we try a shutdown */
      if ( ! this->is_connected() ) return false;

      switch ( read_write )
      {
        case READ:
          result = ::shutdown ( m_fd, SHUT_RD );
          break;
        case WRITE:
          result = ::shutdown ( m_fd, SHUT_WR );
          break;
        case READ_WRITE:
          result = ::shutdown ( m_fd, SHUT_RDWR );
          break;
        default:
          return false;
      }

      return result==0;
    }

    size_t TCP::input_available() throw ()
    {
      int n;
      ioctl ( m_fd, SIOCINQ, &n );
      if ( n >= 0 ) return n;
      return 0;
    }

    int TCP::pending_output_size()
    {
      int rsize;
      int error;
      error = ioctl ( m_fd, SIOCOUTQ, &rsize );
      return rsize;
    }

    void TCP::on_local_address_changed( )
    {
      // if this is an accepted socket, there is nothing to do when the local address is changed
      if ( is_accepted() )
        return ;

      // but, if this is a client side socket, and the socket is already bound, close and reopen to
      // same state with new local port and address
      if ( is_bound() )
        close_and_reopen();
    }

    void TCP::on_remote_address_changed( )
    {
      // if this is an accepted socket, there is nothing to do when the remote address is changed
      if ( is_accepted() )
        return ;

      // but, if this is a client side socket, and the socket is already connected, close and reopen to
      // same state with new local port and address
      if ( is_connected() )
        close_and_reopen();
    }

    size_t TCP::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
    {
      if ( ! is_accepted() && ! is_connected() )
        try
        {
          change_state ( SOCKET_CONNECTED );
        }
        catch ( exception::state::failed )
        {
          throw exception::write::not_connected();
        }

      return Socket::write_data ( data, timeout );
    }

  }
}

