/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ipv4_udp.h"

#include <iostream>
#include <sys/ioctl.h>
#include <linux/sockios.h>
// #include <linux/udp.h>

#include <conexus/utility.h>

#define DEFAULT_IPV4_MULTICAST_HOPS       1
#define DEFAULT_IPV4_MULTICAST_LOOP_BACK  false

namespace Conexus
{
  namespace IPv4
  {

    unsigned int UDP::m_default_hops = DEFAULT_IPV4_MULTICAST_HOPS;
    bool UDP::m_default_loopback = DEFAULT_IPV4_MULTICAST_LOOP_BACK;

    UDP::UDP ( uint32_t local_if, uint16_t localport, uint16_t remoteport ) :
        IP ( SOCK_DGRAM, IPPROTO_UDP ),
        m_write_without_connect ( true ),
        m_suppress_multicast_join ( false ),
        m_multicast_hops ( m_default_hops ),
        m_multicast_loopback ( m_default_loopback )
    {
      m_local_address.set_address ( local_if );
      m_local_address.set_port ( localport );
      m_remote_address.set_port ( remoteport );
    }

    UDP::UDP ( const std::string & local_if, uint16_t localport, uint16_t remoteport ) :
        IP ( SOCK_DGRAM, IPPROTO_UDP ),
        m_write_without_connect ( true ),
        m_suppress_multicast_join ( false ),
        m_multicast_hops ( m_default_hops ),
        m_multicast_loopback ( m_default_loopback )
    {
      m_local_address.set_address ( local_if );
      m_local_address.set_port ( localport );
      m_remote_address.set_port ( remoteport );
    }

    UDP::pointer UDP::create ( uint16_t localport, uint16_t remoteport )
    {
      return pointer ( new UDP ( INADDR_ANY, localport, remoteport ) );
    }

    UDP::pointer UDP::create ( uint32_t local_if, uint16_t localport, uint16_t remoteport )
    {
      return pointer ( new UDP ( local_if, localport, remoteport ) );
    }

    UDP::pointer UDP::create ( const std::string & local_if, uint16_t localport, uint16_t remoteport )
    {
      return pointer ( new UDP ( local_if, localport, remoteport ) );
    }

    UDP::~UDP() throw()
    {
      this->leave();
    }

    size_t UDP::input_available() throw()
    {
      int n;
      ioctl ( m_fd, FIONREAD, &n );
      if ( n >= 0 ) return n;
      return 0;
    }

    void UDP::bind() throw ( bind_exception )
    {
      int i;
      IP::bind();

      if ( ! m_suppress_multicast_join && m_local_address.is_multicast() )
        this->join();

      i = m_multicast_loopback;
      this->set_option ( SOL_IP, IP_MULTICAST_LOOP, i );
      this->set_option ( SOL_IP, IP_MULTICAST_TTL, m_multicast_hops );

      if ( m_local_address.is_broadcast() )
        this->set_option ( SO_BROADCAST, true );
      else
        this->set_option ( SO_BROADCAST, false );
    }

    void UDP::bind ( const Conexus::Address& a ) throw ( bind_exception )
    {
      int i;

      IP::bind ( a );
      if ( ! m_suppress_multicast_join && m_local_address.is_multicast() )
        this->join();

      i = m_multicast_loopback;
      this->set_option ( SOL_IP, IP_MULTICAST_LOOP, i );
      this->set_option ( SOL_IP, IP_MULTICAST_TTL, m_multicast_hops );
    }

    void UDP::close() throw ( close_exception )
    {
      this->leave();
      IP::close();
    }

    int UDP::output_pending()
    {
      int rsize;
      int error;
      error = ioctl ( m_fd, TIOCOUTQ, &rsize );
      return rsize;
    }

    void UDP::connect( ) throw ( connect_exception )
    {
      IP::connect ( m_remote_address );
      if ( ! m_suppress_multicast_join && m_remote_address.is_multicast() )
        this->join();
    }

    void UDP::connect ( const Address& a ) throw ( connect_exception )
    {
      IP::connect ( a );
      if ( ! m_suppress_multicast_join && m_remote_address.is_multicast() )
        this->join();
    }

    bool UDP::suppress_multicast_join() const
    {
      return m_suppress_multicast_join;
    }

    void UDP::set_suppress_multicast_join ( bool value )
    {
      m_suppress_multicast_join = value;
    }

    void UDP::set_write_without_connect ( bool value )
    {
      m_write_without_connect = value;
    }

    bool UDP::write_without_connect() const
    {
      return m_write_without_connect;
    }

    void UDP::add_multicast_interface ( unsigned iface_index )
    {
      std::set<unsigned>::iterator iter;

      iter = m_multicast_interfaces.find ( iface_index );

      if ( iter != m_multicast_interfaces.end() )
        return;

      m_multicast_interfaces.insert ( iface_index );

      if ( this->is_bound() )
        this->join ( iface_index );
    }

    void UDP::add_multicast_interface ( const std::string & iface_name )
    {
      this->add_multicast_interface ( interface_index ( iface_name ) );
    }

    void UDP::remove_multicast_interface ( unsigned iface_index )
    {
      std::set<unsigned>::iterator iter;

      iter = m_multicast_interfaces.find ( iface_index );

      if ( iter == m_multicast_interfaces.end() )
        return;

      m_multicast_interfaces.erase ( iter );
      if ( this->is_bound() )
        this->leave ( iface_index );
    }

    void UDP::remove_multicast_interface ( const std::string & iface_name )
    {
      this->remove_multicast_interface ( interface_index ( iface_name ) );
    }

    bool UDP::is_multicast_interface ( unsigned iface_index )
    {
      return ( m_multicast_interfaces.find ( iface_index ) != m_multicast_interfaces.end() );
    }

    bool UDP::is_multicast_interface ( const std::string& iface_name )
    {
      return this->is_multicast_interface ( interface_index ( iface_name ) );
    }

    unsigned UDP::multicast_hops()
    {
      int hops;

      if ( ! this->is_bound() )
        return m_multicast_hops;

      this->option ( SOL_IP, IP_MULTICAST_TTL, hops );

      return ( unsigned ) hops;
    }

    void UDP::set_multicast_hops ( unsigned hops )
    {
      m_multicast_hops = hops;
      if ( this->is_bound() )
        this->set_option ( SOL_IP, IP_MULTICAST_TTL, hops );
    }

    unsigned int UDP::default_multicast_hops( )
    {
      return m_default_hops;
    }

    void UDP::set_default_multicast_hops ( unsigned int hops )
    {
      m_default_hops = hops;
    }

    bool UDP::multicast_loopback()
    {
      if ( ! this->is_bound() )
        return m_multicast_loopback;

      int i;
      this->option ( SOL_IP, IP_MULTICAST_LOOP, i );
      return i;
    }

    void UDP::set_multicast_loopback ( bool loopback )
    {
      int i = loopback ? 1 : 0;

      m_multicast_loopback = loopback;

      if ( this->is_bound() )
        this->set_option ( SOL_IP, IP_MULTICAST_LOOP, i );
    }

    bool UDP::default_multicast_loopback( )
    {
      return m_default_loopback;
    }

    void UDP::set_default_multicast_loopback ( bool loopback )
    {
      m_default_loopback = loopback;
    }

    //     void UDP::cork( )
    //     {
    //       int i = 1;
    //       ::setsockopt( m_fd, IPPROTO_UDP, UDP_CORK, &i, sizeof(int) );
    //     }
    //
    //     void UDP::uncork( )
    //     {
    //       int i = 0;
    //       ::setsockopt( m_fd, IPPROTO_UDP, UDP_CORK, &i, sizeof(int) );
    //     }
    //
    //     bool UDP::is_corked()
    //     {
    //       int i;
    //       socklen_t len = sizeof(int);
    //       ::getsockopt( m_fd, IPPROTO_UDP, UDP_CORK, &i, &len );
    //       return i;
    //     }

    void UDP::on_local_address_changed( )
    {
      // if the socket is already bound, close and reopen to same state with new local port and address
      if ( this->is_bound() )
        this->close_and_reopen();
    }

    void UDP::on_remote_address_changed( )
    {
      // if the socket is already connected, close and reopen to same state with new remote port
      if ( this->is_connected() )
        this->close_and_reopen();
    }

    void UDP::join()
    {
      std::set<unsigned>::iterator iter;

      if ( ! m_local_address.is_multicast() )
        return;

      if ( ! this->is_bound() )
        return;

      if ( m_multicast_interfaces.size() == 0 )
        m_multicast_interfaces.insert ( 0 );

      for ( iter = m_multicast_interfaces.begin(); iter != m_multicast_interfaces.end(); iter++ )
        this->join ( *iter );
    }

    void UDP::join ( unsigned iface_index )
    {
      struct ip_mreqn multicast_request;

      if ( ! m_local_address.is_multicast() )
        return;

      if ( ! this->is_bound() )
        return;

      multicast_request.imr_multiaddr = m_local_address.sockaddr_in().sin_addr;
      multicast_request.imr_address.s_addr = INADDR_ANY;
      multicast_request.imr_ifindex = iface_index;

      int result = ::setsockopt ( m_fd, IPPROTO_IP, IP_ADD_MEMBERSHIP, &multicast_request, sizeof ( struct ip_mreqn ) );

      if ( result == -1 ) throw_socket_exception ( errno );
    }

    void UDP::leave()
    {
      std::set<unsigned>::iterator iter;

      if ( ! m_local_address.is_multicast() )
        return;

      if ( ! this->is_bound() )
        return;

      for ( iter = m_multicast_interfaces.begin(); iter != m_multicast_interfaces.end(); iter++ )
        this->leave ( *iter );

      m_multicast_interfaces.clear();
    }

    void UDP::leave ( unsigned iface_index )
    {
      struct ip_mreqn multicast_request;

      if ( ! m_local_address.is_multicast() )
        return;

      if ( ! this->is_bound() )
        return;

      multicast_request.imr_multiaddr = m_local_address.sockaddr_in().sin_addr;
      multicast_request.imr_address.s_addr = 0;
      multicast_request.imr_ifindex = iface_index;
      this->set_option ( SOL_IP, IP_DROP_MEMBERSHIP, multicast_request );
    }

    size_t UDP::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
    {
      if ( m_write_without_connect )
      {
        if ( !m_remote_address_set )
          throw exception::write::no_default_remote_address();

        if ( is_closed() )
          try
        {
          change_state ( ENDPOINT_OPENED );
        }
        catch ( exception::state::failed )
        {
          throw exception::write::not_opened();
        }

        return Socket::writeto ( m_remote_address, data, timeout );
      }
      else
      {
        if ( ! is_connected() )
          try
        {
          change_state ( SOCKET_CONNECTED );
        }
        catch ( exception::state::failed )
        {
          throw exception::write::not_connected();
        }

        return Socket::write_data ( data, timeout );
      }
    }

  }
}

