/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ipv6_address.h"
#include "ipv4_address.h"

#include <iostream>

#include <sstream>
#include <cstdlib>
#include <netdb.h>

namespace Conexus
{

  namespace IPv6
  {

    struct in6_addr operator&( const struct in6_addr& first, const struct in6_addr& second )
    {
      struct in6_addr tmp = first;
      tmp &= second;
      return tmp;
    }

    struct in6_addr operator|( const struct in6_addr& first, const struct in6_addr& second )
    {
      struct in6_addr tmp = first;
      tmp |= second;
      return tmp;
    }

    struct in6_addr operator^( const struct in6_addr& first, const struct in6_addr& second )
    {
      struct in6_addr tmp = first;
      tmp ^= second;
      return tmp;
    }

    struct in6_addr operator~( const struct in6_addr& addr )
    {
      struct in6_addr tmp;
      for ( int i = 0; i < 4; i++ )
        tmp.s6_addr32[ i ] = ~addr.s6_addr32[ i ];
      return tmp;
    }

    struct in6_addr& operator&=( struct in6_addr& first, const struct in6_addr& second )
    {
      for ( int i = 0; i < 4; i++ )
        first.s6_addr32[ i ] &= second.s6_addr32[ i ];
      return first;
    }

    struct in6_addr& operator|=( struct in6_addr& first, const struct in6_addr& second )
    {
      for ( int i = 0; i < 4; i++ )
        first.s6_addr32[ i ] |= second.s6_addr32[ i ];
      return first;
    }

    struct in6_addr& operator^=( struct in6_addr& first, const struct in6_addr& second )
    {
      for ( int i = 0; i < 4; i++ )
        first.s6_addr32[ i ] ^= second.s6_addr32[ i ];
      return first;
    }

    bool operator==( const struct in6_addr& first, const struct in6_addr& second )
    {
      return ( IN6_ARE_ADDR_EQUAL( first.s6_addr32, second.s6_addr32 ) );
    }

    bool operator<( const struct in6_addr& first, const struct in6_addr& second )
    {
      for ( int i=0; i < 4; i++ ) {
        if ( first.s6_addr32[i] < second.s6_addr32[i] )
          return true;
        if ( first.s6_addr32[i] > second.s6_addr32[i] )
          return false;
      }
      return false;
    }

    bool operator!=( const struct in6_addr& first, const struct in6_addr& second )
    {
      return ! ( IN6_ARE_ADDR_EQUAL( first.s6_addr32, second.s6_addr32 ) );
    }

    const struct in6_addr Address::LOOPBACK =
      {
        { {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01
          }
        }
      };
    const struct in6_addr Address::IPV4_COMPATIBLE_PREFIX =
      {
        { {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
          }
        }
      };
    const struct in6_addr Address::IPV4_MAPPED_PREFIX =
      {
        { {
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00
          }
        }
      };

    Address::Address( const struct in6_addr& address, uint16_t port ) :
        m_address( address ),
        m_port( port ),
        m_flowinfo( 0 ),
        m_scope_id( 0 ),
        m_prefix_length( 128U )
    {
      m_sockaddr.ipv6.sin6_family = AF_INET6;
      m_sockaddr.ipv6.sin6_addr = m_address;
      m_sockaddr.ipv6.sin6_port = htons( m_port );
    }

    Address::Address( uint32_t host, uint16_t port ) :
        m_address( in6addr_any ),
        m_port( port ),
        m_flowinfo( 0 ),
        m_scope_id( 0 ),
        m_prefix_length( 128U )
    {
      m_sockaddr.ipv6.sin6_family = AF_INET6;
      this->set_address(host);
      m_sockaddr.ipv6.sin6_port = htons( m_port );
    }

    Address::Address( const std::string & host, uint16_t port ) :
        m_address( in6addr_any ),
        m_port( port ),
        m_flowinfo( 0 ),
        m_scope_id( 0 ),
        m_prefix_length( 128U )
    {
      m_sockaddr.ipv6.sin6_family = AF_INET6;
      this->set_address(host);
      m_sockaddr.ipv6.sin6_port = htons( m_port );
    }

    Address::Address( const struct sockaddr_in6 & addr ) :
        m_address( in6addr_any ),
        m_port( 0 ),
        m_flowinfo( 0 ),
        m_scope_id( 0 ),
        m_prefix_length( 128U )
    {
      m_sockaddr.ipv6 = addr;
      m_address = m_sockaddr.ipv6.sin6_addr;
      m_port = htons( m_sockaddr.ipv6.sin6_port );
      m_flowinfo = htonl( m_sockaddr.ipv6.sin6_flowinfo );
      m_scope_id = htonl( m_sockaddr.ipv6.sin6_scope_id );
    }

    Address::Address( const struct sockaddr_storage & addr ) :
        m_address( in6addr_any ),
        m_port( 0 ),
        m_flowinfo( 0 ),
        m_scope_id( 0 ),
        m_prefix_length( 128U )
    {
      m_sockaddr.storage = addr;
      m_address = m_sockaddr.ipv6.sin6_addr;
      m_port = htons( m_sockaddr.ipv6.sin6_port );
      m_flowinfo = htonl( m_sockaddr.ipv6.sin6_flowinfo );
      m_scope_id = htonl( m_sockaddr.ipv6.sin6_scope_id );
    }

    Address::Address( const IPv4::Address& ipv4addr, IPV4_EMBEDDED_TYPE ipv4type ):
        m_address( in6addr_any ),
        m_port( 0 ),
        m_flowinfo( 0 ),
        m_scope_id( 0 ),
        m_prefix_length( 128U )
    {
      this->set_address( ipv4addr.address(), ipv4type );
      this->set_port( ipv4addr.port() );
    }

    Address::pointer Address::create( const struct in6_addr& address, uint16_t port )
    {
      return Address::pointer( new Address( address, port ) );
    }

    Address::pointer Address::create( uint32_t host, uint16_t port )
    {
      return Address::pointer( new Address( host, port ) );
    }

    Address::pointer Address::create( const std::string & host, uint16_t port )
    {
      return Address::pointer( new Address( host, port ) );
    }

    Address::pointer Address::create( const struct sockaddr_in6& addr )
    {
      return Address::pointer( new Address( addr ) );
    }

    Address::pointer Address::create( const struct sockaddr_storage& addr )
    {
      return Address::pointer( new Address( addr ) );
    }

    Address::pointer Address::create(const IPv4::Address& ipv4addr, IPV4_EMBEDDED_TYPE ipv4type)
    {
      return Address::pointer( new Address( ipv4addr, ipv4type ) );
    }

    Address::~Address()
    {}

    struct in6_addr Address::address() const
      {
        return m_sockaddr.ipv6.sin6_addr;
      }

    std::string Address::address_string() const
    {
      return address_to_string( this->address() );
    }

    struct in6_addr Address::subnet_mask() const
      {
        return create_subnet_mask( m_prefix_length );
      }

    std::string Address::subnet_mask_string() const
    {
      return address_to_string( create_subnet_mask( m_prefix_length ) );
    }

    unsigned Address::prefix_length() const
    {
      return m_prefix_length;
    }

    std::string Address::prefix_length_string() const
    {
      std::ostringstream sout;
      sout << m_prefix_length;
      return sout.str();
    }

    std::string Address::address_subnet_mask_string() const
    {
      return this->address_string() + std::string( "/" ) + this->subnet_mask_string();
    }

    std::string Address::cidr_address_string() const
    {
      return this->address_string() + std::string( "/" ) + this->prefix_length_string();
    }

    struct in6_addr Address::prefix() const
      {
        return this->address() & this->subnet_mask();
      }

    std::string Address::prefix_string() const
    {
      return address_to_string( this->prefix() );
    }

    struct in6_addr Address::local_address() const
      {
        return this->address() & ~( this->subnet_mask() );
      }

    std::string Address::local_address_string() const
    {
      return address_to_string( this->local_address() );
    }

    uint16_t Address::port() const
    {
      return m_port;
    }

    std::string Address::port_string() const
    {
      std::ostringstream sout;
      sout << this->port();
      return sout.str();
    }

    void Address::set_address( const std::string& s )
    {
      struct in6_addr addr;
      int prefix;

      string_to_address_prefix( s, addr, prefix );

      if ( prefix >= 0 )
        this->set_prefix_length( prefix );
      this->set_address( addr );
    }

    void Address::set_address( const std::string& address, uint16_t port )
    {
      struct in6_addr addr;
      int prefix;

      string_to_address_prefix( address, addr, prefix );

      if ( prefix >= 0 )
        this->set_prefix_length( prefix );
      m_address = addr;
      m_sockaddr.ipv6.sin6_addr = m_address;

      m_port = port;
      m_sockaddr.ipv6.sin6_port = htons( m_port );

      m_signal_address_changed.emit();
      m_signal_port_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_address( const struct in6_addr& a )
    {
      m_address = a;
      m_sockaddr.ipv6.sin6_addr = m_address;

      m_signal_address_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_address( uint32_t a4, IPV4_EMBEDDED_TYPE type )
    {
      in6_addr a6;
      if ( type == IPV4_MAPPED )
        a6 = IPV4_MAPPED_PREFIX;
      else
        a6 = IPV4_COMPATIBLE_PREFIX;
      a6.s6_addr32[ 3 ] = htonl( a4 );
      m_address = a6;
      m_sockaddr.ipv6.sin6_addr = m_address;

      m_signal_address_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_address_prefix_length( uint32_t address, unsigned prefix_length, IPV4_EMBEDDED_TYPE type )
    {
      m_prefix_length = prefix_length + 96;

      set_address( address, type );

      m_signal_prefix_changed.emit();
    }

    void Address::set_prefix_length( unsigned prefix_length )
    {
      m_prefix_length = std::min( prefix_length, 128U );
      m_signal_prefix_changed.emit();
    }

    void Address::set_prefix_length( const std::string& prefix_length )
    {
      std::istringstream sin( prefix_length );
      unsigned pl;
      sin >> pl;
      this->set_prefix_length( pl );
    }

    void Address::set_port( uint16_t p )
    {
      m_port = p;
      m_sockaddr.ipv6.sin6_port = htons( m_port );

      m_signal_port_changed.emit();
      m_signal_changed.emit();
    }

    void Address::set_port( const std::string& port )
    {
      unsigned i = 0;
      while ( i < port.size() && isspace( port[ i ] ) )
        i++;
      if ( isdigit( port[ i ] ) ) {
        std::istringstream sin( port );
        uint16_t p;
        sin >> p;
        this->set_port( p );
      } else
        this->set_port( servicename_to_service( port ) );
    }

    uint32_t Address::flowinfo()
    {
      return m_flowinfo;
    }

    void Address::set_flowinfo( uint32_t f )
    {
      m_flowinfo = f;
      m_sockaddr.ipv6.sin6_flowinfo = htonl( m_flowinfo );

      m_signal_flowinfo_changed.emit();
      m_signal_changed.emit();
    }

    uint32_t Address::scope_id()
    {
      return m_scope_id;
    }

    void Address::set_scope_id( uint32_t sid )
    {
      m_scope_id = sid;
      m_sockaddr.ipv6.sin6_scope_id = htonl( m_scope_id );

      m_signal_scope_id_changed.emit();
      m_signal_changed.emit();
    }

    std::string Address::hostname() const
    {
      return address_to_hostname( this->address() );
    }

    std::string Address::servicename() const
    {
      return service_to_servicename( this->port() );
    }

    bool Address::is_reserved() const
    {
      return ( m_sockaddr.ipv6.sin6_addr.s6_addr[ 0 ] == 0x00 );
    }

    bool Address::is_nsap() const
    {
      return ( ( m_sockaddr.ipv6.sin6_addr.s6_addr[ 0 ] & 0xFE ) == 0x02 );
    }

    bool Address::is_ipx() const
    {
      return ( ( m_sockaddr.ipv6.sin6_addr.s6_addr[ 0 ] & 0xFE ) == 0x04 );
    }

    bool Address::is_aggregatable_global() const
    {
      return ( ( m_sockaddr.ipv6.sin6_addr.s6_addr[ 0 ] & 0xE0 ) == 0x20 );
    }

    bool Address::is_link_local() const
    {
      return IN6_IS_ADDR_LINKLOCAL( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_site_local() const
    {
      return IN6_IS_ADDR_SITELOCAL( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_multicast() const
    {
      return IN6_IS_ADDR_MULTICAST( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_multicast_interface_local() const
    {
      return IN6_IS_ADDR_MC_NODELOCAL( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_multicast_link_local() const
    {
      return IN6_IS_ADDR_MC_LINKLOCAL( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_multicast_site_local() const
    {
      return IN6_IS_ADDR_MC_SITELOCAL( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_multicast_organization_local() const
    {
      return IN6_IS_ADDR_MC_ORGLOCAL( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_multicast_global() const
    {
      return IN6_IS_ADDR_MC_GLOBAL( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_any() const
    {
      return ( m_sockaddr.ipv6.sin6_addr == in6addr_any );
    }

    bool Address::is_loopback() const
    {
      return IN6_IS_ADDR_LOOPBACK( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_ipv4() const
    {
      return ( is_ipv4_mapped() || is_ipv4_compatible() );
    }

    bool Address::is_ipv4_mapped() const
    {
      return IN6_IS_ADDR_V4MAPPED( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    bool Address::is_ipv4_compatible() const
    {
      return IN6_IS_ADDR_V4COMPAT( m_sockaddr.ipv6.sin6_addr.s6_addr32 );
    }

    struct sockaddr_in6& Address::sockaddr_in()
    {
      return m_sockaddr.ipv6;
    }

    struct sockaddr_in6* Address::sockaddr_in_ptr()
    {
      return &(m_sockaddr.ipv6);
    }

    Address::operator struct sockaddr_in6*()
    {
      return &(m_sockaddr.ipv6);
    }

    socklen_t Address::sockaddr_size() const
    {
      return sizeof( struct sockaddr_in6 );
    }

    Address& Address::operator=( const Address& other )
    {
      this->m_sockaddr.storage = other.m_sockaddr.storage;
      m_prefix_length = other.m_prefix_length;

      m_address = other.m_address;
      m_sockaddr.ipv6.sin6_addr = m_address;

      m_port = other.m_port;
      m_sockaddr.ipv6.sin6_port = htons( m_port );

      m_flowinfo = other.m_flowinfo;
      m_sockaddr.ipv6.sin6_flowinfo = htonl( m_flowinfo );

      m_scope_id = other.m_scope_id;
      m_sockaddr.ipv6.sin6_scope_id = htonl( m_scope_id );

      m_signal_prefix_changed.emit();
      m_signal_address_changed.emit();
      m_signal_port_changed.emit();
      m_signal_flowinfo_changed.emit();
      m_signal_scope_id_changed.emit();
      m_signal_changed.emit();
      return *this;
    }

    Address & Address::operator =( const Conexus::Address & other )
    {
      const struct sockaddr_storage* sa = other.sockaddr_storage_ptr();
      const struct sockaddr_in6* sain6 = (const struct sockaddr_in6*) sa;

      if ( sa->ss_family != AF_INET6 )
        // probably should throw an exception
        return *this;

      m_address = sain6->sin6_addr;
      m_sockaddr.ipv6.sin6_addr = m_address;

      m_port = sain6->sin6_port;
      m_sockaddr.ipv6.sin6_port = m_port;
      m_port = htons( m_port );

      m_flowinfo = sain6->sin6_flowinfo;
      m_sockaddr.ipv6.sin6_flowinfo = m_flowinfo;
      m_flowinfo = htonl( m_flowinfo );

      m_scope_id = sain6->sin6_scope_id;
      m_sockaddr.ipv6.sin6_scope_id = m_scope_id;
      m_scope_id = htonl( m_scope_id );

      m_signal_address_changed.emit();
      m_signal_port_changed.emit();
      m_signal_flowinfo_changed.emit();
      m_signal_scope_id_changed.emit();
      m_signal_changed.emit();
      return *this;
    }

    bool Address::operator==( const Address& other )
    {
      return (m_port == other.m_port && m_address == other.m_address && m_flowinfo == other.m_flowinfo && m_scope_id == other.m_scope_id );
    }

    sigc::signal<void> Address::signal_changed( )
    {
      return m_signal_changed;
    }

    sigc::signal<void> Address::signal_address_changed( )
    {
      return m_signal_address_changed;
    }

    sigc::signal<void> Address::signal_port_changed( )
    {
      return m_signal_port_changed;
    }

    sigc::signal<void> Address::signal_flowinfo_changed( )
    {
      return m_signal_flowinfo_changed;
    }

    sigc::signal<void> Address::signal_scope_id_changed( )
    {
      return m_signal_scope_id_changed;
    }

    sigc::signal<void> Address::signal_prefix_changed( )
    {
      return m_signal_prefix_changed;
    }

    std::string address_to_string( const struct in6_addr& addr )
    {
      char buffer[ INET6_ADDRSTRLEN ];
      if ( inet_ntop( AF_INET6, &addr, buffer, INET6_ADDRSTRLEN ) == NULL )
        throw exception::address::bad_string();
      return std::string( buffer );
    }

    struct in6_addr string_to_address( const std::string& saddr )
    {
      struct in6_addr addr;
      if ( inet_pton( AF_INET6, saddr.c_str(), &addr ) <= 0 )
        throw exception::address::bad_string();
      return addr;
    }

    void string_to_address_prefix( const std::string& s, struct in6_addr& address, int& prefix )
    {
      std::string addrstr, prefixstr;
      size_t split;

      prefix = -1;

      split = s.find( '/', 0 );
      if ( split == std::string::npos )
      {
        addrstr = s;
      } else
      {
        addrstr = s.substr( 0, split );
        prefixstr = s.substr( split + 1, std::string::npos );
      }

      try
      {
        address = string_to_address( addrstr );
      } catch ( exception::address::bad_string )
      {
        address = hostname_to_address( addrstr );
      }

      if ( prefixstr != std::string() )
      {
        std::istringstream sin( prefixstr );
        sin >> prefix;
        if ( IN6_IS_ADDR_V4MAPPED( address.s6_addr32 ) ||
             IN6_IS_ADDR_V4COMPAT( address.s6_addr32 )
           )
          prefix += 96;
        if ( prefix > 128 )
          prefix = 128;
      }
    }

    bool is_valid_hostname( const std::string& name )
    {
      try {
        hostname_to_address( name );
        return true;
      } catch ( address_exception ) {
        return false;
      }
    }

    std::string address_to_hostname( const struct in6_addr& address ) throw ( address_exception )
    {
      if ( IN6_IS_ADDR_V4MAPPED( address.s6_addr32 ) || IN6_IS_ADDR_V4COMPAT( address.s6_addr32 ) )
      {
        struct sockaddr_in sa;
        char ctmp[ NI_MAXHOST ];
        int result;
        sa.sin_family = AF_INET;
        sa.sin_port = 0x00;
        sa.sin_addr.s_addr = address.s6_addr32[ 3 ];
        result = getnameinfo( ( const struct sockaddr* ) & sa, sizeof( struct sockaddr_in ), ctmp, NI_MAXHOST, NULL, 0x00, 0x00 );
        if ( result != 0 )
          throw_exception( result, EXCEPTION_ADDRESS );
        return std::string( ctmp );
      } else
      {
        struct sockaddr_in6 sa;
        char ctmp[ NI_MAXHOST ];
        int result;
        std::string name;
        sa.sin6_family = AF_INET6;
        sa.sin6_port = 0x00;
        sa.sin6_addr = address;
        result = getnameinfo( ( const struct sockaddr* ) & sa, sizeof( struct sockaddr_in6 ), ctmp, NI_MAXHOST, NULL, 0x00, 0x00 );
        if ( result != 0 )
          throw_exception( result, EXCEPTION_ADDRESS );
        return std::string( ctmp );
      }
    }

    std::string service_to_servicename( uint16_t service ) throw ( address_exception )
    {
      struct sockaddr_in6 sa;
      char ctmp[ NI_MAXSERV ];
      int result;
      sa.sin6_family = AF_INET6;
      sa.sin6_port = htons( service );
      sa.sin6_addr = in6addr_any;
      result = getnameinfo( ( const struct sockaddr* ) & sa, sizeof( struct sockaddr_in6 ), NULL, 0x00, ctmp, NI_MAXSERV, 0x00 );
      if ( result != 0 )
        throw_exception( result, EXCEPTION_ADDRESS );
      return std::string( ctmp );
    }

    struct in6_addr hostname_to_address( const std::string& hostname ) throw ( address_exception )
    {
      struct addrinfo hints, *info, *info2;
      int result;
      struct in6_addr address;
      struct sockaddr_in* sa;
      struct sockaddr_in6* sa6;
      hints.ai_flags = 0x00;
      hints.ai_family = PF_UNSPEC;
      hints.ai_socktype = 0x00;
      hints.ai_protocol = 0x00;
      hints.ai_addrlen = 0x00;
      hints.ai_addr = NULL;
      hints.ai_canonname = NULL;
      hints.ai_next = NULL;
      result = getaddrinfo( hostname.c_str(), NULL, &hints, &info );
      if ( result != 0 )
        throw_exception( result, EXCEPTION_ADDRESS );
      info2 = info;
      // try and find the first IPv6 record
      while ( info2->ai_next != NULL && info2->ai_family != PF_INET6 )
        info2 = info2->ai_next;
      // but, if we haven't found an IPv6 record, let's try and find an IPv4 record
      if ( info2->ai_family == PF_INET6 )
      {
        sa6 = ( struct sockaddr_in6* ) info2->ai_addr;
        address = sa6->sin6_addr;
      } else
      {
        info2 = info;
        while ( info2->ai_next != NULL && info2->ai_family != PF_INET )
          info2 = info2->ai_next;
        if ( info2->ai_family == PF_INET ) {
          sa = ( struct sockaddr_in* ) info2->ai_addr;
          address = Address::IPV4_MAPPED_PREFIX;
          address.s6_addr32[ 3 ] = sa->sin_addr.s_addr;
        } else
          address = in6addr_any;
      }
      freeaddrinfo( info );
      return address;
    }

    uint16_t servicename_to_service( const std::string& servicename ) throw ( address_exception )
    {
      struct addrinfo hints, *info;
      int result;
      uint16_t port;
      struct sockaddr_in* sa;
      hints.ai_flags = 0x00;
      hints.ai_family = PF_UNSPEC;
      hints.ai_socktype = 0x00;
      hints.ai_protocol = 0x00;
      hints.ai_addrlen = 0x00;
      hints.ai_addr = NULL;
      hints.ai_canonname = NULL;
      hints.ai_next = NULL;
      result = getaddrinfo( NULL, servicename.c_str(), &hints, &info );
      if ( result != 0 )
        throw_exception( result, EXCEPTION_ADDRESS );
      sa = ( struct sockaddr_in* ) info->ai_addr;
      port = htons( sa->sin_port );
      freeaddrinfo( info );
      return port;
    }

    bool validate_subnet_mask( const struct in6_addr& subnet_mask ) throw ( )
    {
      int i, k;
      uint8_t tmp;
      // Skip through all the octets that are all 1's
      for ( i = 0; i < 16 && subnet_mask.s6_addr[ i ] == 0xFF; i++ )
        ;

      // If i is 16, then we have a mask that is all 1's
      if ( i == 16 )
        return true;

      // Get a copy of the last octet
      tmp = subnet_mask.s6_addr[ i ];

      // Shift off all the leading 1's
      for ( k = 0; k < 8 && ( tmp&0x80 ); k++ )
        tmp <<= 1;

      // Everything left better be a 0
      if ( tmp != 0x00 )
        return false;

      // And everything left in the rest of the octets better be 0 as well
      while ( ++i < 16 )
        if ( subnet_mask.s6_addr[ i ] != 0x00 )
          return false;

      // If we made it this far, we have a winner
      return true;
    }

    struct in6_addr create_subnet_mask( unsigned prefix_length )
    {
      struct in6_addr tmp = in6addr_any;
      int i;

      // Cap our prefix to the number of bits in an IPv6 address
      if ( prefix_length > 128 )
        prefix_length = 128;

      // Loop, incrementing the index and reducing prefix length by 8 on
      // each iteration
      for ( i = 0; prefix_length >= 8; i++, prefix_length -= 8 )
        tmp.s6_addr[ i ] = 0xFF;

      // Here we need to make a partial octet mask
      if ( prefix_length > 0 )
      {
        uint8_t mask = 0xFF;
        mask <<= 8 - prefix_length;
        tmp.s6_addr[ i ] = mask;
      }

      return tmp;
    }

  }

}
