/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_IPV6ADDRESS_H
#define CONEXUS_IPV6ADDRESS_H

#include <string>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <conexus/ipaddress.h>
#include <conexus/except.h>

namespace Conexus
  {

    namespace IPv4 {
      class Address;
    }

  /**
   * @namespace Conexus::IPv6
   * The conexus::IPv6 namespace contains classes capable of communicating at
   * OSI Layers 3 (Network) and 4 (Transport) which includes support for
   * raw IP, TCP/IP and UDP/IP communications [ For more information see IP(7) ]
   */
  namespace IPv6
    {

      /** @relates Conexus::IPv4::Address */
      typedef enum IPV4_EMBEDDED_TYPE { IPV4_COMPATIBLE, IPV4_MAPPED } IPV4_EMBEDDED_TYPE;

      /** @relates Conexus::IPv4::Address */
      struct in6_addr operator&( const struct in6_addr& first, const struct in6_addr& second );

      /** @relates Conexus::IPv4::Address */
      struct in6_addr operator|( const struct in6_addr& first, const struct in6_addr& second );

      /** @relates Conexus::IPv4::Address */
      struct in6_addr operator^( const struct in6_addr& first, const struct in6_addr& second );

      /** @relates Conexus::IPv4::Address */
      struct in6_addr operator~( const struct in6_addr& addr );

      /** @relates Conexus::IPv4::Address */
      struct in6_addr& operator&=( struct in6_addr& first, const struct in6_addr& second );

      /** @relates Conexus::IPv4::Address */
      struct in6_addr& operator|=( struct in6_addr& first, const struct in6_addr& second );

      /** @relates Conexus::IPv4::Address */
      struct in6_addr& operator^=( struct in6_addr& first, const struct in6_addr& second );

      /** @relates Conexus::IPv4::Address */
      bool operator==( const struct in6_addr&, const struct in6_addr& );

      /** @relates Conexus::IPv4::Address */
      bool operator<( const struct in6_addr&, const struct in6_addr& );

      /** @relates Conexus::IPv4::Address */
      bool operator!=( const struct in6_addr&, const struct in6_addr& );

    /**
     * IPv6 network address.
     *
     * This class represents an IPv6 address and includes support for accessing the address
     * as a traditional sockaddr_in (a socket API IPv6 network address C structure).
     *
     * Host byte order and network byte order: all methods expect parameters and return values
     * in host byte order with the following exception; if you request a sockaddr_in representation
     * of this address, the sockaddr_in C structure will be filled with values that are already in
     * network byte order and ready for use with raw BSD API socket calls.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup IPv6
     */
      class Address: public IPAddress
      {
      public:

        /** IPv6 loopback address */
        static const struct in6_addr LOOPBACK;

        /** IPv6 prefix for IPv4 compatible addresses */
        static const struct in6_addr IPV4_COMPATIBLE_PREFIX;

        /** IPv6 prefix for IPv4 mapped addresses */
        static const struct in6_addr IPV4_MAPPED_PREFIX;

        /** Class scope smart pointer typedef */
        typedef ConexusPointer<Address> pointer;

        /**
         * Default constructor, or if parameters are used constructs an address from
         * a numeric address and an optional port number.
         *
         * @param address Numeric address in local host byte order.
         * @param port Port number in local host byte order.
         */
        Address(const struct in6_addr& address = in6addr_any, uint16_t port = 0);

        Address(uint32_t host, uint16_t port=0);

        Address(const std::string& host, uint16_t port=0);

        /** Constructs from an IPv6 sockaddr in */
        Address(const struct sockaddr_in6& addr);

        /** Constructs from a generic sockaddr_storage */
        Address(const struct sockaddr_storage& addr);

        /** Constructs from an IPv4 address with either a mapped or compatible prefix */
        Address(const IPv4::Address& ipv4addr, IPV4_EMBEDDED_TYPE ipv4type = IPV4_MAPPED);

        /** Creates a smart pointer from an address and port number */
        static pointer create(const struct in6_addr& address = in6addr_any, uint16_t port = 0);

        static pointer create( uint32_t host, uint16_t port );

        static pointer create( const std::string& host, uint16_t port );

        /** Creates a smart pointer from a generic sockaddr_in */
        static pointer create(const struct sockaddr_in6& addr);

        /** Creates a smart pointer from a generic sockaddr_storage */
        static pointer create(const struct sockaddr_storage& addr);

        /** Creates a smart pointer from an IPv4 address with either a mapped or compatible prefix */
        static pointer create(const IPv4::Address& ipv4addr, IPV4_EMBEDDED_TYPE ipv4type = IPV4_MAPPED);

        /** Destructor */
        virtual ~Address();

        /** Returns the IPv6 address. */
        struct in6_addr address() const;

        /** Returns the IPv6 address as a string. */
        std::string address_string() const;

        /** Returns the subnet mask associated with this address. */
        struct in6_addr subnet_mask() const;

        /** Returns the netmask associated with this address as a string. */
        std::string subnet_mask_string() const;

        /** Returns the prefix length of the subnet mask associated with this address. */
        unsigned prefix_length() const;

        /** Returns the prefix length of the subnet mask associated with this address as a string. */
        std::string prefix_length_string() const;

        /** Returns the IPv6 address as an address/subnetmask string. */
        std::string address_subnet_mask_string() const;

        /** Returns the IPv6 address as a CIDR address/prefixlen string. */
        std::string cidr_address_string() const;

        /** Returns the prefix part of this address. */
        struct in6_addr prefix() const;

        /** Returns the prefix part of this address as a string. */
        std::string prefix_string() const;

        /** Returns the local part of this address. */
        struct in6_addr local_address() const;

        /** Returns the local part of this address as a string. */
        std::string local_address_string() const;

        /** Returns the port number associated with this address. */
        uint16_t port() const;

        /** Returns the port number associated with this address as a string. */
        std::string port_string() const;

        /** Set the address to the specified value. */
        void set_address(const struct in6_addr& address);

        /**
         * Sets the address to the specified value.
         * If this is not a numeric specification, a DNS query will be performed.
         * Accepts addresses of the following forms:
         *   # aaa.aaa.aaa.aaa[/bb][:ppppp | :servicename]
         *   # hostname[/bb][:ppppp | :servicename]
         * Where:
         *   - aaa.aaa.aaa.aaa is an address in dotted decimal form
         *   - hostname is a host name
         *   - bb is an optional prefix length
         *   - ppppp is an optional port specification
         *   - servicename is an optional service name specification
         */
        void set_address(const std::string& address);

        /** Sets the address and port to the specified value. */
        virtual void set_address(const std::string& address, uint16_t port);

        /** Set the address to an embedded IPv4 address in the IPv6 domain. */
        void set_address(uint32_t address, IPV4_EMBEDDED_TYPE type=IPV4_MAPPED);

        /** Sets the address to an absolute value with a specified CIDR prefix length. */
        void set_address_prefix_length(uint32_t address, unsigned prefix_length, IPV4_EMBEDDED_TYPE type=IPV4_MAPPED);

        /** Sets the prefix length to the specified value. */
        void set_prefix_length(unsigned prefix_length);

        /** Sets the prefix length to the specified value. */
        void set_prefix_length(const std::string& prefix_length);

        /** Sets the port portion of this address. */
        void set_port(uint16_t port);

        /** Sets the port to the specified value. */
        void set_port(const std::string& port);

        uint32_t flowinfo();

        void set_flowinfo( uint32_t f );

        uint32_t scope_id();

        void set_scope_id( uint32_t sid );

        /** Performs a DNS query on the currently set address and returns the hostname string. */
        std::string hostname() const;

        /** Performs a service name lookup on the currently set port and returns the servicename string */
        std::string servicename() const;

        /** True if this address is a reserved address with binary prefix 0000 0000 */
        bool is_reserved() const;

        /** True if this address is a reserved NSAP address with binary prefix 0000 0001 */
        bool is_nsap() const;

        /** True if this address is a reserved IPX address with binary prefix 0000 001 */
        bool is_ipx() const;

        /** True if this address is an aggregatable global unicast address with binary prefix 001 */
        bool is_aggregatable_global() const;

        /** True if this address is a link-local unicast address with binary prefix 1111 1110 10 */
        bool is_link_local() const;

        /** True if this address is a site-local unicast address with binary prefix 1111 1110 11 */
        bool is_site_local() const;

        /** True if this is a multicast address with binary prefix 1111 1111 */
        bool is_multicast() const;

        /**
         * True if this is a multicast interface local address.
         *
         * Interface-Local scope spans only a single interface on a node
         * and is useful only for loopback transmission of multicast.
         */
        bool is_multicast_interface_local() const;

        /**
         * True if this is a multicast link local address.
         *
         * Link-Local multicast scope spans the same topological region as
         * the corresponding unicast scope.
         */
        bool is_multicast_link_local() const;

        /**
         * True if this is a multicast site local address.
         *
         * Site-Local scope is intended to span a single site.
         */
        bool is_multicast_site_local() const;

        /**
         * True if this is a multicast organization local address.
         *
         * Organization-Local scope is intended to span multiple sites
         * belonging to a single organization.
         */
        bool is_multicast_organization_local() const;

        /** True if this is a multicast global address. */
        bool is_multicast_global() const;

        /** True if this is the any address 0:0:0:0:0:0:0:0 */
        bool is_any() const;

        /** True if this is the loopback address 0:0:0:0:0:0:0:0 */
        bool is_loopback() const;

        /** True if this is an IPv4 address */
        bool is_ipv4() const;

        /** True if this is an IPv4 compatible address */
        bool is_ipv4_compatible() const;

        /** True if this is an IPv4 mapped address */
        bool is_ipv4_mapped() const;

        /** Overrides the virtual parent method and provides the size in bytes of the underlying sockaddr_in structure. */
        socklen_t sockaddr_size() const;

        /** Provides a non-constant reference to the underlying sockaddr_in structure. */
        struct sockaddr_in6& sockaddr_in();

        /** Provides a non-constant pointer to the underlying sockaddr_in structure. */
        struct sockaddr_in6* sockaddr_in_ptr();

        operator struct sockaddr_in6*();

        /** Assignment operator */
        Address& operator=(const Address& other);

        Address& operator=( const Conexus::Address& other );

        /** Tests address equality */
        bool operator==(const Address& other);

        /** True if this is an IPv4 compatible address */
        virtual bool is_ipv4() { return this->is_ipv4_compatible(); }

        /** Always true for IPv6::Address */
        virtual bool is_ipv6() { return true; }

        virtual sigc::signal<void> signal_changed( );

        virtual sigc::signal<void> signal_address_changed( );

        virtual sigc::signal<void> signal_port_changed( );

        virtual sigc::signal<void> signal_flowinfo_changed( );

        virtual sigc::signal<void> signal_scope_id_changed( );

        virtual sigc::signal<void> signal_prefix_changed( );

        protected:
          struct in6_addr m_address;

          uint16_t m_port;

          uint32_t m_flowinfo;

          uint32_t m_scope_id;

          /** Netmask prefix length */
          unsigned m_prefix_length;

          sigc::signal<void> m_signal_changed;

          sigc::signal<void> m_signal_address_changed;

          sigc::signal<void> m_signal_port_changed;

          sigc::signal<void> m_signal_flowinfo_changed;

          sigc::signal<void> m_signal_scope_id_changed;

          sigc::signal<void> m_signal_prefix_changed;
      };

      /**
       * @relates Conexus::IPv4::Address
       * Given an IPv6 address, returns a string representation.
       * This is just a very thin convenience wrapper around inet_ntop() for IPv6 addresses.
       */
      std::string address_to_string(const struct in6_addr&);

      /**
       * @relates Conexus::IPv4::Address
       * Given a string, converts to an IPv6 address.
       * This is just a very thin convenience wrapper around inet_pton() for IPv6 addresses.
       */
      struct in6_addr string_to_address(const std::string&);

      /**
       * @relates Conexus::IPv4::Address
       * Convert a string to an address and a prefix length
       */
      void string_to_address_prefix(const std::string&, struct in6_addr& address, int& prefix);

      /**
       * @relates Conexus::IPv4::Address
       * Determine whether the provided string is a valid hostname by performing
       * a DNS query.
       */
      bool is_valid_hostname(const std::string&);

      /**
       * @relates Conexus::IPv4::Address
       * Given an address, look up the hostname.
       * Unlike the other functions in this section, this is not quite a thin wrapper around getnameinfo().
       * This function will try to first look for the name in the IPv6 domain. However, getnameinfo()
       * returns a success on an IPv4 mapped address, but returns the address string rather than
       * the hostname. Therefore, this function checks to see if any ':' characters are present
       * in the returned name and tries an IPv4 lookup to get the IPv4 hostname.
       */
      std::string address_to_hostname( const struct in6_addr& address ) throw (address_exception);

      /**
       * @relates Conexus::IPv4::Address
       * Given a service port number, returns a string name of the service.
       * This is just a very thin convenience wrapper around getnameinfo().
       */
      std::string service_to_servicename( uint16_t service ) throw (address_exception);

      /**
       * @relates Conexus::IPv4::Address
       * Given a hostname, get the IPv6 address.
       * This function will performa a hostname lookup and return either the IPv6 address of
       * the host, or an IPv4 mapped address if an IPv6 address is not available.
       */
      struct in6_addr hostname_to_address( const std::string& hostname ) throw (address_exception);

      /**
       * @relates Conexus::IPv4::Address
       * Given a service name, returns the port number of the service.
       * This is just a very thin convenience wrapper around getaddrinfo().
       */
      uint16_t servicename_to_service( const std::string& servicename ) throw (address_exception);

      /**
       * @relates Conexus::IPv4::Address
       * Checks to make sure this is a valid subnet mask; i.e. all 1's followed by all 0's
       */
      bool validate_subnet_mask( const struct in6_addr& subnet_mask ) throw();

      /**
       * @relates Conexus::IPv4::Address
       * Given a prefix length, return a subnet mask.
       */
      struct in6_addr create_subnet_mask(unsigned prefix_length);

  }

}

#endif
