/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_IPV6TCP_H
#define CONEXUS_IPV6TCP_H

#include <conexus/ipv6_ip.h>

/**
 * \example ipv6_tcp_client.cpp
 * Conexus::IPv6::TCP Client Example
 *
 * \example ipv6_tcp_echo_client.cpp
 * Conexus::IPv6::TCP Echo Client Example
 */

namespace Conexus
{

  namespace IPv6
  {

    /**
     * This class encapsulates a socket used for TCP/IP communications.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup IPv6
     * @ingroup Endpoint
     */
    class TCP : public IP
    {

      protected:

        friend class TCPServer;

        TCP( uint32_t address, uint16_t port );

        TCP( const struct in6_addr& address=in6addr_any, uint16_t port=0 );

        TCP( const std::string& address, uint16_t port );

      public:

        typedef ConexusPointer<TCP> pointer;

        CONEXUS_ENDPOINT_GENERIC_CREATE();

        static pointer create( uint16_t port = 0 );

        static pointer create( uint32_t address, uint16_t port );

        static pointer create( const struct in6_addr& address, uint16_t port=0 );

        static pointer create( const std::string& address, uint16_t port=0 );

        virtual ~TCP() throw();

        /**
         * Shut down the entirety or part of a TCP connection
         *
         * @param read_write use enum ReadWrite values to indicate which portion of the TCP connection (READ, WRITE or READWRITE) should be shut down
         * @return true if the requested portion of the TCP connection was shut down
         */
        virtual bool shutdown( ReadWrite read_write=READ_WRITE );


        virtual size_t input_available() throw();

        /**
         * Returns the size in bytes of the local send queue.
         */
        int pending_output_size();

      protected:

        /**
         * Protected constructor to be used by TCPServer to create TCP connections.
         * The second bool parameter is used to differentiate the public method from the private method.
         */
        TCP( int accepted_fd, bool );

        static TCP::pointer create( int accepted_fd, bool );

        virtual void on_local_address_changed( );
        virtual void on_remote_address_changed( );

        /**
         * Overloads the Socket ancestor send method to ensure that the socket is connected
         * before trying to send.
         */
        virtual size_t write_data( const Data data, Timeout timeout ) throw ( write_exception );

    };

  }

}

#endif
