/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ipv6_udpposet.h"

#include <sys/ioctl.h>
#include <linux/sockios.h>

namespace Conexus
{
  namespace IPv6
  {

    UDPPoset::UDPPoset( const struct in6_addr& local_if, uint16_t localport ) :
        UDP( local_if, localport, 0 )
    {
      m_write_without_connect = true;
    }

    UDPPoset::UDPPoset( const std::string& local_if, uint16_t localport ) :
        UDP( local_if, localport, 0 )
    {
      m_write_without_connect = true;
    }

    UDPPoset::pointer UDPPoset::create( uint16_t localport )
    {
      return pointer( new UDPPoset( in6addr_any, localport ) );
    }

    UDPPoset::pointer UDPPoset::create( const struct in6_addr & local_if, uint16_t localport )
    {
      return pointer( new UDPPoset( local_if, localport ) );
    }

    UDPPoset::pointer UDPPoset::create( const std::string & local_if, uint16_t localport )
    {
      return pointer( new UDPPoset( local_if, localport ) );
    }

    UDPPoset::~UDPPoset() throw()
    {}

    void UDPPoset::bind() throw ( bind_exception )
    {
      IP::bind();
    }

    void UDPPoset::bind( const Conexus::Address& a ) throw ( bind_exception )
    {
      IP::bind( a );
    }

    void UDPPoset::close() throw ( close_exception )
    {
      IP::close();
    }

  size_t UDPPoset::write_data(const Data data, Timeout timeout) throw (write_exception)
  {
    ssize_t bytes_transmitted = 0;
    if ( is_closed() )
      try {
        change_state( ENDPOINT_OPENED );
      } catch ( exception::state::failed ) {
        throw exception::write::not_opened();
      }

    std::multimap<int, Address>::reverse_iterator iter;
    for (iter = m_addresses.rbegin(); iter != m_addresses.rend(); iter++)
      bytes_transmitted += Socket::writeto( iter->second, data, timeout );

    return bytes_transmitted;
  }

  void UDPPoset::connect( ) throw( connect_exception )
    {
      throw exception::connect::no_udpposet_connect();
    }

    void UDPPoset::connect( const Address& a ) throw ( connect_exception )
    {
      throw exception::connect::no_udpposet_connect();
    }

    void UDPPoset::set_write_without_connect( bool value )
    {
      return ;
    }

    void UDPPoset::on_local_address_changed( )
    {
      // if the socket is already bound, close and reopen to same state with new local port and address
      if ( is_bound() )
        close_and_reopen();
    }

    void UDPPoset::on_remote_address_changed( )
  {}

    bool UDPPoset::add_destination( const Address& destination, int priority, bool singleton ) {
      if ( destination.is_broadcast() || destination.is_multicast() )
        return false;
      if ( singleton )
        remove_destination( destination );
      m_addresses.insert( std::pair<int,Address>(priority, destination) );
      m_signal_destination_added.emit(destination, priority);
      return true;
    }

    bool UDPPoset::add_destination( uint32_t host, uint16_t port, int priority, bool singleton )
    {
      return add_destination( Address(host, port), priority, singleton );
    }

    bool UDPPoset::add_destination( const struct in6_addr& host, uint16_t port, int priority, bool singleton )
    {
      return add_destination( Address(host, port), priority, singleton );
    }

    bool UDPPoset::add_destination( const std::string & host, uint16_t port, int priority, bool singleton )
    {
      return add_destination( Address(host, port), priority, singleton );
    }

    void UDPPoset::remove_destination( const Address & destination, int priority )
    {
      Destinations::iterator iter, iter2;
      std::pair<Destinations::iterator,Destinations::iterator> range;

      range = m_addresses.equal_range(priority);
      iter = range.first;
      while ( iter != range.second ) {
        if ( iter->second == destination ) {
          iter2 = iter++;

          // make sure the signal is emitted while the element is still valid
          m_signal_destination_removed.emit(iter2->second, iter2->first);

          m_addresses.erase(iter2, iter);
        }
        else
          iter++;
      }
    }

    void UDPPoset::remove_destination( const Address & destination )
    {
      Destinations::iterator iter, iter2;

      iter = m_addresses.begin();
      while ( iter != m_addresses.end() ) {
        if ( iter->second == destination ) {
          iter2 = iter++;

          // make sure the signal is emitted while the element is still valid
          m_signal_destination_removed.emit(iter2->second, iter2->first);

          m_addresses.erase(iter2, iter);
        }
        else
          iter++;
      }
    }

    bool UDPPoset::is_destination( const Address & destination )
    {
      Destinations::iterator iter;

      for (iter = m_addresses.begin(); iter != m_addresses.end(); iter++)
        if ( iter->second == destination )
          return true;
      return false;
    }

    bool UDPPoset::is_destination( const Address & destination, int priority )
    {
      Destinations::iterator iter;
      std::pair<Destinations::iterator,Destinations::iterator> range;

      range = m_addresses.equal_range(priority);
      for (iter = range.first; iter != range.second; iter++)
        if ( iter->second == destination )
          return true;
      return false;
    }

    std::set< int > UDPPoset::priorities( const Address & destination )
    {
      std::set<int> found;

      Destinations::iterator iter;

      for (iter = m_addresses.begin(); iter != m_addresses.end(); iter++)
        if ( iter->second == destination )
          found.insert(iter->first);

      return found;
    }

    const UDPPoset::Destinations& UDPPoset::destinations() {
      return m_addresses;
    }

    sigc::signal< void, Address, int > UDPPoset::signal_destination_added( )
    {
      return m_signal_destination_added;
    }

    sigc::signal< void, Address, int > UDPPoset::signal_destination_removed( )
    {
      return m_signal_destination_removed;
    }

  }
}

