/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUS_IPV6UDPPOSET_H
#define CONEXUS_IPV6UDPPOSET_H

#include <conexus/ipv6_udp.h>

#include <map>
#include <set>

/**
 * \example ipv6_udpposet_client.cpp
 * This example client demonstrates sending to three destinations
 * using Conexus::IPv6::UDPPoset.
 *
 * \example ipv6_udpposet_server.cpp
 * This example server demonstrates receiving on three ports
 * using Conexus::IPv6::UDPPoset.
 */

namespace Conexus
{

  namespace IPv6
  {

    /**
     * This class aggregates a single data transmission to
     * multiple UDP addresses. Although multicasting is generally a better solution, sometimes
     * multicasting is not possible with certain frameworks; hence UDPPoset.
     *
     * The advantage of UDPPoset over a vector of UDP sockets is that each instantiation
     * of a UDP socket will use a local port. In contrast, the UDPPoset class will use
     * a single port for all outgoing UDP traffic. Like the UDP class it inherits from,
     * it is capable of receiving transmissions, but only for the single local port.
     *
     * UDPPoset will only accept UDP unicast addresses; not multicast; not broadcast.
     *
     * The name poset, or partially ordered set, was used since UDP destinations of
     * equal priority have no guarantees on ordering. Priority is a numeric value,
     * and transmissions will occur from the highest priority first to the lowest
     * priority last.
     *
     * @author Rick L Vinyard Jr <rvinyard@cs.nmsu.edu>
     * @ingroup IPv6
     * @ingroup Endpoint
     */
    class UDPPoset : public UDP
    {
      protected:

        /** Constructor */
        UDPPoset( const struct in6_addr& local_if, uint16_t localport = 0 );

        /** Constructor */
        UDPPoset( const std::string& local_if, uint16_t localport = 0 );

      public:
        /** TODO Revisit and see if possible to add multicast support */
        /** TODO examine adding socket pools */

        /** Class typedef to destinations container */
        typedef std::multimap<int,Address> Destinations;

        /** Class typedef to smart pointer */
        typedef ConexusPointer<UDPPoset> pointer;

        CONEXUS_ENDPOINT_GENERIC_CREATE();

        /** Returns a smart pointer to a new object */
        static UDPPoset::pointer create( uint16_t localport = 0 );

        /** Returns a smart pointer to a new object */
        static UDPPoset::pointer create( const struct in6_addr& local_if, uint16_t localport = 0 );

        /** Returns a smart pointer to a new object */
        static UDPPoset::pointer create( const std::string& local_if, uint16_t localport = 0 );

        /** Destructor */
        virtual ~UDPPoset() throw();

        /** Bypasses parent UDP methods and goes straight to grandparent IP methods */
        virtual void bind() throw ( bind_exception );

        /** Bypasses parent UDP methods and goes straight to grandparent IP methods */
        virtual void bind( const Conexus::Address& a ) throw ( bind_exception );

        /** Bypasses parent UDP methods and goes straight to grandparent IP methods */
        virtual void close() throw ( close_exception );

        /**
         * UDPPoset cannot connect; always throws exception::connect::no_udpposet_connect()
         *
         * Since UDPPoset relies on \c sendto() to perform transmission, it cannot connect
         * since \c sendto() ignores the \c to parameter when connected.
         */
        virtual void connect() throw ( connect_exception );

        /**
         * UDPPoset cannot connect; always throws exception::connect::no_udpposet_connect()
         *
         * Since UDPPoset relies on \c sendto() to perform transmission, it cannot connect
         * since \c sendto() ignores the \c to parameter when connected.
         */
        virtual void connect( const Address& a ) throw ( connect_exception );

        /** Has no effect, write without connect must always be true */
        virtual void set_write_without_connect( bool value = true );

        /**
         * Add the destination with the given priority.
         *
         * @param destination The destination address to add to the poset
         * @param priority The priority level to add this destination to
         * @param singleton If true, this will be the only priority level the destination will occur
         * at, and there will only be a single occurrence at this priority level.
         *
         * The underlying model is a multimap from priority to destination. Unless the singleton
         * parameter is used duplicate destinations will be allowed for the same priority level
         * as well as multiple priority levels.
         *
         * What this means is that multiple transmissions will occur to the same destination. However,
         * the use of the singleton parameter, which defaults to true, will ensure the uniqueness
         * of a destination in the transmission set.
         *
         * Examples:
         *
         * \code
         * UDPPoset my_udpposet;
         * my_udpposet.add_destination( address1, 5 );
         * my_udpposet.add_destination( address1, 3 );
         * \endcode
         * Result: \b one transmission to address1 at priority level 3
         *
         * \code
         * UDPPoset my_udpposet;
         * my_udpposet.add_destination( address1, 5, false );
         * my_udpposet.add_destination( address1, 3, false );
         * \endcode
         * Result: \b two transmissions to address1 at priority levels 5 and 3
         *
         * \code
         * UDPPoset my_udpposet;
         * my_udpposet.add_destination( address1, 5, false );
         * my_udpposet.add_destination( address1, 5, false );
         * \endcode
         * Result: \b two transmissions to address1 at priority level 5
         *
         * \code
         * UDPPoset my_udpposet;
         * my_udpposet.add_destination( address1, 5, false );
         * my_udpposet.add_destination( address1, 3, false );
         * my_udpposet.add_destination( address1, 1 );
         * \endcode
         * Result: \b one transmission to address1 at priority level 1
         *
         * \code
         * UDPPoset my_udpposet;
         * my_udpposet.add_destination( address1, 5, false );
         * my_udpposet.add_destination( address1, 3, false );
         * my_udpposet.add_destination( address1, 1 );
         * my_udpposet.add_destination( address1, 0, false );
         * \endcode
         * Result: \b two transmissions to address1 at priority levels 1 and 0
         *
         */
        bool add_destination( const Address& destination, int priority=0, bool singleton=true );

        bool add_destination( uint32_t host, uint16_t port, int priority = 0, bool singleton = true );

        bool add_destination( const struct in6_addr& host, uint16_t port, int priority = 0, bool singleton = true );

        bool add_destination( const std::string& host, uint16_t port, int priority = 0, bool singleton = true );

        /** Removes all instances of a destination for a specific priority level */
        void remove_destination(const Address& destination, int priority );

        /** Removes all instances of a destination for all priority levels */
        void remove_destination(const Address& destination );

        /** True if the address is a destination at any priority level */
        bool is_destination( const Address& destination );

        /** True if the address is a destination at the specified priority level */
        bool is_destination( const Address& destination, int priority );

        /**
         * Returns a set of the priority levels this destination is associated with
         * or a vector of size zero if not present as a destination.
         */
        std::set<int> priorities( const Address& destination );

        /** Returns a constant reference to the map used to store destinations */
        const Destinations& destinations();

        /** Signal emitted when a destination is added */
        sigc::signal<void,Address,int> signal_destination_added();

        /** Signal emitted when a destination is removed */
        sigc::signal<void,Address,int> signal_destination_removed();

      protected:
        /** Destination addresses */
        Destinations m_addresses;

        /** Signal emitted when a destination is added */
        sigc::signal<void,Address,int> m_signal_destination_added;

        /** Signal emitted when a destination is removed */
        sigc::signal<void,Address,int> m_signal_destination_removed;

        /** Do nothing when parent callback occurs */
        virtual void on_local_address_changed( );

        /** Do nothing when parent callback occurs */
        virtual void on_remote_address_changed( );

      /**
       * Overloads the Socket ancestor send method to ensure that the socket is connected
       * before trying to send.
       */
      virtual size_t write_data(const Data data, Timeout timeout) throw (write_exception);

    };

  }

}

#endif
