/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "ll_packet.h"

#include <netinet/in.h>
#include <sys/ioctl.h>
#include <net/if.h>

namespace Conexus
{
  namespace LL
  {

    Packet::Packet( int protocol, int type ) :
        Socket( PF_PACKET, type, htons( protocol ) )
    {

    }

    Packet::pointer Packet::create( int protocol, int type )
    {
      return pointer( new Packet( protocol, type ) );
    }

    Packet::~Packet() throw()
    {}


    void Packet::bind() throw ( bind_exception )
    {
      m_address.set_protocol( m_protocol );
      m_address.set_interface( 1 );
      Socket::bind( m_address );
    }

    void Packet::bind( Address& a ) throw ( bind_exception )
    {
      Socket::bind( a );
    }

    size_t Packet::write_data( const Data data, Timeout timeout ) throw ( write_exception )
    {
      return Socket::writeto( m_address, data, timeout );
    }

    size_t Packet::input_available() throw()
    {
      return ETH_FRAME_LEN;
    }

    void Packet::set_interface( int ifindex )
    {
      set_interface( interface_name( ifindex ) );
    }

    void Packet::set_interface( const std::string ifname )
    {
      if ( is_closed() )
        change_state( ENDPOINT_OPENED );
      struct ifreq req;
      strncpy( req.ifr_name, ifname.c_str(), IFNAMSIZ );
      if ( ioctl( m_fd, SIOCGIFINDEX, &req ) >= 0 ) {
        m_address.m_sockaddr.ll.sll_ifindex = req.ifr_ifindex;
        if ( ioctl( m_fd, SIOCGIFHWADDR, &req ) >= 0 ) {
          m_address.m_sockaddr.ll.sll_hatype = req.ifr_hwaddr.sa_family;
          m_address.m_sockaddr.ll.sll_halen = 8;
          memcpy( m_address.m_sockaddr.ll.sll_addr, req.ifr_hwaddr.sa_data, 8 );
        }
      }
    }

    int Packet::interface_index( const std::string name )
    {
      if ( is_closed() )
        change_state( ENDPOINT_OPENED );
      int ifindex = -1;
      struct ifreq req;
      strncpy( req.ifr_name, name.c_str(), IFNAMSIZ );
      if ( ioctl( m_fd, SIOCGIFINDEX, &req ) >= 0 )
        ifindex = req.ifr_ifindex;
      return ifindex;
    }

    std::string Packet::interface_name( int ifindex )
    {
      if ( is_closed() )
        change_state( ENDPOINT_OPENED );
      struct ifreq req;
      std::string ifname;
      req.ifr_ifindex = ifindex;
      if ( ioctl( m_fd, SIOCGIFNAME, &req ) >= 0 )
        ifname = req.ifr_name;
      return ifname;
    }


#define MAX_REQ_SIZE 100

    std::vector<std::string> Packet::interface_names()
    {
      if ( is_closed() )
        change_state( ENDPOINT_OPENED );
      struct ifreq* req;
      struct ifconf conf;
      char buf[ sizeof( struct ifreq ) * MAX_REQ_SIZE ];
      std::vector<std::string> ifnames;

      conf.ifc_len = sizeof( buf );
      conf.ifc_buf = ( caddr_t ) buf;

      if ( ioctl( m_fd, SIOCGIFCONF, ( caddr_t ) & conf ) >= 0 )
        for ( req = ( struct ifreq* ) conf.ifc_buf; ( caddr_t ) req < conf.ifc_buf + conf.ifc_len; req++ )
          ifnames.push_back( std::string( req->ifr_name ) );

      return ifnames;
    }

  } // namespace LL
} // namespace Conexus
