/***************************************************************************
 *   Copyright (C) 2007 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "messagequeue.h"

namespace Conexus
{

  MessageQueue::MessageQueue ( const std::string& name, unsigned flags, mode_t mode, bool close_on_destruct ) :
      Endpoint ( close_on_destruct ),
      m_name ( name ),
      m_flags ( flags ),
      m_unlink_on_destruction ( false ),
      m_max_messages ( 0 ),
      m_max_message_size ( 0 ),
      m_mode ( mode ),
      m_queue ( 0 )
  {
  }

  MessageQueue::pointer MessageQueue::create ( const std::string& name,
          unsigned flags,
          mode_t mode,
          bool close_on_destruct
                                             )
  {
    return pointer ( new MessageQueue ( name, flags, mode, close_on_destruct ) );
  }

  MessageQueue::~MessageQueue()
  {
    if ( m_unlink_on_destruction )
      ::mq_unlink ( m_name.c_str() );
  }

  void MessageQueue::open( ) throw ( open_exception )
  {
    struct mq_attr mqa;
    mqd_t status;
    unsigned flags = 0;

    if ( ( m_flags & QUEUE_READ ) && ( m_flags & QUEUE_WRITE ) )
      flags |= O_RDWR;
    else if ( m_flags & QUEUE_READ )
      flags |= O_RDONLY;
    else if ( m_flags & QUEUE_WRITE )
      flags |= O_WRONLY;
    else
      throw exception::open::no_rd_wr_rdwr();

    if ( m_flags & QUEUE_NONBLOCKING )
      flags |= O_NONBLOCK;

    if ( m_flags & QUEUE_CREATE )
      flags |= O_CREAT;

    if ( m_flags & QUEUE_EXCLUSIVE_CREATE )
      flags |= O_EXCL | O_CREAT;

    if ( flags & O_CREAT )
      m_queue = ::mq_open ( m_name.c_str(), flags, m_mode, NULL );
    else
      m_queue = ::mq_open ( m_name.c_str(), flags );

    if ( m_queue == -1 )
    {
      perror ( strerror ( errno ) );
      throw_open_exception ( errno );
    }

    status = ::mq_getattr ( m_queue, &mqa );

    if ( status == 0 )
    {
      m_max_messages = mqa.mq_maxmsg;
      m_max_message_size = mqa.mq_msgsize;
    }

    this->set_state_opened();
  }

  void MessageQueue::close ( bool force ) throw ( close_exception )
  {
    mqd_t status;

    status = ::mq_close ( m_queue );
    if ( status == -1 )
      throw_close_exception ( errno );

    this->set_state_closed();
  }

  const std::string & MessageQueue::name( ) const
  {
    return m_name;
  }

  void MessageQueue::set_name ( const std::string & name )
  {
    if ( name != m_name )
      m_name = name;
  }

  unsigned MessageQueue::flags( ) const
  {
    return m_flags;
  }

  void MessageQueue::set_flags ( unsigned flags )
  {
    m_flags = flags;
  }

  bool MessageQueue::unlink_on_destruction( )
  {
    return m_unlink_on_destruction;
  }

  void MessageQueue::set_unlink_on_destruction ( bool u )
  {
    m_unlink_on_destruction = u;
  }

  long int MessageQueue::max_messages( )
  {
    return m_max_messages;
  }

  long int MessageQueue::max_message_size( )
  {
    return m_max_message_size;
  }

  size_t MessageQueue::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
  {
    mqd_t result;

    if ( ! data ) return 0;

    if ( this->is_closed() ) this->open();

    // why don't we need to check to see if the open succeeded?
    // because if it didn't an exception was thrown and we wouldn't be here

    if ( timeout.is_null() )
    {
      result = ::mq_send ( m_queue, (const char*)data, data.size(), 0 );
      if ( result == 0 ) return data.size();
    }
    else
    {
      struct timespec ts_timeout = timeout;
      result = ::mq_timedsend ( m_queue, (const char*)data, data.size(), 0, &ts_timeout );
      if ( result == 0 ) return data.size();
    }

    throw_write_exception ( errno );

    return 0;
  }

  Data MessageQueue::read_data ( size_t s, Timeout timeout ) throw ( read_exception )
  {
    mqd_t result;
    unsigned priority;

    Data data = Data ( m_max_message_size );

    if ( this->is_closed() ) this->open();

    if ( timeout.is_null() ) timeout = m_default_read_timeout;

    if ( timeout.is_zero() )
    {
      result = ::mq_receive ( m_queue, data, data.size(), &priority );
    }
    else
    {
      struct timespec ts_timeout = timeout;
      result = ::mq_timedreceive ( m_queue, data, data.size(), &priority, &ts_timeout );
    }

    switch (result)
    {
      case -1:
        throw_read_exception ( errno );
      case 0:
        throw exception::read::disconnected();
      default:
        data.resize ( result );
        if ( m_timestamp ) data.set_current_time();
        return data;
    }
  }

}


