/***************************************************************************
 *   Copyright (C) 2007 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSOBJECT_H
#define CONEXUSOBJECT_H

#include <string>
#include <sigc++/sigc++.h>

#include <conexus/pointer.h>

namespace Conexus
{
  /**
   * This class provides a common inheritance hierarchy for all conexus
   * objects.
   *
   * In addition to providing a common hierarchy this base class ensures
   * that all descendants are also descendants of sigc::trackable allowing
   * automatic signal disconnection on destruction.
   *
   * This base class also provides a common mechanism for naming conexus
   * objects.
   * 
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   */
  class Object: public sigc::trackable
  {
    protected:
      /**
       * The only constructor accepts a name parameter and is protected to ensure
       * that only children derived from \c Object will call the constructor directly.
       */
      Object(const std::string& name=std::string()): m_name(name) { }
      
    public:
      /** This typedef provides a common notation for smart pointers */
      typedef ConexusPointer<Object> pointer;

      /** Destructor does nothing */
      ~Object() { }

      /** Gets the object's name */
      const std::string& name() { return m_name; }

      /** Sets the object's name and emits \c signal_name_changed after the name has changed. */
      void set_name( const std::string& n ) { m_name = n; m_signal_name_changed.emit(); }

      /** Accessor for the signal emitted when then name is changed */
      sigc::signal<void> signal_name_changed() { return m_signal_name_changed; }
      
    private:
      /** Signal emitted when the name is changed */
      sigc::signal<void> m_signal_name_changed;

      /** Internal storage for the object name */
      std::string m_name;
  };

}

#endif
