/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "pipe.h"

#include <unistd.h>
#include <fcntl.h>

namespace Conexus
{

  Pipe::Pipe( ) :
      FileDescriptor(),
      m_write_fd( -1 )
  {}

  Pipe::pointer Pipe::create( )
  {
    return pointer( new Pipe( ) );
  }

  Pipe::~Pipe()
  {}

  void Pipe::open( ) throw( open_exception )
  {
    int fds[ 2 ];
    int result;

    result = pipe( fds );

    switch ( result ) {
      case 0:
        break;
      case EFAULT:
        break;
      case EMFILE:
        break;
      case ENFILE:
        break;
    }

    m_fd = fds[ 0 ];
    m_write_fd = fds[ 1 ];
    this->set_state_opened();
    m_state |= PIPE_READ | PIPE_WRITE;
  }

  void Pipe::close( bool force ) throw ( close_exception )
  {
    int i;
    if ( m_state & PIPE_READ ) {
      i = ::close( m_fd );
      if ( i == -1 )
        throw_close_exception( errno );
    }

    if ( m_state & PIPE_WRITE ) {
      i = ::close( m_write_fd );
      if ( i == -1 )
        throw_close_exception( errno );
    }

    this->set_state_closed();
  }

  void Pipe::close_read( bool force ) throw ( close_exception )
  {
    int i;
    if ( m_state & PIPE_READ ) {
      i = ::close( m_fd );
      if ( i == -1 )
        throw_close_exception( errno );
      m_state &= ~PIPE_READ;
    }
  }

  void Pipe::close_write( bool force ) throw ( close_exception )
  {
    int i;
    if ( m_state & PIPE_WRITE ) {
      i = ::close( m_write_fd );
      if ( i == -1 )
        throw_close_exception( errno );
      m_state &= ~PIPE_WRITE;
    }
  }

  size_t Pipe::input_available() throw()
  {
    return PIPE_BUF;
  }

  size_t Pipe::write_data( const Data data, Timeout timeout ) throw ( write_exception )
  {
    int fcntl_flags;
    int fcntl_result;
    int write_result;

    // Do we have valid data?
    if ( ! data ) throw_write_exception ( EINVAL );

    // Is the file descriptor open?
    if ( this->is_closed() ) this->open();

    // We need the current flags whether this is going to be blocking or non-blocking
    fcntl_flags = ::fcntl ( m_write_fd, F_GETFL );
    if ( fcntl_flags == -1 ) throw_write_exception ( errno );

    // We need to do some prep work based on whether the timeout indicates
    // a blocking write, a non-blocking immediate write, or a blocking with timeout

    // Only a non-blocking immediate write needs O_NONBLOCK set
    if ( timeout.is_zero() )
    {
      // Switch to non-blocking if necessary
      if ( ! ( fcntl_flags & O_NONBLOCK ) )
      {
        fcntl_flags |= O_NONBLOCK;
        fcntl_result = ::fcntl ( m_write_fd, F_SETFL, fcntl_flags );
        if ( fcntl_result == -1 ) throw_write_exception ( errno );
      }
    }
    else
    {
      // Switch to blocking if necessary
      if ( fcntl_flags & O_NONBLOCK )
      {
        fcntl_flags &= ~O_NONBLOCK;
        fcntl_result = ::fcntl ( m_write_fd, F_SETFL, fcntl_flags );
        if ( fcntl_result == -1 ) throw_write_exception ( errno );
      }

      // If we have a timeout value, then we need to wait in select()
      if ( ! timeout.is_null() )
      {
        fd_set writefd;
        struct timeval tv_timeout = timeout;
        int select_fd_set;
        
        FD_ZERO ( &writefd );
        FD_SET ( m_write_fd, &writefd );
        select_fd_set = ::select ( m_write_fd + 1, NULL, &writefd, NULL, &tv_timeout );
        if ( select_fd_set == 0 ) throw_write_exception(ETIMEDOUT);
        else if ( select_fd_set == -1 ) throw_write_exception(errno);
      }
    }

    write_result = ::write( m_write_fd, data, data.size() );

    if ( write_result == -1 ) throw_write_exception(errno);

    return write_result;
  }

  Data Pipe::read_data( size_t size, Timeout timeout ) throw ( read_exception )
  {
    Data d = FileDescriptor::read_data( size, timeout );

    if ( ! d ) throw ( exception::read::disconnected() );

    return d;
  }

}

