/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "serial.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <iostream>

namespace Conexus
  {

  Serial::Serial(const std::string& device, ReadWrite mode):
      m_rwmode(mode)
  {
    if ( m_rwmode == READ_WRITE_UNCHANGED )
      throw(Conexus::exception::state::invalid_read_write());
    this->set_device(device);
  }

  Serial::pointer Serial::create(const std::string& device, ReadWrite rwmode)
  {
    return pointer( new Serial( device, rwmode ) );
  }

  Serial::~Serial()
  {}

  void Serial::open( const std::string name, ReadWrite rwmode ) throw (open_exception)
  {
    int flags=O_NOCTTY|O_NDELAY;

    if ( is_open() )
      change_state(ENDPOINT_CLOSED);

    if ( rwmode != READ_WRITE_UNCHANGED )
      m_rwmode = rwmode;

    switch (m_rwmode)
    {
      case READ:
        flags |= O_RDONLY;
        break;
      case WRITE:
        flags |= O_WRONLY;
        break;
      case READ_WRITE:
        flags |= O_RDWR;
        break;
      default:
        /* do nothing */
        break;
    }
    
    m_fd = ::open(name.c_str(), flags);

    if ( m_fd == -1 )
      throw open_exception(errno);

    this->set_state_opened();
  }

  void Serial::open( ) throw( open_exception )
  {
    if (m_device.size() == 0)
      throw(exception::open::no_device_set());
    this->open( m_device );
  }

  ReadWrite Serial::rw_mode( )
  {
    return m_rwmode;
  }

  std::string Serial::device( )
  {
    return m_device;
  }

  void Serial::set_device( const std::string& d, ReadWrite rwmode ) throw (open_exception)
  {
    long unsigned tempstate;

    // if the device is the same and the state is the same, then leave
    if (m_device == d && (m_rwmode==rwmode || rwmode == READ_WRITE_UNCHANGED ))
      return;

    // set the new device
    m_device = d;
    if ( rwmode != READ_WRITE_UNCHANGED ) m_rwmode = rwmode;

    tempstate = m_state;

    // close the old file (changes the state so this needs to be after check above)
    this->close();

    // only potentially reopen the file if the device is not an empty string
    if ( m_device.size() > 0 )
      this->change_state(tempstate);

    m_signal_device_changed.emit();
  }

  sigc::signal<void> Serial::signal_device_changed()
  {
    return m_signal_device_changed;
  }

}

