/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSSERIAL_H
#define CONEXUSSERIAL_H

#include <conexus/filedescriptor.h>

/**
 * \example serial_send.cpp
 * Conexus::Serial send example
 */


namespace Conexus
{

  /**
   * This class encapsulates a raw serial port. It is likely that it will be extended one
   * level deeper in the inheritance hierarchy and replaced by an ancestor that provides
   * file I/O to be used with traditional files and with Unix named pipes.
   *
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   * @ingroup Endpoint
   */
  class Serial : public FileDescriptor
  {
    protected:

      Serial( const std::string& device = std::string(), ReadWrite rwmode = READ_WRITE );

    public:

      typedef ConexusPointer<Serial> pointer;

      CONEXUS_ENDPOINT_GENERIC_CREATE();

      static Serial::pointer create( const std::string& device = std::string(), ReadWrite rwmode = READ_WRITE );

      virtual ~Serial();

      /**
       * Opens the named serial port device in the specified mode.
       */
      virtual void open( const std::string name, ReadWrite rwmode = READ_WRITE_UNCHANGED ) throw ( open_exception );

      /**
       * Will attempt to open the serial port if a device has been specified. Otherwise an
       * exception will be thrown.
       */
      virtual void open() throw ( open_exception );

      ReadWrite rw_mode();

      /**
       * Will return the string representing the device associated with this serial object.
       */
      std::string device();

      /**
       * Sets the string representing the device associated with this serial object.
       *
       * If the serial device is currently open and d is not equal to the current device,
       * the current device will be closed.
       *
       * Also, if auto_open is set, the new device will be immediately opened.
       *
       * Additionally, this method checks whether the parameter d is the same as the current device,
       * and parameter mode is the same as the current mode.
       * If both are equal, the device will not be closed and reopened unless force=true.
       */
      void set_device( const std::string& d, ReadWrite rwmode = READ_WRITE_UNCHANGED ) throw ( open_exception );

      sigc::signal<void> signal_device_changed();

    protected:
      std::string m_device;
      ReadWrite m_rwmode;

      sigc::signal<void> m_signal_device_changed;
  };

}

#endif
