/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#ifndef CONEXUSSERVER_H
#define CONEXUSSERVER_H

#include <set>
#include <queue>

#include <sys/select.h>

#include <sigc++/sigc++.h>

#include <glibmm/dispatcher.h>
#include <glibmm/thread.h>

#include <conexus/enums.h>
#include <conexus/data.h>
#include <conexus/object.h>

namespace Conexus
{

  /**
   * @author Rick L Vinyard Jr
   * @ingroup conexus
   */
  class Server: public Object
  {
    protected:

      Server();

    public:

      typedef ConexusPointer<Server> pointer;

      virtual ~Server();

      virtual void start();
      virtual void start( bool use_dispatcher );
      virtual void stop();

      /**
       * Sets the responsiveness to add_*, remove_* and stop commands in microseconds;
       * lower is more responsive; larger is less responsive; default is 500000 microseconds (or 500 ms).
       *
       * In general this refers to the timeout period of the @em select() call used internally.
       * Each cycle of the server I/O thread acquires several mutex locks at the beginning
       * of the cycle and releases them at the end of the cycle.
       *
       * This value shortens the cycle period, thereby making the server more responsive to commands at
       * the cost of increasing overhead.
         */
      virtual void set_responsiveness( long r );

      /**
       * Returns the server responsiveness level.
       */
      virtual long responsiveness();

      virtual bool is_running();

      sigc::signal<void,StartStop> signal_start_stop();

    protected:
      Glib::Thread* m_thread;
      bool m_terminate;
      bool m_running;
      Glib::Dispatcher* m_dispatcher;
      bool m_use_dispatcher;
      struct timeval m_timeval;
      sigc::signal<void,StartStop> m_signal_start_stop;

      virtual void service_thread_main() = 0;

      virtual void pre_start() { }

      virtual void post_start() { }

      virtual void pre_stop() { }

      virtual void post_stop() { }

      virtual void emit() = 0;

    private:
      void service_thread_main_proxy();

      void emit_proxy();

  };

}

#endif
