/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "socket.h"

#include <errno.h>
#include <sys/socket.h>
#include <unistd.h>
#include <fcntl.h>

namespace Conexus
{

  Socket::Socket ( int domain, int type, int protocol ) throw () :
      FileDescriptor(),
      m_domain ( domain ),
      m_type ( type ),
      m_protocol ( protocol )
  {}

  Socket::pointer Socket::create ( int domain, int type, int protocol ) throw ()
  {
    return pointer ( new Socket ( domain, type, protocol ) );
  }

  Socket::~Socket() throw ()
  {}

  int Socket::domain() throw ()
  {
    return m_domain;
  }

  void Socket::set_domain ( int d ) throw ()
  {
    m_domain = d;
  }

  int Socket::type() throw ()
  {
    return m_type;
  }

  void Socket::set_type ( int t ) throw ()
  {
    m_type = t;
  }

  int Socket::protocol() throw ()
  {
    return m_protocol;
  }

  void Socket::set_protocol ( int p ) throw ()
  {
    m_protocol = p;
  }

  void Socket::open() throw ( open_exception )
  {
    if ( this->is_open() )
      this->change_state ( ENDPOINT_CLOSED );

    m_fd = ::socket ( m_domain, m_type, m_protocol );
    if ( m_fd == -1 )
      throw_open_exception ( errno );

    this->set_state_opened();
  }

  void Socket::close ( bool force ) throw ( close_exception )
  {
    int i;

    if ( this->is_closed() && !force )
      return ;

    i = ::close ( m_fd );
    if ( i == -1 )
      throw_close_exception ( errno );

    this->set_state_closed();
  }


  void Socket::bind() throw ( bind_exception )
  {
    throw exception::bind::no_socket_bind();
  }

  void Socket::bind ( const Address& a ) throw ( bind_exception )
  {
    int i;

    if ( this->is_closed() )
      try
      {
        this->change_state ( ENDPOINT_OPENED );
      }
      catch ( exception::state::failed )
      {
        throw exception::bind::not_open();
      }

    i = ::bind ( m_fd, a.sockaddr_ptr(), a.sockaddr_size() );
    if ( i == -1 )
      throw_bind_exception ( errno );

    this->set_state_bound();
  }

  void Socket::connect() throw ( connect_exception )
  {
    throw exception::connect::no_socket_connect();
  }

  void Socket::connect ( const Address& a ) throw ( connect_exception )
  {
    int i;

    if ( this->is_closed() )
      try
      {
        this->change_state ( ENDPOINT_OPENED );
      }
      catch ( exception::state::failed )
      {
        throw exception::connect::not_open();
      }

    if ( a.is_broadcast() )
      this->set_option ( SO_BROADCAST, true );
    else
      this->set_option ( SO_BROADCAST, false );

    i = ::connect ( m_fd, a.sockaddr_ptr(), a.sockaddr_size() );
    if ( i == -1 && errno != EINPROGRESS )
      throw_connect_exception ( errno );

    this->set_state_connected();
  }

  void Socket::listen ( int backlog )
  {
    int i;

    if ( ! this->is_bound() )
      try
      {
        this->change_state ( SOCKET_BOUND );
      }
      catch ( exception::state::failed )
      {
        throw exception::listen::not_bound();
      }

    i = ::listen ( m_fd, backlog );

    if ( i == -1 )
      throw_listen_exception ( errno );

    this->set_state_listening();
  }

  size_t Socket::writeto ( Address& a, const void* data, size_t size, Timeout timeout ) throw ( write_exception )
  {
    // TODO use timeout parameter
    ssize_t ret;

    if ( data == NULL || size <= 0 )
      return 0;

    if ( this->is_closed() )
      try
      {
        this->change_state ( ENDPOINT_OPENED );
      }
      catch ( exception::state::failed )
      {
        throw exception::write::not_socket();
      }

    if ( a.is_broadcast() )
      this->set_option ( SO_BROADCAST, true );
    else
      this->set_option ( SO_BROADCAST, false );

    ret = ::sendto ( m_fd, data, size, 0, a.sockaddr_ptr(), a.sockaddr_size() );

    if ( ret == -1 )
      throw_write_exception ( errno );

    return ret;
  }

  size_t Socket::writeto ( Address& a, const Data data, Timeout timeout ) throw ( write_exception )
  {
    return this->writeto ( a, data, data.size(), timeout );
  }

  void Socket::set_option ( int option, bool b, long required_state )
  {
    int i = b;

    try
    {
      this->change_state ( required_state );
    }
    catch ( exception::state::failed )
    {
      return ;
    }

    int result = ::setsockopt ( m_fd, SOL_SOCKET, option, &i, sizeof ( int ) );
    if ( result == -1 ) throw_socket_exception ( errno );
  }

  void Socket::set_reuse_address ( bool set )
  {
    return set_option ( SO_REUSEADDR, set, ENDPOINT_OPENED );
  }

  void Socket::read_thread_main()
  {
    // ensure that the state is at least BOUND
    if ( ! is_bound() )
      try
      {
        change_state ( SOCKET_BOUND );
      }
      catch ( exception::state::failed )
      {
        throw exception::read::not_bound();
      }

    FileDescriptor::read_thread_main();
  }

  void Socket::change_state ( long states ) throw ( state_exception )
  {
    // check for inconsistent states
    // if it's CLOSED it can't be BOUND, CONNECTED, LISTENING or ACCEPTED
    if ( ( states & ENDPOINT_CLOSED ) && ( ( states & SOCKET_BOUND ) || ( states & SOCKET_CONNECTED ) || ( states & SOCKET_LISTENING ) || ( states & SOCKET_ACCEPTED ) ) )
      throw exception::state::inconsistent();

    // call for CLOSED, OPENED and UNCHANGED
    // also checks for existing state
    // if it does the job, go ahead and return
    Endpoint::change_state ( states );

    if ( ! this->is_bound() && states & SOCKET_BOUND )
      this->bind();
    if ( ! this->is_connected() && states & SOCKET_CONNECTED )
      this->connect();
    if ( ! this->is_listening() && states & SOCKET_LISTENING )
      this->listen();
  }

  sigc::signal<void> Socket::signal_bound()
  {
    return m_signal_bound;
  }

  sigc::signal<void> Socket::signal_connected()
  {
    return m_signal_connected;
  }

  sigc::signal<void> Socket::signal_listening()
  {
    return m_signal_listening;
  }

  bool Socket::is_bound()
  {
    return m_state & SOCKET_BOUND;
  }

  bool Socket::is_connected()
  {
    return m_state & SOCKET_CONNECTED;
  }

  bool Socket::is_listening()
  {
    return m_state & SOCKET_LISTENING;
  }

  bool Socket::is_accepted()
  {
    return m_state & SOCKET_ACCEPTED;
  }


  void Socket::set_state_closed()
  {
    m_state &= ~SOCKET_BOUND;
    m_state &= ~SOCKET_CONNECTED;
    m_state &= ~SOCKET_LISTENING;
    m_state &= ~SOCKET_ACCEPTED;
    if ( this->is_closed() )
      return ;
    Endpoint::set_state_closed();
  }

  void Socket::set_state_bound()
  {
    if ( this->is_bound() )
      return ;
    m_state |= SOCKET_BOUND;
    m_signal_bound.emit();
  }

  void Socket::set_state_connected()
  {
    if ( this->is_connected() )
      return ;
    m_state |= SOCKET_CONNECTED;
    m_signal_connected.emit();
  }

  void Socket::set_state_listening()
  {
    if ( this->is_listening() )
      return ;
    m_state |= SOCKET_LISTENING;
    m_signal_listening.emit();
  }

  size_t Socket::write_data ( const Data data, Timeout timeout ) throw ( write_exception )
  {
    if ( ! data ) return 0;

    if ( ! this->is_connected() )
      try
      {
        this->change_state ( SOCKET_CONNECTED );
      }
      catch ( exception::state::failed )
      {
        throw exception::write::not_connected();
      }

    return FileDescriptor::write_data ( data, timeout );
  }

  Data Socket::read_data ( size_t size, Timeout timeout ) throw ( read_exception )
  {
    if ( ! this->is_bound() )
      try
      {
        this->change_state ( SOCKET_BOUND );
      }
      catch ( exception::state::failed )
      {
        throw exception::read::not_bound();
      }

    Data d = FileDescriptor::read_data ( size, timeout );

    if ( d.size() == 0 ) throw exception::read::disconnected();

    return d;
  }

}

