/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "tty.h"

#include <termios.h>
#include <sys/ioctl.h>
#include <iostream>


namespace Conexus
  {

  TTY::TTY(const std::string& device, ReadWrite rwmode):
      Serial(device, rwmode)
  {
    memset( &m_termios, 0x00, sizeof(struct termios) );
    m_termios.c_cflag |= CS8 | CLOCAL | CREAD | CRTSCTS;
    m_termios.c_iflag |= IGNBRK;
    m_termios.c_oflag |= NL0 | CR0 | TAB0 | BS0 | VT0 | FF0;
    m_termios.c_cc[VMIN] = 1;
    m_termios.c_cc[VTIME] = 0;
  }

  TTY::pointer TTY::create(const std::string& device, ReadWrite rwmode) {
    return pointer( new TTY(device, rwmode) );
  }

  TTY::~TTY()
  {
    ::tcsetattr(m_fd, TCSANOW, &m_origtermios);
  }

  void TTY::open() throw (open_exception)
  {
    Serial::open();
  }

  void TTY::open(const std::string name, ReadWrite rwmode)  throw (open_exception)
  {
    Serial::open(name, rwmode);
    // TODO revisit the effet of commenting out next line
//     set_state_opened();
    ::tcgetattr(m_fd, &m_origtermios);
    this->flush();
    this->tcsetattr( SET_NOW );
  }

  void TTY::close(bool force) throw (close_exception) {
    if ( is_closed() && !force )
      return;

    if ( m_reset_on_close )
      ::tcsetattr(m_fd, TCSANOW, &m_origtermios);
    FileDescriptor::close();
  }

  void TTY::drain( )
  {
    ::tcdrain(m_fd);
  }

  void TTY::flush_input( )
  {
    ::tcflush(m_fd, TCIFLUSH);
  }

  void TTY::flush_output( )
  {
    ::tcflush(m_fd, TCOFLUSH);
  }

  void TTY::flush( )
  {
    ::tcflush(m_fd, TCIOFLUSH);
  }

  void TTY::suspend_output( )
  {
    ::tcflow(m_fd, TCOOFF);
  }

  void TTY::restart_output( )
  {
    ::tcflow(m_fd, TCOON);
  }

  void TTY::suspend_input( )
  {
    ::tcflow(m_fd, TCIOFF);
  }

  void TTY::restart_input( )
  {
    ::tcflow(m_fd, TCION);
  }

  void TTY::set_input_speed( unsigned speed, SetOption option )
  {
    ::cfsetispeed( &m_termios, unsigned2speed(speed) );
    this->tcsetattr(option);
    m_signal_input_speed.emit();
  }

  void TTY::set_output_speed( unsigned speed, SetOption option )
  {
    ::cfsetospeed( &m_termios, unsigned2speed(speed) );
    this->tcsetattr(option);
    m_signal_output_speed.emit();
  }

  void TTY::set_speed( unsigned speed, SetOption option )
  {
    ::cfsetispeed(&m_termios, unsigned2speed(speed));
    ::cfsetospeed(&m_termios, unsigned2speed(speed));
    this->tcsetattr(option);
    m_signal_input_speed.emit();
    m_signal_output_speed.emit();
  }

  unsigned TTY::input_speed( )
  {
    return speed2unsigned( ::cfgetispeed(&m_termios) );
  }

  unsigned TTY::output_speed( )
  {
    return speed2unsigned( ::cfgetospeed(&m_termios) );
  }

  void TTY::set_parity(Parity parity, ParityError error, SetOption option )
  {
    if ( parity == PARITY_NONE )
      m_termios.c_cflag &= ~PARENB;
    else
      {
        m_termios.c_cflag |= PARENB;
        if ( parity == PARITY_ODD )
          m_termios.c_cflag |= PARODD;
        else
          m_termios.c_cflag &= ~PARODD;

        switch ( error )
          {
          case PARITY_ERROR_DISCARD: /** Discards all parity error bytes */
            m_termios.c_iflag &= ~PARMRK;
            m_termios.c_iflag &= ~INPCK;
            m_termios.c_iflag |= IGNPAR;
            break;
          case PARITY_ERROR_MARK:    /** Delivers three bytes (0xFF, 0x00 and c; where c is the error byte */
            m_termios.c_iflag &= ~IGNPAR;
            m_termios.c_iflag &= ~INPCK;
            m_termios.c_iflag |= PARMRK;
            break;
          case PARITY_ERROR_ZERO:    /** Delivers 0x00 when a parity error is detected */
            m_termios.c_iflag &= ~IGNPAR;
            m_termios.c_iflag &= ~PARMRK;
            m_termios.c_iflag |= INPCK;
            break;
          case PARITY_ERROR_IGNORE:  /** Ignores all parity errors and delivers all data as is */
            m_termios.c_iflag &= ~IGNPAR;
            m_termios.c_iflag &= ~PARMRK;
            m_termios.c_iflag &= ~INPCK;
            break;
          }
      }
    this->tcsetattr(option);
    m_signal_parity.emit();
  }

  Parity TTY::parity()
  {
    if ( m_termios.c_cflag & PARENB )
      {
        if ( m_termios.c_cflag & PARODD )
          return PARITY_ODD;
        else
          return PARITY_EVEN;
      }
    return PARITY_NONE;
  }

  ParityError TTY::parity_error()
  {
    if ( m_termios.c_cflag & PARENB )
      {
        if ( m_termios.c_iflag & IGNPAR )
          return PARITY_ERROR_DISCARD;
        if ( m_termios.c_iflag & PARMRK )
          return PARITY_ERROR_MARK;
        if ( m_termios.c_iflag & INPCK )
          return PARITY_ERROR_ZERO;
      }
    return PARITY_ERROR_IGNORE;
  }

  unsigned TTY::byte_size()
  {
    if ( m_termios.c_cflag & CS5 )
      return 5;
    if ( m_termios.c_cflag & CS6 )
      return 6;
    if ( m_termios.c_cflag & CS7 )
      return 7;
    if ( m_termios.c_cflag & CS8 )
      return 8;
    return 0;
  }

  void TTY::set_byte_size( unsigned size, SetOption option )
  {
    if ( size < 5 || size > 8)
      return;

    switch (size)
      {
      case 5:
        m_termios.c_cflag = (m_termios.c_cflag & ~CSIZE) | CS5;
        break;
      case 6:
        m_termios.c_cflag = (m_termios.c_cflag & ~CSIZE) | CS6;
        break;
      case 7:
        m_termios.c_cflag = (m_termios.c_cflag & ~CSIZE) | CS7;
        break;
      case 8:
        m_termios.c_cflag = (m_termios.c_cflag & ~CSIZE) | CS8;
        break;
      }
    this->tcsetattr(option);
    m_signal_byte_size.emit();
  }

  unsigned TTY::stop_bits() {
    if ( m_termios.c_cflag & CSTOPB )
      return 2;
    else
      return 1;
  }

  void TTY::set_stop_bits( unsigned size, SetOption option ) {
    if ( size == 1 )
      m_termios.c_cflag = m_termios.c_cflag & ~CSTOPB;
    else
      m_termios.c_cflag = m_termios.c_cflag | CSTOPB;
    this->tcsetattr(option);
    m_signal_stop_bits.emit();
  }

  FlowControl TTY::flow_control()
  {
    if ( m_termios.c_cflag & CRTSCTS  &&  !(m_termios.c_cflag & (IXON|IXOFF)) )
      return FLOW_HARDWARE;
    if ( m_termios.c_cflag & (IXON|IXOFF)  &&  !(m_termios.c_cflag&CRTSCTS) )
      return FLOW_SOFTWARE;
    if ( !( m_termios.c_cflag & (CRTSCTS|IXON|IXOFF) ) )
      return FLOW_NONE;
    return FLOW_CUSTOM;
  }

  void TTY::set_flow_control(FlowControl flow_control, SetOption option)
  {
    switch ( flow_control ) {
      case FLOW_NONE:
        m_termios.c_cflag &= ~CRTSCTS;
        m_termios.c_iflag &= ~(IXON|IXOFF);
        break;
      case FLOW_HARDWARE:
        m_termios.c_cflag |= CRTSCTS;
        m_termios.c_iflag &= ~(IXON|IXOFF);
        break;
      case FLOW_SOFTWARE:
        m_termios.c_cflag &= ~CRTSCTS;
        m_termios.c_iflag |= IXON|IXOFF;
        break;
      case FLOW_CUSTOM:
        break;
    }

    this->tcsetattr(option);
    m_signal_flow_control.emit();
  }

  bool TTY::receiver_enabled()
  {
    return m_termios.c_cflag & CREAD;
  }

  void TTY::set_receiver_enabled(bool enable, SetOption option)
  {
    if (enable)
      m_termios.c_cflag |= CREAD;
    else
      m_termios.c_cflag &= ~CREAD;
    this->tcsetattr(option);
    m_signal_receiver.emit();
  }

  bool TTY::carrier_detect_enabled()
  {
    return m_termios.c_cflag & CLOCAL;
  }

  void TTY::set_carrier_detect(bool enable, SetOption option)
  {
    if (enable)
      m_termios.c_cflag &= ~CLOCAL;
    else
      m_termios.c_cflag |= CLOCAL;
    this->tcsetattr(option);
    m_signal_carrier_detect.emit();
  }

  void TTY::set_input_modes(tcflag_t iflag, SetOption option)
  {
    m_termios.c_iflag = iflag;
    this->tcsetattr(option);
    m_signal_input_speed.emit();
    m_signal_parity.emit();
    m_signal_flow_control.emit();
  }

  void TTY::set_output_modes(tcflag_t oflag, SetOption option)
  {
    m_termios.c_oflag = oflag;
    this->tcsetattr(option);
    m_signal_output_speed.emit();
  }

  void TTY::set_control_modes(tcflag_t cflag, SetOption option)
  {
    m_termios.c_cflag = cflag;
    this->tcsetattr(option);
    m_signal_parity.emit();
    m_signal_byte_size.emit();
    m_signal_stop_bits.emit();
    m_signal_flow_control.emit();
    m_signal_receiver.emit();
    m_signal_carrier_detect.emit();
  }

  void TTY::set_local_modes(tcflag_t lflag, SetOption option)
  {
    m_termios.c_lflag = lflag;
    this->tcsetattr(option);
  }

  void TTY::set_control_characters(int index, cc_t value, SetOption option)
  {
    m_termios.c_cc[index] = value;
    this->tcsetattr(option);
  }

  struct termios TTY::attributes()
    {
      struct termios attr;
      tcgetattr(m_fd, &attr);
      return attr;
    }

  bool TTY::reset_on_close() {
    return m_reset_on_close;
  }

  void TTY::set_reset_on_close(bool reset) {
    m_reset_on_close = reset;
  }

  void TTY::tcsetattr( SetOption option )
  {
    int i;
    if ( this->is_open() ) {
      i = ::tcsetattr(m_fd, option, &m_termios);
      if ( i == -1 )
        std::cout << "error from tcsetattr" << std::endl;
    }
  }

  void TTY::inject(char data)
  {
    ::ioctl(m_fd, TIOCSTI, &data);
  }

  void TTY::inject(const char* buffer, size_t bufsize)
  {
    for (size_t i=0; i < bufsize; i++)
      inject(buffer[i]);
  }

  void TTY::inject(const std::string& buffer)
  {
    for (size_t i=0; i < buffer.size(); i++)
      inject(buffer[i]);
  }

  sigc::signal<void> TTY::signal_input_speed() {
    return m_signal_input_speed;
  }

  sigc::signal<void> TTY::signal_output_speed() {
    return m_signal_output_speed;
  }

  sigc::signal<void> TTY::signal_parity() {
    return m_signal_parity;
  }

  sigc::signal<void> TTY::signal_byte_size() {
    return m_signal_byte_size;
  }

  sigc::signal<void> TTY::signal_stop_bits() {
    return m_signal_stop_bits;
  }

  sigc::signal<void> TTY::signal_flow_control() {
    return m_signal_flow_control;
  }

  sigc::signal<void> TTY::signal_carrier_detect() {
    return m_signal_carrier_detect;
  }

  sigc::signal<void> TTY::signal_receiver() {
    return m_signal_receiver;
  }

  speed_t TTY::unsigned2speed(unsigned speed)
  {
    switch (speed)
      {
      case 0:
        return B0;
      case 50:
        return B50;
      case 75:
        return B75;
      case 110:
        return B110;
      case 134:
        return B134;
      case 150:
        return B150;
      case 200:
        return B200;
      case 300:
        return B300;
      case 600:
        return B600;
      case 1200:
        return B1200;
      case 1800:
        return B1800;
      case 2400:
        return B2400;
      case 4800:
        return B4800;
      case 9600:
        return B9600;
      case 19200:
        return B19200;
      case 38400:
        return B38400;
      case 57600:
        return B57600;
      case 115200:
        return B115200;
      case 230400:
        return B230400;
      case 460800:
        return B460800;
      case 500000:
        return B500000;
      case 576000:
        return B576000;
      case 921600:
        return B921600;
      case 1000000:
        return B1000000;
      case 1152000:
        return B1152000;
      case 1500000:
        return B1500000;
      case 2000000:
        return B2000000;
      case 2500000:
        return B2500000;
      case 3000000:
        return B3000000;
      case 3500000:
        return B3500000;
      case 4000000:
        return B4000000;
      default:
        return B0;
      }
  }

  unsigned TTY::speed2unsigned(speed_t speed)
  {
    switch (speed)
      {
      case B0:
        return 0;
      case B50:
        return 50;
      case B75:
        return 75;
      case B110:
        return 110;
      case B134:
        return 134;
      case B150:
        return 150;
      case B200:
        return 200;
      case B300:
        return 300;
      case B600:
        return 600;
      case B1200:
        return 1200;
      case B1800:
        return 1800;
      case B2400:
        return 2400;
      case B4800:
        return 4800;
      case B9600:
        return 9600;
      case B19200:
        return 19200;
      case B38400:
        return 38400;
      case B57600:
        return 57600;
      case B115200:
        return 115200;
      case B230400:
        return 230400;
      case B460800:
        return 460800;
      case B500000:
        return 500000;
      case B576000:
        return 576000;
      case B921600:
        return 921600;
      case B1000000:
        return 1000000;
      case B1152000:
        return 1152000;
      case B1500000:
        return 1500000;
      case B2000000:
        return 2000000;
      case B2500000:
        return 2500000;
      case B3000000:
        return 3000000;
      case B3500000:
        return 3500000;
      case B4000000:
        return 4000000;
      default:
        return 0;
      }
  }

  size_t TTY::input_available() throw() {
    int n;
    ioctl( m_fd, TIOCINQ, &n );
    if ( n >= 0 )
      return n;
    return 0;
  }

}
