/***************************************************************************
 *   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
 *   rvinyard@cs.nmsu.edu                                                  *
 *                                                                         *
 *   This file is part of the conexus library.                             *
 *                                                                         *
 *   The conexus library is free software; you can redistribute it and/or  *
 *   modify it under the terms of the GNU General Public License           *
 *   version 3 as published by the Free Software Foundation.               *
 *                                                                         *
 *   The conexus library is distributed in the hope that it will be        *
 *   useful, but WITHOUT ANY WARRANTY; without even the implied warranty   *
 *   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU   *
 *   General Public License for more details.                              *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this software. If not see <http://www.gnu.org/licenses/>.  *
 ***************************************************************************/
#include "utility.h"

#include <netinet/in.h>
#include <sys/ioctl.h>
#include <net/if.h>

// #include <netpacket/packet.h>
// #include <net/ethernet.h>
// #include <netdb.h>
// #include <sys/socket.h>
// #include <netinet/in.h>
// #include <arpa/inet.h>

#include <glibmm/thread.h>

#include <conexus/file.h>
#include <conexus/ipv4_tcp.h>
#include <conexus/ipv4_tcpserver.h>
#include <conexus/ipv4_udpposet.h>
#include <conexus/ipv4_udp.h>
#include <conexus/ipv6_tcp.h>
#include <conexus/ipv6_tcpserver.h>
#include <conexus/ipv6_udp.h>
#include <conexus/ipv6_udpposet.h>
#include <conexus/ll_packet.h>
#include <conexus/messagequeue.h>
#include <conexus/pipe.h>
#include <conexus/serial.h>
#include <conexus/tty.h>

namespace Conexus
{

  bool __use_dispatcher_default = false;
  bool __initialized = false;
  Factory::pointer __default_factory(Factory::create());
  Registry::pointer __default_registry(Registry::create());


  unsigned interface_index( const std::string& name )
  {
    unsigned index = if_nametoindex( name.c_str() );
    if ( index == 0 )
      throw interface_index_exception();
    return index;
  }

  std::string interface_name( unsigned ifindex )
  {
    char index_name[IF_NAMESIZE];

    if ( index_name != if_indextoname( ifindex, index_name ) )
      throw interface_index_exception();

    return std::string( index_name );
  }

  std::vector<std::pair<unsigned,std::string> > interfaces()
  {
    struct if_nameindex *index, *i;
    std::vector<std::pair<unsigned,std::string> > ifs;

    index = if_nameindex();
    for ( i = index; i != NULL; i++ )
      ifs.push_back( std::pair<unsigned,std::string>(i->if_index, i->if_name) );
    if_freenameindex(index);

    return ifs;
  }

  std::vector<std::string> interface_names()
  {
    struct if_nameindex *index, *i;
    std::vector<std::string> ifs;

    index = if_nameindex();
    for ( i = index; i != NULL; i++ )
      ifs.push_back( std::string(i->if_name) );
    if_freenameindex(index);

    return ifs;
  }

#define MAX_REQ_SIZE 100

  std::vector< std::pair < std::string , std::string > > interface_ip_addresses( )
  {
    std::vector< std::pair<std::string, std::string> > ifaddrs;

    int sd = socket( PF_PACKET, SOCK_RAW, 0 );
    struct ifreq* req;
    struct ifconf conf;
    char buf[ sizeof( struct ifreq ) * MAX_REQ_SIZE ];
    std::pair<std::string, std::string> interface;
    struct sockaddr_in* sa;

    conf.ifc_len = sizeof( buf );
    conf.ifc_buf = ( caddr_t ) buf;

    if ( ioctl( sd, SIOCGIFCONF, ( caddr_t ) & conf ) >= 0 )
      for ( req = ( struct ifreq* ) conf.ifc_buf; ( caddr_t ) req < conf.ifc_buf + conf.ifc_len; req++ ) {
        interface.first = std::string( req->ifr_name );
        if ( req->ifr_addr.sa_family == AF_INET ) {
          sa = ( struct sockaddr_in* ) & ( req->ifr_addr );
          interface.second = std::string( inet_ntoa( sa->sin_addr ) );
          ifaddrs.push_back( interface );
        }
      }

    close( sd );
    return ifaddrs;
  }

  void init( bool use_dispatcher_default )
  {

    __use_dispatcher_default = use_dispatcher_default;
    if ( !Glib::thread_supported() )
      Glib::thread_init();

    __initialized = true;

    Factory& factory = default_factory();

    factory.add( "File", File::generic_create );
    factory.add( "IPv4::TCP", IPv4::TCP::generic_create );
    factory.add( "IPv4::TCPServer", IPv4::TCPServer::generic_create );
    factory.add( "IPv4::UDP", IPv4::UDP::generic_create );
    factory.add( "IPv4::UDPPoset", IPv4::UDPPoset::generic_create );
    factory.add( "IPv6::TCP", IPv4::TCP::generic_create );
    factory.add( "IPv6::TCPServer", IPv4::TCPServer::generic_create );
    factory.add( "IPv6::UDP", IPv4::UDP::generic_create );
    factory.add( "IPv6::UDPPoset", IPv4::UDPPoset::generic_create );
    factory.add( "Packet", LL::Packet::generic_create );
    factory.add( "MessageQueue", MessageQueue::generic_create );
    factory.add( "Pipe", Pipe::generic_create );
    factory.add( "Serial", Serial::generic_create );
    factory.add( "TTY", TTY::generic_create );
  }

  bool initialized()
  {
    return __initialized;
  }

  bool use_dispatcher_default( )
  {
    return __use_dispatcher_default;
  }

  void set_use_dispatcher_default( bool use_dispatcher_default )
  {
    __use_dispatcher_default = use_dispatcher_default;
  }

  void set_default_multicast_hops( unsigned hops )
  {
    IPv4::UDP::set_default_multicast_hops( hops );
    IPv6::UDP::set_default_multicast_hops( hops );
  }

  void set_default_multicast_loopback( bool loopback )
  {
    IPv4::UDP::set_default_multicast_loopback( loopback );
    IPv6::UDP::set_default_multicast_loopback( loopback );
  }

  Factory& default_factory( )
  {
    return *__default_factory;
  }

  bool set_default_factory( Factory::pointer new_factory )
  {
    if ( ! new_factory )
      return false;
    __default_factory = new_factory;
    return true;
  }

  Registry & default_registry( )
  {
    return *__default_registry;
  }

  bool set_default_registry( Registry::pointer new_registry )
  {
    if ( ! new_registry )
      return false;
    __default_registry = new_registry;
    return true;
  }

}

