/* Host and Port Entry
 *
 * Allows the user to enter a host name or an IP address and a port. In addition to providing an entry
 * widget, this widget can optionally display the hostname and IP address beneath the entry widget as
 * well as an optional image that indicates whether the current entry has been found via a DNS lookup.
 *
 * As with all of the Conexus::Gtk widgets, these widgets extend the standard Gtk widgets by allowing
 * a Conexus endpoint widget to be attached. Thus, changes to the endpoint will result in changes
 * to the widget, and vice-versa. In essence, the widget becomes the 'view' on the endpoint 'model'
 * a la the model-view abstraction.
 *
 * This widget allows IPv4 and IPv6 endpoints to be attached.
 *
 * This demo provides three checkboxes to turn on/off the optional features.
 */

#include "gtkmm.h"

#include <conexus-gtkmm.h>

class Example_HostPortEntry : public Gtk::Window
{
  public:
    Example_HostPortEntry();
    virtual ~Example_HostPortEntry();

  protected:
    /* Member widgets */
    Conexus::Gtk::HostPortEntry m_HostPortEntry;

    typedef enum BUTTON{ HOST_LABEL, ADDRESS_LABEL, IMAGE, SERVICE_LABEL } BUTTON;
    Gtk::CheckButton m_host_label_button;
    Gtk::CheckButton m_address_label_button;
    Gtk::CheckButton m_image_button;
    Gtk::CheckButton m_servicename_button;
    Gtk::RadioButton m_layout_vertical;
    Gtk::RadioButton m_layout_horizontal;
    Gtk::VBox m_vbox;

    void on_button_clicked(BUTTON which);
    void on_layout_changed(Conexus::Gtk::HostPortEntry::Layout);
};

/* Called by DemoWindow */
Gtk::Window* do_HostPortEntry()
{
  return new Example_HostPortEntry();
}

Example_HostPortEntry::Example_HostPortEntry()
{
  Gtk::RadioButtonGroup layout_group;
  Gtk::HBox* hbox;

  set_title("Host Entry");
  set_border_width(10);

  /* Add the HostPortEntry widget */
  m_vbox.pack_start( m_HostPortEntry );

  /* And the other widgets for this demo */
  m_vbox.pack_start( *Gtk::manage( new Gtk::HSeparator() ) );
  m_host_label_button.set_label( "Show hostname label" );
  m_address_label_button.set_label( "Show IP address label" );
  m_image_button.set_label( "Show entry validation image" );
  m_servicename_button.set_label( "Show service name label" );
  m_host_label_button.set_active( true );
  m_address_label_button.set_active( true );
  m_image_button.set_active( true );
  m_servicename_button.set_active( true );
  m_host_label_button.signal_toggled().connect( sigc::bind( sigc::mem_fun( *this, &Example_HostPortEntry::on_button_clicked), HOST_LABEL) );
  m_address_label_button.signal_toggled().connect( sigc::bind( sigc::mem_fun( *this, &Example_HostPortEntry::on_button_clicked), ADDRESS_LABEL) );
  m_image_button.signal_toggled().connect( sigc::bind( sigc::mem_fun( *this, &Example_HostPortEntry::on_button_clicked), IMAGE) );
  m_servicename_button.signal_toggled().connect( sigc::bind( sigc::mem_fun( *this, &Example_HostPortEntry::on_button_clicked), SERVICE_LABEL) );
  m_vbox.pack_start( m_host_label_button );
  m_vbox.pack_start( m_address_label_button );
  m_vbox.pack_start( m_image_button );
  m_vbox.pack_start( m_servicename_button );

  m_layout_vertical.set_label( "Vertical" );
  m_layout_vertical.set_group( layout_group );
  m_layout_vertical.set_active();
  m_layout_horizontal.set_label( "Horizontal" );
  m_layout_horizontal.set_group( layout_group );
  m_layout_vertical.signal_toggled().connect( sigc::bind( sigc::mem_fun( *this, &Example_HostPortEntry::on_layout_changed ), Conexus::Gtk::HostPortEntry::VERTICAL ) );
  m_layout_horizontal.signal_toggled().connect( sigc::bind( sigc::mem_fun( *this, &Example_HostPortEntry::on_layout_changed ), Conexus::Gtk::HostPortEntry::HORIZONTAL ) );
  m_vbox.pack_start( *( Gtk::manage( new Gtk::HSeparator() ) ) );
  hbox = Gtk::manage( new Gtk::HBox() );
  hbox->pack_start( *( Gtk::manage( new Gtk::Label( "Layout: " ) ) ) );
  hbox->pack_start( m_layout_vertical );
  hbox->pack_start( *( Gtk::manage( new Gtk::VSeparator() ) ) );
  hbox->pack_start( m_layout_horizontal );
  m_vbox.pack_start( *hbox );

  this->add( m_vbox );

  show_all();
}

Example_HostPortEntry::~Example_HostPortEntry()
{
}

void Example_HostPortEntry::on_button_clicked( BUTTON which ) {
  switch ( which ) {
    case HOST_LABEL:
      m_HostPortEntry.host_entry().set_show_hostname_label( m_host_label_button.get_active() );
      break;
    case ADDRESS_LABEL:
      m_HostPortEntry.host_entry().set_show_address_label( m_address_label_button.get_active() );
      break;
    case IMAGE:
      m_HostPortEntry.host_entry().set_show_host_image( m_image_button.get_active() );
      break;
    case SERVICE_LABEL:
      m_HostPortEntry.port_entry().set_show_servicename_label( m_servicename_button.get_active() );
  }
}

void Example_HostPortEntry::on_layout_changed( Conexus::Gtk::HostPortEntry::Layout layout ) {
    m_HostPortEntry.set_layout( layout );
}

