/* TTY Widget Demonstration
 *
 * This example demonstrates a working TTY widget with full I/O. This example will attempt
 * to configure and access the selected serial port.
 */

#include "gtkmm.h"

#include <conexus-gtkmm.h>

class Example_TTYWidgetDemo : public Gtk::Window
{
  public:
    Example_TTYWidgetDemo();
    virtual ~Example_TTYWidgetDemo();

  protected:
    /* Member widgets */
    Conexus::TTY::pointer m_serial; /* A serial endpoint to transmit on */
    Conexus::Gtk::TTYWidget m_TTYWidgetDemo; /* A configuration widget */

    /* Other Gtk widgets */
    Gtk::Button m_serial_transmit; /* Every time this button is clicked 0123456789 is transmitted */
    Gtk::TextView m_serial_transmit_history; /* An output text window showing the user when the transmits occurred */
    Gtk::TextView m_serial_receive_history; /* An output text window of everything received */
    Gtk::VBox m_vbox; /* Used to stack the widgets in the main window */
    Gtk::Frame m_transmit_frame, m_receive_frame;
    Gtk::ScrolledWindow m_transmit_scroll, m_receive_scroll;

    /* Callbacks */
    void on_serial_transmit(); /* Updates the transmit history whenever anything is transmitted */
    void on_serial_receive( Conexus::Data d ); /* Updates the receive history whenever anything is received */
    void on_device_set();

};

/* Called by DemoWindow */
Gtk::Window* do_TTYWidget_demo()
{
  return new Example_TTYWidgetDemo();
}

Example_TTYWidgetDemo::Example_TTYWidgetDemo()
{
  m_serial = Conexus::TTY::create();

  /* Connect a serial endpoint to the widget and add to main box */
  m_TTYWidgetDemo.connect( m_serial );
  m_vbox.add(m_TTYWidgetDemo);
  m_vbox.pack_start( *Gtk::manage( new Gtk::HSeparator() ) ); /* Put a separator after this widget */

  /* Setup the transmit button and callback */
  m_serial_transmit.set_label( "Transmit serial data" );
  m_serial_transmit.signal_clicked().connect( sigc::mem_fun( *this, &Example_TTYWidgetDemo::on_serial_transmit ) );
  m_vbox.pack_start( m_serial_transmit );
  m_vbox.pack_start( *Gtk::manage( new Gtk::HSeparator() ) );

  /* Setup the transmit box */
  m_serial_transmit_history.set_editable(false);
  m_transmit_scroll.set_size_request(-1, 150);
  m_transmit_scroll.set_policy( Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC );
  m_transmit_scroll.add( m_serial_transmit_history );
  m_transmit_frame.set_label( "Transmit Status" );
  m_transmit_frame.add( m_transmit_scroll );
  m_vbox.pack_start( m_transmit_frame );
  m_vbox.pack_start( *Gtk::manage( new Gtk::HSeparator() ) );

  /* Setup the receive box */
  m_serial_receive_history.set_editable(false);
  m_receive_scroll.set_size_request(-1, 150);
  m_receive_scroll.set_policy( Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC );
  m_receive_scroll.add( m_serial_receive_history );
  m_receive_frame.set_label( "Received data" );
  m_receive_frame.add( m_receive_scroll );
  m_vbox.pack_start( m_receive_frame );
  m_serial->signal_device_changed().connect( sigc::mem_fun( *this, &Example_TTYWidgetDemo::on_device_set ) );
  m_serial->signal_data().connect( sigc::mem_fun( *this, &Example_TTYWidgetDemo::on_serial_receive ) );

  this->add( m_vbox );

  show_all();
}

Example_TTYWidgetDemo::~Example_TTYWidgetDemo()
{
}

void Example_TTYWidgetDemo::on_serial_transmit( ) {
  /* We need some data to transmit */
  char data[] = "0123456789\n";

  /* Get the Gtk TextBuffer widget buffer */
  Glib::RefPtr<Gtk::TextBuffer> buffer = m_serial_transmit_history.get_buffer();

  /* Get the current time and build a result string */
  time_t t = time(NULL);
  std::string status = ctime(&t);
  status.erase(status.end()-1);
  status += ": ";
  if ( m_serial->device() == std::string() ) {
    buffer->insert( buffer->end(), status + std::string("Device not set\n") );
    return;
  }

  /* Actually output the data  */
  /* This is the only line really needed here to transmit, the rest is just for the GUI */
  m_serial->write( data, strlen( data ) );

  /* Update the buffer when the transmit is successful */
  status += std::string("Transmitted on ") + m_serial->device() + std::string("\n");
  buffer->insert( buffer->end(), status );
}

void Example_TTYWidgetDemo::on_serial_receive( Conexus::Data d ) {
  Glib::RefPtr<Gtk::TextBuffer> buffer = m_serial_receive_history.get_buffer();
  buffer->insert( buffer->end(), (const char*)(const uint8_t*)d, (const char*)(const uint8_t*)d + d.size() );
}

void Example_TTYWidgetDemo::on_device_set() {
  Glib::RefPtr<Gtk::TextBuffer> buffer = m_serial_receive_history.get_buffer();
  time_t t;
  std::string status;

  if (m_serial->is_running())
    m_serial->stop();

  if ( m_TTYWidgetDemo.device_widget().device() == std::string() ) {
    m_serial->close();
    return;
  }

  t = time(NULL);
  status = ctime(&t);
  status.erase(status.end()-1);
  m_serial->start();
  status += std::string(": Receiving data on ") + m_serial->device() + std::string("\n");
  buffer->insert( buffer->end(), status );
}


